@@ -10,6 +10,7 @@
 
 #include "config-options.h"
 #include "po/gettext.h"
+#include "hackerlab/bugs/exception.h"
 #include "hackerlab/cmd/main.h"
 #include "libarch/patch-logs.h"
 #include "libarch/namespace.h"
@@ -31,8 +32,6 @@ static t_uchar * usage = N_("[options] [limit]");
   OP (opt_version, "V", "version", 0, \
       N_("Display a release identifier string\n" \
       "and exit.")) \
-  OP (opt_archive, "A", "archive", 1, \
-      N_("Override `my-default-archive'")) \
   OP (opt_reverse, "r", "reverse", 0, \
       N_("sort revisions from newest to oldest")) \
   OP (opt_summary, "s", "summary", 0, \
@@ -105,7 +104,6 @@ arch_cmd_abrowse (t_uchar * program_name, int argc, char * argv[])
 {
   int o;
   struct opt_parsed * option;
-  t_uchar * default_archive = 0;
   int reverse = 0;
   int wanted_headers = 0;
   int show_kind = 0;
@@ -141,12 +139,6 @@ arch_cmd_abrowse (t_uchar * program_name, int argc, char * argv[])
           safe_printfmt (2, "ill-formed argument for `%s' (`%s')\n", option->opt_string, option->arg_string);
           goto usage_error;
 
-        case opt_archive:
-          {
-            default_archive = str_save (0, option->arg_string);
-            break;
-          }
-
         case opt_reverse:
           {
             reverse = 1;
@@ -271,12 +263,8 @@ arch_cmd_abrowse (t_uchar * program_name, int argc, char * argv[])
   if (argc > 2)
     goto usage_error;
 
-  default_archive = arch_my_default_archive (default_archive);
-
-
   {
     t_uchar * limit_spec = 0;
-    t_uchar * archive = 0;
     t_uchar * limit_category = 0;
     t_uchar * limit_branch = 0;
     t_uchar * limit_version = 0;
@@ -300,25 +288,37 @@ arch_cmd_abrowse (t_uchar * program_name, int argc, char * argv[])
         limit_spec = str_save (0, argv[1]);
       }
 
-    if (limit_spec && arch_valid_archive_name (limit_spec))
+    if (!limit_spec)
+	limit_spec = arch_my_default_archive (NULL);
+
+    if (!limit_spec)
       {
-        archive = limit_spec;
-        limit_spec = 0;
-      }
-    else if (limit_spec && !arch_valid_package_name (limit_spec, arch_maybe_archive, arch_req_category, 1))
-      {
-        safe_printfmt (1, "abrowse: invalid limit spec (%s)\n", limit_spec);
-        exit (1);
-      }
-    else if (limit_spec)
-      {
-        archive = arch_parse_package_name (arch_ret_archive, default_archive, limit_spec);
-      }
-    else
-      {
-        archive = arch_my_default_archive (default_archive);
+	arch_project_tree_t * tree = arch_project_tree_new (talloc_context, ".");
+	if (tree->archive)
+	    limit_spec = str_save (0, tree->archive);
+	arch_project_tree_delete (tree);
       }
 
+    if (!str_length (limit_spec))
+	goto usage_error;
+
+      {
+	/* TODO RBC 20050328 check the project tree */
+	t_uchar * temp_spec = NULL;
+	arch = arch_archive_connect_branch (limit_spec, &temp_spec);
+	if (!arch)
+	  {
+	    safe_printfmt (2, "Could not connect to archive for browsing '%s'\n", limit_spec);
+	    exit (2);
+	  }
+	lim_free (0, limit_spec);
+	if (arch_valid_package_name (temp_spec, arch_req_archive, arch_req_category, 1))
+	    limit_spec = arch_parse_package_name (arch_ret_non_archive, NULL, temp_spec);
+	else
+	    limit_spec = NULL;
+	lim_free (0, temp_spec);
+      }
+	
     /********************************
      * sanity check
      */
@@ -430,18 +430,11 @@ arch_cmd_abrowse (t_uchar * program_name, int argc, char * argv[])
       }
 
 
-    /********************************
-     * connect to the archive
-     */
-
-    arch = arch_archive_connect (archive, 0);
-
-
     /****************************************************************
      * produce output
      */
 
-    safe_printfmt (1, "%s\n", archive);
+    safe_printfmt (1, "%s\n", arch->official_name);
     safe_flush (1);
 
     {
@@ -611,26 +604,26 @@ arch_cmd_abrowse (t_uchar * program_name, int argc, char * argv[])
 					  int is_cached;
 					  int has_ancestry;
                                           int s;
+                                          arch_patch_id * pkg_revision_patch;
 
                                           version_stem = str_alloc_cat (0, versions[v][0], "--");
 
                                           package_revision = str_alloc_cat (0, version_stem, revisions[r][0]);
-                                          arch_revision_type (&this_type, &is_cached, &has_ancestry, arch, package_revision);
+                                          pkg_revision_patch = arch_patch_id_new_archive (arch->official_name, package_revision);
+                                          arch_revision_type (&this_type, &is_cached, &has_ancestry, arch, pkg_revision_patch);
 
                                           if (this_type == arch_continuation_revision)
                                             {
-                                              t_uchar * tagged = 0;
-
-                                              tagged = arch_get_continuation (arch, package_revision);
+                                              arch_patch_id * tagged = arch_get_continuation (arch, pkg_revision_patch);
 
 					      safe_printfmt (1, "        %s", revisions[r][0]);
 					      if (show_kind)
-						safe_printfmt (1, " (tag of %s)", tagged);
+						safe_printfmt (1, " (tag of %s)", arch_patch_id_patch_id (tagged));
 					      if (show_cacherevs)
 						safe_printfmt (1, "%s", " [cached]");
 					      safe_printfmt (1, "\n");
 
-                                              lim_free (0, tagged);
+                                              talloc_free (tagged);
                                             }
                                           else
                                             {
@@ -639,10 +632,13 @@ arch_cmd_abrowse (t_uchar * program_name, int argc, char * argv[])
                                                   enum arch_revision_type s_type;
 						  int s_is_cached;
 						  int has_ancestry;
+                                                  arch_patch_id * pkg_revision_patch;
 
                                                   lim_free (0, package_revision);
                                                   package_revision = str_alloc_cat (0, version_stem, revisions[s][0]);
-                                                  arch_revision_type (&s_type, &s_is_cached, &has_ancestry, arch, package_revision);
+                                                  pkg_revision_patch = arch_patch_id_new_archive (arch->official_name, package_revision);
+                                                  arch_revision_type (&s_type, &s_is_cached, &has_ancestry, arch, pkg_revision_patch);
+                                                  talloc_free (pkg_revision_patch);
 
 						  if ((show_cacherevs && (s_is_cached != is_cached))
 						      || (show_kind && (s_type != this_type)))
@@ -677,6 +673,7 @@ arch_cmd_abrowse (t_uchar * program_name, int argc, char * argv[])
 
                                           lim_free (0, version_stem);
                                           lim_free (0, package_revision);
+                                          talloc_free (pkg_revision_patch);
                                         }
 
                                       safe_flush (1);
@@ -689,7 +686,7 @@ arch_cmd_abrowse (t_uchar * program_name, int argc, char * argv[])
                                       assoc_table headers = 0;
 
                                       rev = str_alloc_cat_many (0, versions[v][0], "--", revisions[r][0], str_end);
-				      fqrev = arch_fully_qualify (archive, rev);
+				      fqrev = arch_fully_qualify (arch->official_name, rev);
                                       if (!show_kind && !show_cacherevs)
                                         {
                                           safe_printfmt (1, "        %s\n", full ? fqrev : revisions[r][0]);
@@ -699,8 +696,9 @@ arch_cmd_abrowse (t_uchar * program_name, int argc, char * argv[])
                                           enum arch_revision_type type;
 					  int is_cached;
 					  int has_ancestry;
+                                          arch_patch_id * revision = arch_patch_id_new_archive (arch->official_name, rev);
 
-                                          arch_revision_type (&type, &is_cached, &has_ancestry, arch, rev);
+                                          arch_revision_type (&type, &is_cached, &has_ancestry, arch, revision);
 
 					  safe_printfmt (1, "        %s   ", full ? fqrev : revisions[r][0]);
 					  if (show_kind)
@@ -710,21 +708,22 @@ arch_cmd_abrowse (t_uchar * program_name, int argc, char * argv[])
 						case arch_import_revision:
 						  {
 						    safe_printfmt (1, "(initial import)");
+                                                    talloc_free (revision);
 						    break;
 						  }
 						case arch_simple_revision:
 						  {
 						    safe_printfmt (1, "(simple changeset)");
+                                                    talloc_free (revision);
 						    break;
 						  }
 						case arch_continuation_revision:
 						  {
-						    t_uchar * continuation_of = 0;
+						    arch_patch_id * continuation_of = arch_get_continuation (arch, revision);
+						    safe_printfmt (1, "(tag revision of %s)", arch_patch_id_patch_id (continuation_of));
 
-						    continuation_of = arch_get_continuation (arch, rev);
-						    safe_printfmt (1, "(tag revision of %s)", continuation_of);
-
-						    lim_free (0, continuation_of);
+						    talloc_free (continuation_of);
+                                                    talloc_free (revision);
 						    break;
 						  }
 						}
@@ -806,7 +805,6 @@ arch_cmd_abrowse (t_uchar * program_name, int argc, char * argv[])
 
 
     lim_free (0, limit_spec);
-    lim_free (0, archive);
     lim_free (0, limit_category);
     lim_free (0, limit_branch);
     lim_free (0, limit_version);
