@@ -20,5 +20,5 @@ extern t_uchar arch_cmd_diff_help[];
 
 
 extern int arch_cmd_diff (t_uchar * program_name, int argc, char * argv[]);
-extern char * arch_diff_default_output_dir(char const *tree_root, t_uchar * dirname);
+extern t_uchar * arch_diff_default_output_dir(void * context, t_uchar const *tree_root, t_uchar * dirname);
 #endif  /* INCLUDE__LIBARCH__CMD_DIFF_H */
