@@ -69,8 +69,6 @@ OP (opt_log_date, "D", 0 , 0, \
 OP (opt_patch_regex, 0, "patch-regex REGEX", 1, \
     N_("Only show revisions that contain [REGEX]\n" \
     "(implies -s)")) \
-OP (opt_archive, "A", "archive", 1, \
-    N_("Use [archive] instead of default")) \
 OP (opt_unhide_sealed, 0, "show-sealed", 0, \
     N_("Do not hide sealed branches")) 
 
@@ -116,7 +114,7 @@ arch_cmd_rbrowse (t_uchar * program_name, int argc, char * argv[])
   t_uchar * errname;
   t_uchar * snap_file = NULL;
   t_uchar * since_filename = NULL;
-  t_uchar * default_archive = NULL;
+  t_uchar * archive_name = NULL;
   t_uchar * patch_regex = NULL;
   t_uchar * limit_regex = NULL;
   rel_table since_output = NULL;
@@ -169,11 +167,6 @@ arch_cmd_rbrowse (t_uchar * program_name, int argc, char * argv[])
               since_filename = str_save(0, option->arg_string);
               break;
             }
-        case opt_archive:
-            {
-              default_archive = str_save(0, option->arg_string);
-              break;
-            }
         case opt_log_summary:
             {
               summarized_headers |= arch_include_summary;
@@ -225,19 +218,21 @@ arch_cmd_rbrowse (t_uchar * program_name, int argc, char * argv[])
     limit_regex = str_save (0, ".*");
   else
     {
-      t_uchar *slash;
-
       /* Did the user give us something resembling archive/regex ? */
-      if ( 0 == (slash = str_chr_index (argv[1], '/')))
+      if ( 0 == str_chr_index (argv[1], '/'))
         limit_regex = str_save (0, argv[1]);
       else
         {
-
-          if (default_archive)
-            panic ("Archive specified twice! ");
-
-          limit_regex = str_save(0, argv[1]);
-          default_archive = str_separate (&limit_regex, "/");
+	  /* get the real namespace */
+	  struct arch_archive * arch = arch_archive_connect_branch (argv[1], &archive_name);
+	  if (!arch)
+	    {
+	      safe_printfmt (2, "Cannot connect to archive %s\n", argv[1]);
+	      exit (2);
+	    }
+	  arch_archive_close (arch);
+	  limit_regex = archive_name;
+	  archive_name = str_separate (&limit_regex, "/");
 
           if (! limit_regex)
             limit_regex = str_save (0, ".*");
@@ -245,8 +240,8 @@ arch_cmd_rbrowse (t_uchar * program_name, int argc, char * argv[])
         }
     }
 
-  if (!default_archive)
-    default_archive = arch_my_default_archive (default_archive);
+  if (!archive_name)
+    archive_name = arch_my_default_archive (NULL);
 
   if (regcomp(&limit_regex_needle, limit_regex, REG_EXTENDED))
     panic ("Invalid regular expression given for limit.");
@@ -260,12 +255,12 @@ arch_cmd_rbrowse (t_uchar * program_name, int argc, char * argv[])
   if (since_filename)
     saved_since = rbrowse_load_since (since_filename);
 
-  if (!arch_valid_archive_name (default_archive))
+  if (!arch_valid_archive_name (archive_name))
     panic ("Invalid archive given");
 
   if (! all_archives)
     {
-      rel_add_records (&archive_list, rel_make_record (default_archive, 0), 0);
+      rel_add_records (&archive_list, rel_make_record (archive_name, 0), 0);
     }
   else
     {
@@ -281,12 +276,12 @@ arch_cmd_rbrowse (t_uchar * program_name, int argc, char * argv[])
       struct arch_archive * arch = 0;
       rel_table categories = 0;
 
-
+      /* FIXME-REMOVENAMES this can go in baz 1.4 RBC 20050329 */
       if (str_cmp("-MIRROR", str_chr_rindex(archive_list[w][0], '-')) == 0 ||
           str_cmp("-SOURCE", str_chr_rindex(archive_list[w][0], '-')) == 0 )
         continue;
 
-      arch = arch_archive_connect(archive_list[w][0], 0);
+      arch = arch_archive_connect_branch (archive_list[w][0], NULL);
 
       categories = arch_archive_categories(arch);
 
@@ -369,7 +364,7 @@ arch_cmd_rbrowse (t_uchar * program_name, int argc, char * argv[])
                       if (!shown_archive)
                         {
                           shown_archive = 1;
-                          safe_printfmt (1,"%s\n", archive_list[w][0]);
+                          safe_printfmt (1,"%s\n", arch->official_name);
                         }
                       if (!shown_category)
                         {
