@@ -30,14 +30,14 @@ main (int argc, char * argv[])
 {
   char ** array = NULL;
 
-  ar_push_charstar(&array, "foo");
-  ar_push_charstar(&array, "bar");
+  ar_push_char_star(&array, "foo");
+  ar_push_char_star(&array, "bar");
 
-  invariant(2 == ar_size ((void *)array, 0, sizeof (char *)));
+  invariant(2 == ar_size_char_star (array));
   invariant(!str_cmp(array[0], "foo"));
   invariant(!str_cmp(array[1], "bar"));
 
-  ar_free ((void **)&array, 0);
+  ar_free_char_star (&array);
 
   return 0;
 }
