@@ -11,10 +11,11 @@
  */
 
 
+#include "hackerlab/bugs/exception.h"
 #include "hackerlab/bugs/panic.h"
-#include "hackerlab/arrays/ar.h"
 #include "hackerlab/char/str.h"
 #include "hackerlab/fs/file-names.h"
+#include "hackerlab/os/errno.h"
 #include "hackerlab/os/sys/wait.h"
 #include "hackerlab/vu/safe.h"
 #include "libawk/associative.h"
@@ -27,84 +28,69 @@
 #include "libarch/libraries.h"
 #include "libarch/namespace.h"
 #include "libarch/merge-points.h"
-#include "libarch/patch-id.h"
 #include "libarch/ancestry.h"
 
 
-/* TODO this belongs as the underlying mechanism for arch_archive_connect.. name, 0, ?, ? */
-typedef struct _arch_archive_connection_cache
-{
-    struct arch_archive **cache;
-} arch_archive_connection_cache;
-
-static void arch_archive_connection_cache_init (arch_archive_connection_cache * cache);
-static void arch_archive_connection_cache_finalise (arch_archive_connection_cache * cache, struct arch_archive * dontclose);
-static struct arch_archive * arch_archive_connection_cache_find_or_maybe_connect (arch_archive_connection_cache * cache, t_uchar * name, int soft_errors);
-static void arch_archive_connection_cache_add (arch_archive_connection_cache * cache, struct arch_archive * arch);
-static t_uchar * arch_ancestry_from_revision (t_uchar const * patch_id, int soft_errors);
-static void arch_ancestry_note_ancestor (t_uchar const *patch_id, t_uchar const *ancestor);
+static arch_patch_id * arch_ancestry_from_revision (void * context, arch_patch_id * patch_id, int soft_errors);
 
 
 
 rel_table
-arch_trace_ancestry (struct arch_archive * arch, t_uchar * archive, t_uchar * revision, int merges)
+arch_trace_ancestry (void * context, struct arch_archive * arch, arch_patch_id * starting_patch_id, int merges)
 {
   rel_table answer = 0;
   int done = 0;
   arch_archive_connection_cache cache;
+  arch_patch_id * patch_id = arch_patch_id_copy (starting_patch_id);
 
   arch_archive_connection_cache_init (&cache);
-  archive = str_save (0, archive);
-  revision = str_save (0, revision);
 
   if (arch)
       arch_archive_connection_cache_add (&cache, arch);
 
   while (!done)
     {
-      t_uchar * fq_rev = 0;
       struct arch_archive * this_arch = 0;
 
-      fq_rev = arch_fully_qualify (archive, revision);
-      this_arch = arch_archive_connection_cache_find_or_maybe_connect (&cache, archive, 0);
+      this_arch = arch_archive_connection_cache_find_or_maybe_connect (&cache, arch_patch_id_archive (patch_id), 0);
 
       if (!this_arch)
         {
-          rel_add_records (&answer, rel_make_record (fq_rev, "???", 0), 0);
+          rel_add_records (&answer, rel_make_record (arch_patch_id_patch_id(patch_id), "???", 0), 0);
           done = 1;
         }
       else
         {
-          t_uchar * ancestor = 0;
+          arch_patch_id * ancestor = NULL;
 
           if (!merges)
-            rel_add_records (&answer, rel_make_record (fq_rev, fq_rev, 0), 0);
+            rel_add_records (&answer, rel_make_record (arch_patch_id_patch_id(patch_id), arch_patch_id_patch_id(patch_id), 0), 0);
           else
             {
               enum arch_revision_type type;
 
               /* FIXME: grab ancestry here if it exists */
-              arch_revision_type (&type, 0, NULL, this_arch, revision);
+              arch_revision_type (&type, 0, NULL, this_arch, patch_id);
 
               if (type == arch_continuation_revision)
                 {
-                  rel_add_records (&answer, rel_make_record (fq_rev, fq_rev, 0), 0);
+                  rel_add_records (&answer, rel_make_record (arch_patch_id_patch_id(patch_id), arch_patch_id_patch_id(patch_id), 0), 0);
                 }
               else
                 {
                   rel_table merges = 0;
                   int x;
 
-                  merges = arch_archive_merge_points (this_arch, revision, 0, 0, 0);
+                  merges = arch_archive_merge_points (this_arch, arch_patch_id_revision (patch_id), 0, 0, 0);
                   for (x = 0; x < rel_n_records (merges); ++x)
                     {
-                      rel_add_records (&answer, rel_make_record (fq_rev, merges[x][1], 0), 0);
+                      rel_add_records (&answer, rel_make_record (arch_patch_id_patch_id(patch_id), merges[x][1], 0), 0);
                     }
                   rel_free_table (merges);
                 }
             }
 
-          ancestor = arch_patch_ancestor (fq_rev, 0, 0);
+          ancestor = arch_patch_ancestor (talloc_context, patch_id, 0);
 
           if (!ancestor)
             {
@@ -112,81 +98,20 @@ arch_trace_ancestry (struct arch_archive * arch, t_uchar * archive, t_uchar * re
             }
           else
             {
-              lim_free (0, archive);
-              lim_free (0, revision);
-
-              archive = arch_parse_package_name (arch_ret_archive, 0, ancestor);
-              revision = arch_parse_package_name (arch_ret_non_archive, 0, ancestor);
+              talloc_free (patch_id);
+              patch_id = ancestor;
             }
-
-          lim_free (0, ancestor);
         }
 
-      lim_free (0, fq_rev);
     }
   
   arch_archive_connection_cache_finalise (&cache, arch);
-
-  lim_free (0, archive);
-  lim_free (0, revision);
+  talloc_steal (context, ar_base(answer));
   return answer;
 }
 
 
 
-void
-arch_archive_connection_cache_init (arch_archive_connection_cache * cache)
-{
-    cache->cache = NULL;
-}
-
-void 
-arch_archive_connection_cache_finalise (arch_archive_connection_cache * cache, struct arch_archive * dontclose)
-{
-    int position;
-
-    for (position = 0; position < ar_size ((void *)cache->cache, 0, sizeof (struct arch_archive *)); ++position)
-      {
-        if (cache->cache[position] != dontclose)
-          arch_archive_close (cache->cache[position]);
-      }
-    ar_free ((void **)&(cache->cache), 0);
-    cache->cache = NULL;
-}
-
-static struct arch_archive *
-arch_archive_connection_cache_find_or_maybe_connect (arch_archive_connection_cache * cache, t_uchar *name, int soft_errors)
-{
-  int position;
-  t_uchar * loc = 0;
-  struct arch_archive * arch = 0;
-
-  for (position = 0; position < ar_size (cache->cache, 0, sizeof (struct arch_archive *)); ++position)
-    {
-      if (!str_cmp (name, cache->cache[position]->name))
-        return (cache->cache)[position];
-    }
-
-  loc = arch_archive_location (name, 1);
-
-  if (loc)
-    {
-      arch = arch_archive_connect_ext (name, 0, soft_errors);
-      invariant (!arch || !str_cmp (name, arch->official_name));
-      if (arch)
-          arch_archive_connection_cache_add (cache, arch);
-    }
-
-  lim_free (0, loc);
-  return arch;
-}
-
-void
-arch_archive_connection_cache_add (arch_archive_connection_cache * cache, struct arch_archive * arch)
-{
-  *(struct arch_archive **)ar_push ((void **)&(cache->cache), 0, sizeof (struct arch_archive *)) = arch;
-}
-
 static int
 ancestry_parse_baz_1_is_patch_id (t_uchar const *record)
 {
@@ -243,11 +168,11 @@ ancestry_parse(t_uchar const *content)
 }
 
 static rel_table
-partial_ancestry_from_full(rel_table full, t_uchar const *patch_id)
+partial_ancestry_from_full(rel_table full, arch_patch_id * patch_id)
 {
     rel_table result = NULL;
     int row = 0;
-    while (row < rel_n_records (full) && str_cmp (patch_id, full[row][0]))
+    while (row < rel_n_records (full) && str_cmp (arch_patch_id_patch_id (patch_id), full[row][0]))
            ++row;
     for (; row < rel_n_records (full); ++row)
       {
@@ -259,59 +184,29 @@ partial_ancestry_from_full(rel_table full, t_uchar const *patch_id)
 static assoc_table arch_ancestors = NULL;
 static arch_archive_connection_cache global_cache = { NULL };
 
-static struct arch_archive *
-cached_archive_connection (t_uchar const *patch_id, int soft_errors)
-{
-    t_uchar * archive = arch_parse_package_name(arch_ret_archive, 0, patch_id);
-    struct arch_archive *arch = NULL;
-    t_uchar * location;
-    location = arch_archive_location (archive, soft_errors);
-    if (location)
-        arch = arch_archive_connection_cache_find_or_maybe_connect (&global_cache, archive, soft_errors);
-    lim_free (0, archive);
-    lim_free (0, location);
-    return arch;
-}
-
-/* when the archive caches moves to archive.c, this moves to 
- * cached_archive.c
+/**
+ * \brief get a cached patch ancestor.
+ * \throw ENOENT if there is no cached entry
+ * NULL for terminal revisions
+ * a patch id for a specific ancestor.
  */
-static t_uchar *
-arch_patch_id_query (t_uchar const * patch_id, t_uchar const *key)
+static arch_patch_id *
+arch_cached_patch_ancestor (void * context, arch_patch_id * patch_id)
 {
-  arch_patch_id patch; 
-  t_uchar *result = NULL;
-  
-  arch_patch_id_init (&patch, patch_id); 
-  
-  result = arch_cache_query_new (&patch, key);
-  
-  arch_patch_id_finalise (&patch);
-  return result;
-}
-
-/* returns:
- * PARTIAL for no known ancestory
- * NULL for no ancestor
- * other string for a specific ancestor.
- */
-static t_uchar *
-arch_cached_patch_ancestor (t_uchar const * patch_id)
-{
-    t_uchar *result = assoc_ref (arch_ancestors, (t_uchar *)patch_id);
-    invariant (str_cmp (patch_id, ARCH_ANCESTRY_PATCH_PARTIAL));
+    t_uchar *result = assoc_ref (arch_ancestors, arch_patch_id_patch_id (patch_id));
+    invariant (str_cmp (arch_patch_id_patch_id (patch_id), ARCH_ANCESTRY_PATCH_PARTIAL));
     if (result && str_cmp (result, ARCH_ANCESTRY_PATCH_PARTIAL))
       {
-        /* penultimate ancestor */
         if (!str_length (result))
           {
+            /* penultimate ancestor - "" in the hash */
             return NULL;
           }
-        return str_save(0, result);
+        return talloc_steal (context, arch_patch_id_new (result));
       }
     /* memory cache miss */
       {
-        t_uchar * ancestor_query = arch_patch_id_query (patch_id, "ancestor");
+        t_uchar * ancestor_query = arch_cache_query_new (patch_id, "ancestor");
         t_uchar * temp_result = NULL;
         if (ancestor_query && arch_cache_has_answer (ancestor_query))
           {
@@ -320,88 +215,109 @@ arch_cached_patch_ancestor (t_uchar const * patch_id)
             result = trim_surrounding_ws (temp_result);
           }
         lim_free (0, ancestor_query);
-        if (!result)
+        if (!result || !str_cmp (result, ARCH_ANCESTRY_PATCH_PARTIAL))
+          {
             lim_free (0, temp_result);
+            /* not there either, return partial */
+            Throw (exception (ENOENT, "No cache entry"));
+          }
         else
           {
-            result = str_save (0, result);
-            lim_free (0, temp_result);
-            /* penultimate ancestor */
+            arch_patch_id * answer;
             if (!str_length (result))
-              {
-                lim_free (0, result);
-                return NULL;
-              }
-            return result;
+                answer = NULL;
+            else
+                answer = arch_patch_id_new (result);
+            lim_free (0, temp_result);
+            return talloc_steal (context, answer);
           }
       }
-    /* not there either, return partial */
-    return str_save (0, ARCH_ANCESTRY_PATCH_PARTIAL);
 }
 
-/* refactored arch_ancestor_revision
- * will:
- * exit on failure unless soft_errors is set.
- * return PARTIAL on failures otherwise.
- * return the ancestor patch id, and 
- * return NULL for the terminal
- * ancestor.
+
+/** 
+ * \brief find the ancestory for patch_id
  *
  * find the ancestor for patch_id from:
  *   - known ancestry data
  *   - the archive
  *
  * maintains the in-core ancestry cache.
+ * \throw ENOENT if the ancestor cannot be found, or any unexpected child 
+ * function exceptions raised during its operation.
+ * \param context the talloc context to put results in
+ * \param patch_id the patch to get the ancestor of
+ * \param cached_only - only look in the baz cache.
+ * \return the patch id. NULL is returned if there is no ancestor.
  */
-
-t_uchar *
-arch_patch_ancestor (t_uchar const * patch_id, int soft_errors, int cached_only)
+arch_patch_id *
+arch_patch_ancestor (void * context, arch_patch_id * patch_id,  int cached_only)
 {
     /* cache lookup */
-    t_uchar *result = arch_cached_patch_ancestor (patch_id);
-    if (!result || (str_cmp (result, ARCH_ANCESTRY_PATCH_PARTIAL) ))
-        /* penultimate ancestor or known ancestor */
-        return result;
-    lim_free (0, result);
-    result = NULL;
-    /* cache miss */
-    if (!cached_only)
+    arch_patch_id * answer = NULL;
+    t_uchar *result = NULL;
+    struct arch_archive *arch;
+    int found = -1;
+    struct exception * e;
+    Try
       {
-        result = arch_ancestry_from_revision (patch_id, soft_errors);
-        if (!result || (str_cmp (result, ARCH_ANCESTRY_PATCH_PARTIAL)))
-            /* penultimate ancestor or known ancestor */
-            return result;
-        lim_free (0, result);
-        result = NULL;
-        /* old api .. */
-        struct arch_archive *arch = cached_archive_connection (patch_id, soft_errors);
-
-        if (arch)
-          {
-            t_uchar * revision = arch_parse_package_name(arch_ret_non_archive, 0, patch_id);
-            /* FIXME: teach arch_ancestor_revision about soft errors */
-            result = arch_ancestor_revision (arch, revision);
-            lim_free (0, revision);
-            if (!result) 
-              {
-                /* penultimate ancestor */
-                arch_ancestry_note_ancestor (patch_id, "");
-                return NULL;
-              }
-            else /* when ancestor_revision does soft errors..
-                    if (str_cmp (result, ARCH_ANCESTRY_PATCH_PARTIAL)) */
-              {
-                arch_ancestry_note_ancestor (patch_id, result);
-                return result;
-              }
-          }
-        
+        answer = arch_cached_patch_ancestor (context, patch_id);
+        found = 1;
       }
-    /* no cache hit, and couldn't get anything from the archive */
-    if (soft_errors)
-        return str_save (0, ARCH_ANCESTRY_PATCH_PARTIAL);
-    else
-        exit(2);
+    Catch (e)
+      {
+        /* dont mask critical errors */
+        if (e->code < 0)
+            Throw (e);
+        talloc_free (e);
+        if (cached_only)
+            Throw (exception (ENOENT, "Could not retrieve ancestor"));
+        found = 0;
+      }
+    if (found)
+        return talloc_steal (context, answer);
+
+    /* cache miss */
+    Try 
+      {
+        answer = arch_ancestry_from_revision (context, patch_id, 1);
+        found = 1;
+      }
+    Catch (e)
+      {
+        /* dont mask critical errors */
+        if (e->code < 0)
+            Throw (e);
+        talloc_free (e);
+        found = 0;
+      }
+    if (found)
+        return talloc_steal (context, answer);
+
+    /* old api .. */
+    arch = arch_archive_connection_cache_find_or_maybe_connect (&global_cache, arch_patch_id_archive (patch_id), 1);
+
+    if (arch)
+      {
+        /* FIXME: teach arch_ancestor_revision about soft errors */
+        result = arch_ancestor_revision (arch, arch_patch_id_revision (patch_id));
+        if (!result) 
+          {
+            /* penultimate ancestor */
+            arch_ancestry_note_ancestor (patch_id, "");
+            return NULL;
+          }
+        else /* when ancestor_revision does soft errors..
+                if (str_cmp (result, ARCH_ANCESTRY_PATCH_PARTIAL)) */
+          {
+            arch_ancestry_note_ancestor (patch_id, result);
+            answer = talloc_steal (context, arch_patch_id_new (result));
+            lim_free (0, result);
+            return answer;
+          }
+      }
+
+    Throw (exception (ENOENT, "Could not retrieve ancestor"));
 }
 
 /* we've observed the ancestor patch_id:
@@ -410,19 +326,27 @@ arch_patch_ancestor (t_uchar const * patch_id, int soft_errors, int cached_only)
  * $other_patch_id for a specific ancestor
  */
 void
-arch_ancestry_note_ancestor (t_uchar const *patch_id, t_uchar const *ancestor)
+arch_ancestry_note_ancestor (arch_patch_id * patch_id, t_uchar const *ancestor)
 {
-    t_uchar *current = arch_cached_patch_ancestor (patch_id);
-    if (current && !str_cmp(current, ARCH_ANCESTRY_PATCH_PARTIAL))
+    struct exception * e;
+    invariant (ancestor != NULL);
+    invariant (str_cmp (arch_patch_id_patch_id (patch_id), ARCH_ANCESTRY_PATCH_PARTIAL));
+    
+    Try
+      arch_cached_patch_ancestor (talloc_context, patch_id);
+    Catch (e)
       {
-        invariant (ancestor != NULL);
-        invariant (str_cmp (patch_id, ARCH_ANCESTRY_PATCH_PARTIAL));
-        t_uchar * ancestor_query = arch_patch_id_query (patch_id, "ancestor");
+        t_uchar * ancestor_query;
+        /* dont mask critical errors */
+        if (e->code < 0)
+            Throw (e);
+        talloc_free (e);
+
+        ancestor_query = arch_cache_query_new (patch_id, "ancestor");
         arch_cache_put_line (ancestor_query, ancestor);
-        assoc_set (&arch_ancestors, (t_uchar *)patch_id, (t_uchar *)ancestor);
+        assoc_set (&arch_ancestors, arch_patch_id_patch_id (patch_id), (t_uchar *)ancestor);
         lim_free (0, ancestor_query);
       }
-    lim_free (0, current);
 }
 
 static void
@@ -432,9 +356,17 @@ arch_ancestry_insert_ancestry (rel_table revisions)
     if (!revisions)
         return;
     for (row=0; row < rel_n_records (revisions) - 1; ++row)
-        arch_ancestry_note_ancestor (revisions[row][0], revisions[row + 1][0]);
+      {
+        arch_patch_id * temp = arch_patch_id_new (revisions[row][0]);
+        arch_ancestry_note_ancestor (temp, revisions[row + 1][0]);
+        talloc_free (temp);
+      }
     if (str_cmp(ARCH_ANCESTRY_PATCH_PARTIAL, revisions[rel_n_records (revisions) - 1][0]))
-        arch_ancestry_note_ancestor (revisions[rel_n_records (revisions) - 1][0], "");
+      {
+        arch_patch_id * temp = arch_patch_id_new (revisions[rel_n_records (revisions) - 1][0]);
+        arch_ancestry_note_ancestor (temp, "");
+        talloc_free (temp);
+      }
 }
 
 /* append locally available ancestry data 
@@ -445,6 +377,7 @@ ancestry_append_available (rel_table current_ancestry)
 {
     rel_table answer = NULL;
     t_uchar * fq_current;
+    arch_patch_id * current;
     int position = 0;
     if (!rel_n_records(current_ancestry))
         return NULL;
@@ -459,16 +392,29 @@ ancestry_append_available (rel_table current_ancestry)
             fq_current = NULL;
       }
     
-    fq_current = str_save (0, answer[rel_n_records(answer) - 1][0]); 
-    while (fq_current && str_cmp (fq_current, ARCH_ANCESTRY_PATCH_PARTIAL))
       {
-        t_uchar *next = arch_patch_ancestor (fq_current, 1,1);
-        if (next)
-            rel_add_records (&answer, rel_make_record (next, 0), 0);
-        lim_free (0, fq_current);
-        fq_current = next;
+        struct exception * e;
+        Try 
+          {
+            current = talloc_steal (talloc_context, arch_patch_id_new (answer[rel_n_records(answer) - 1][0])); 
+            while (current && str_cmp (arch_patch_id_patch_id (current), ARCH_ANCESTRY_PATCH_PARTIAL))
+              {
+                arch_patch_id *next = arch_patch_ancestor (talloc_context, current, 1);
+                if (next)
+                    rel_add_records (&answer, rel_make_record (arch_patch_id_patch_id (next), 0), 0);
+                talloc_free (current);
+                current = next;
+              }
+            talloc_free (current);
+          }
+        Catch (e)
+          {
+            /* dont mask critical errors */
+            if (e->code < 0)
+                Throw (e);
+            talloc_free (e);
+          }
       }
-    lim_free (0, fq_current);
     return answer;
 
 }
@@ -478,9 +424,9 @@ ancestry_append_available (rel_table current_ancestry)
  * maybe_ancestor.
  */
 static rel_table
-ancestry_scan (t_uchar const * patch_id, rel_table maybe_ancestor, int minimum_depth, int soft_errors)
+ancestry_scan (t_uchar const * patch_id, rel_table maybe_ancestor, int minimum_depth)
 {
-  t_uchar * fq_current = str_save (0, patch_id);
+  arch_patch_id * current = arch_patch_id_new (patch_id);
   rel_table answer = NULL;
   rel_table final = NULL;
   int depth = 0;
@@ -493,43 +439,41 @@ ancestry_scan (t_uchar const * patch_id, rel_table maybe_ancestor, int minimum_d
     {
       /* pre-condition:
        *
-       * fq_current tell use the patch_id to append to the result.
+       * current tell use the patch_id to append to the result.
        */
       
       int partial = 0;
-      t_uchar * fq_next_revision = 0;
-      
-      fq_next_revision = arch_patch_ancestor (fq_current, 1, 0);
-      if (!fq_next_revision)
-          /* end of ancestry */
+      arch_patch_id * next_revision = NULL;
+      struct exception * e;
+      Try 
         {
-          lim_free (0, fq_next_revision);
-          fq_next_revision = NULL;
+          next_revision = arch_patch_ancestor (NULL, current, 0);
         }
-      if (fq_next_revision && !str_cmp (ARCH_ANCESTRY_PATCH_PARTIAL, fq_next_revision))
-          /* couldn't search further */
+      Catch (e)
         {
-          lim_free (0, fq_next_revision);
-          rel_add_records(&answer, rel_make_record (ARCH_ANCESTRY_PATCH_PARTIAL, 0), 0);
-          fq_next_revision = NULL; 
+          /* dont mask critical errors */
+          if (e->code < 0)
+              Throw (e);
+          talloc_free (e);
           partial = -1;
+          rel_add_records(&answer, rel_make_record (ARCH_ANCESTRY_PATCH_PARTIAL, 0), 0);
         }
-      lim_free (0, fq_current);
-      if (fq_next_revision)
+      
+      talloc_free (current);
+      if (partial || !next_revision)
+          current = NULL;
+      else
           /* normal ancestor */
         {
-          rel_add_records (&answer, rel_make_record (fq_next_revision, 0), 0);
-          fq_current = str_save (0, fq_next_revision);
+          rel_add_records (&answer, rel_make_record (arch_patch_id_patch_id (next_revision), 0), 0);
+          current = next_revision;
         }
-      else
-          fq_current = NULL;
-      lim_free (0, fq_next_revision);
       ++depth;
       
-      if (!fq_current || partial)
+      if (!current || partial)
           break;
     }
-  lim_free (0, fq_current);
+  talloc_free (current);
   final = ancestry_append_available (answer);
   rel_free_table (answer);
   arch_archive_connection_cache_finalise (&global_cache, NULL);
@@ -573,7 +517,7 @@ ancestry_write (rel_table answer, int outfd, int maximum_depth)
 
 /* retrieve the patch ancestry for for patch_id from tree tree */
 static rel_table
-patch_ancestry_from_tree (struct arch_project_tree *tree, t_uchar * patch_id, int minimum_depth, int soft_errors)
+patch_ancestry_from_tree (struct arch_project_tree *tree, arch_patch_id * patch_id, int minimum_depth)
 {
     rel_table answer = NULL;
     if (arch_project_tree_file_exists(tree, ARCH_PROJECT_TREE_ANCESTRY_FILE))
@@ -630,8 +574,18 @@ patch_ancestry_from_tree (struct arch_project_tree *tree, t_uchar * patch_id, in
  *    from there.
  */
 
+/**
+ * \brief get the revision-history for a specific patch.
+ * \throw E* if no ancestry is available. 
+ * network access to revisions is presumed flakey and recoverable 
+ * errors are caught. unknown errors are propogated up.
+ * \param optional_tree a tree to gather ancestry from
+ * \param patch_id the patch to gather ancestry for
+ * \param minimum depth the minimum amount of history to attempt to 
+ * retrieve.
+ */
 rel_table
-patch_ancestry (struct arch_project_tree *optional_tree, t_uchar const * patch_id, int minimum_depth, int soft_errors)
+patch_ancestry (void * context, struct arch_project_tree *optional_tree, arch_patch_id * patch_id, int minimum_depth)
 {
     rel_table answer = NULL;
     /* look in the local tree */
@@ -645,73 +599,36 @@ patch_ancestry (struct arch_project_tree *optional_tree, t_uchar const * patch_i
         answer = partial_ancestry_from_full (candidate, patch_id);
         if (!answer)
           {
-            answer = ancestry_scan (patch_id, candidate, minimum_depth, soft_errors);
+            answer = ancestry_scan (arch_patch_id_patch_id (patch_id), candidate, minimum_depth);
             do_snap = -1;
           }
 
         rel_free_table (candidate);
         
         if (!answer)
-            
-          {
-            if (soft_errors)
-                return NULL;
-            else
-              {
-                safe_printfmt (2, "Empty or invalid cached ancestry file\n");
-                exit (2);
-              }
-          }
-        if (do_snap)
-            ancestry_snap (optional_tree, answer);
-        return answer;
+            Throw (exception (ENOENT, "Empty or invalid cached ancestry file"));
       }
+    else
     /* look in a reference tree */
       {
-        arch_patch_id patch;
-        t_uchar *pristine;
-        struct arch_project_tree pristine_tree;
+        arch_project_tree_t * pristine_tree;
 
-        arch_patch_id_init (&patch, patch_id);
-        pristine = arch_library_find (NULL, &patch, 1);
-        arch_patch_id_finalise (&patch);
+        pristine_tree = arch_library_find (NULL, patch_id, 1);
         
-        if (pristine)
+        if (pristine_tree)
           {
-            arch_project_tree_init (&pristine_tree, pristine);
-            answer = patch_ancestry_from_tree (&pristine_tree, (t_uchar *)patch_id, minimum_depth, soft_errors);
-            arch_project_tree_finalise(&pristine_tree);
+            answer = patch_ancestry_from_tree (pristine_tree, patch_id, minimum_depth);
+            arch_project_tree_delete(pristine_tree);
           }
 
-        lim_free (0, pristine);
       }
-    if (answer)
-        goto finished;
-    /* look for a cached ancestry file in the revision */
-    /* answer = ... */
-
-    /* Check we got enough */
-    if (minimum_depth > 0 && answer && 
-        !str_cmp (answer[rel_n_records(answer) - 1][0], ARCH_ANCESTRY_PATCH_PARTIAL) &&
-        rel_n_records(answer) < minimum_depth)
-      {
-        /* TODO finish the min depth stuff
-         * recursion isn't the best, may make sense to turn these few calls
-         * into a loop...
-         *
-        int minimum = minimum_depth - rel_n_records(answer) ;
-        rel_table temp = patch_ancestry (NULL, answer[rel_n_records(answer) - 2][0], remaining, soft_errors);
-        remove the PARTIAL record and append the result we got to answer.
-        goto finished;
-        */
-      }
-    
     /* scan the network */
-    answer = ancestry_scan (patch_id, NULL, minimum_depth, soft_errors);
-    /* TODO: assert answer ? */
-finished:
+    answer = ancestry_scan (arch_patch_id_patch_id (patch_id), NULL, minimum_depth);
+    if (!answer)
+        Throw (exception (ENOENT, "Empty or invalid cached ancestry file"));
     if (optional_tree && answer)
         ancestry_snap (optional_tree, answer);
+    talloc_steal (context, ar_base (answer));
     return answer;
 }
 
@@ -719,14 +636,14 @@ finished:
  * -1 depth means all known,
  *  0 means just the revisions data*/
 void
-ancestry_upload_patch (struct arch_archive * arch, t_uchar const *fqrevision, int max_depth)
+ancestry_upload_patch (struct arch_archive * arch, arch_patch_id * patch_id, int max_depth)
 {
     int infd = -1;
     /* working file */
-    t_uchar *tmp_file = tmp_file_name (".", ",,ancestry");
-    t_uchar *tmp_gz_file = tmp_file_name (".", ",,ancestry");
+    t_uchar *tmp_file = talloc_tmp_file_name (talloc_context, ".", ",,ancestry");
+    t_uchar *tmp_gz_file = talloc_tmp_file_name (tmp_file, ".", ",,ancestry");
     /* generate ancestry data */
-    rel_table ancestry = patch_ancestry (NULL, fqrevision, max_depth, 0);
+    rel_table ancestry = patch_ancestry (tmp_file, NULL, patch_id, max_depth);
     invariant (rel_n_records (ancestry) > 0);
     /* create a ancestry file */
       {
@@ -765,59 +682,45 @@ ancestry_upload_patch (struct arch_archive * arch, t_uchar const *fqrevision, in
     /* upload */
       {
         t_uchar *error = NULL;
-        t_uchar *revision = arch_parse_package_name (arch_ret_non_archive, 0, fqrevision);
-        arch_archive_put_ancestry (&error, arch, revision, infd);
-        lim_free (0, revision);
+        arch_archive_put_ancestry (&error, arch, arch_patch_id_revision (patch_id), infd);
       }
     
     safe_close (infd);
     safe_unlink (tmp_gz_file);
-    lim_free (0, tmp_file);
-    lim_free (0, tmp_gz_file);
+    talloc_free (tmp_file);
 }
 
-/* insert/update an ancestry file for a branch
- * -1 depth means all known,
- *  0 means just the revisions data
+/**
+ * \brief get the ancestry data cache from a revision in the archive
+ * \throw system errors only
+ * \param patch_id the patch to get ancestors for
+ * \param soft errors exit / return PARTIAL on error
  */
-
-void
-ancestry_upload_branch (struct arch_archive * arch, t_uchar const *fqrevision, int max_depth)
-{
-    /* generate ancestry data */
-    rel_table ancestry = patch_ancestry (NULL, fqrevision, max_depth, 0);
-    invariant (rel_n_records (ancestry) > 0);
-}
-
-t_uchar *
-arch_ancestry_from_revision (t_uchar const * patch_id, int soft_errors)
+arch_patch_id *
+arch_ancestry_from_revision (void * context, arch_patch_id * patch_id, int soft_errors)
 {
     /* retrieve tarball */
-    struct arch_archive *arch = cached_archive_connection (patch_id, soft_errors);
-    t_uchar *tmp_gz_file = tmp_file_name (".", ",,ancestry");
-    t_uchar *tmp_file = tmp_file_name (".", ",,ancestry");
-    
+    struct arch_archive *arch = arch_archive_connection_cache_find_or_maybe_connect (&global_cache, arch_patch_id_archive (patch_id), soft_errors);
+    t_uchar *tmp_gz_file = talloc_tmp_file_name (talloc_context, ".", ",,ancestry");
+    t_uchar *tmp_file = talloc_tmp_file_name (tmp_gz_file, ".", ",,ancestry");
     
     if (arch)
       {
         int ignore;
         int outfd;
-        t_uchar * revision = arch_parse_package_name(arch_ret_non_archive, 0, patch_id);
         int has_ancestry;
         
-        arch_revision_type (NULL, NULL, &has_ancestry, arch, revision);
+        arch_revision_type (NULL, NULL, &has_ancestry, arch, patch_id);
         if (!has_ancestry)
           {
-            lim_free (0, revision);
             goto finished;
           }
 
         vu_unlink (&ignore, tmp_gz_file);
         outfd = safe_open (tmp_gz_file, (O_WRONLY | O_CREAT | O_EXCL), 0444);
 
-        arch_archive_get_ancestry (outfd, arch, revision);
+        arch_archive_get_ancestry (outfd, arch, arch_patch_id_revision (patch_id));
         safe_close (outfd);
-        lim_free (0, revision);
       }
     else
       goto finished;
@@ -859,10 +762,9 @@ arch_ancestry_from_revision (t_uchar const * patch_id, int soft_errors)
     safe_unlink (tmp_file);
 
 finished:
-    lim_free (0, tmp_file);
-    lim_free (0, tmp_gz_file);
+    talloc_free (tmp_gz_file);
 
-    return arch_cached_patch_ancestor (patch_id);
+    return arch_cached_patch_ancestor (context, patch_id);
 }
 
 /* tag: Tom Lord Thu Jun 26 18:41:53 2003 (ancestry.c)
