@@ -9,27 +9,29 @@
  */
 
 
+#include "hackerlab/bugs/exception.h"
 #include "hackerlab/fs/file-names.h"
+#include "hackerlab/mem/talloc.h"
 #include "hackerlab/vu/safe.h"
 #include "libfsutils/tmp-files.h"
 #include "libfsutils/rmrf.h"
-#include "libarch/build-revision.h"
+#include "libarch/local-cache.h"
 #include "libarch/archive-cache.h"
 
 
 void
-arch_archive_cache (int chatter_fd, struct arch_archive * arch, t_uchar * archive, t_uchar * revision, t_uchar * cache_dir)
+arch_archive_cache (int chatter_fd, struct arch_archive * arch, t_uchar * archive, t_uchar * revision, arch_project_tree_t * cache)
 {
   t_uchar * tmp_dir = 0;
   t_uchar * pristine_dir = 0;
 
-  tmp_dir = tmp_file_name (".", ",,archive-cache-revision");
+  tmp_dir = talloc_tmp_file_name (talloc_context, ".", ",,archive-cache-revision");
   pristine_dir = file_name_in_vicinity (0, tmp_dir, revision);
 
   safe_mkdir (tmp_dir, 0777);
   safe_mkdir (pristine_dir, 0777);
 
-  arch_build_revision (chatter_fd, pristine_dir, arch, archive, revision, cache_dir);
+  pristine_dir = arch_find_or_make_tmp_local_copy  (chatter_fd, tmp_dir, cache, NULL, arch, archive, revision);
   
   {
     t_uchar * errstr;
@@ -43,7 +45,7 @@ arch_archive_cache (int chatter_fd, struct arch_archive * arch, t_uchar * archiv
 
   rmrf_file (tmp_dir);
 
-  lim_free (0, tmp_dir);
+  talloc_free (tmp_dir);
   lim_free (0, pristine_dir);
 }
 
