@@ -10,10 +10,12 @@
 
 
 #include "po/gettext.h"
+#include "hackerlab/bugs/exception.h"
 #include "hackerlab/bugs/panic.h"
 #include "hackerlab/os/errno.h"
 #include "hackerlab/os/errno-to-string.h"
 #include "hackerlab/mem/mem.h"
+#include "hackerlab/mem/talloc.h"
 #include "hackerlab/char/str.h"
 #include "hackerlab/hash/md5.h"
 #include "hackerlab/hash/sha1.h"
@@ -38,7 +40,7 @@
 
 
 /* __STDC__ prototypes for static functions */
-static void pfs_close (struct arch_archive * a);
+static int pfs_close (void * arch_archive_void);
 static t_uchar * pfs_archive_version (struct arch_archive * a);
 static rel_table pfs_categories (struct arch_archive * a);
 static rel_table pfs_branches (struct arch_archive * a, t_uchar * category);
@@ -47,12 +49,13 @@ static rel_table pfs_revisions (struct arch_archive * a, t_uchar * version);
 static t_uchar * pfs_archive_log (struct arch_archive * a, t_uchar * revision);
 static int pfs_revision_type  (enum arch_revision_type * type, int * archive_cached,
                                int *has_ancestry,
-                                struct arch_archive * a, t_uchar * revision);
+                                struct arch_archive * a, t_uchar const * revision);
 static int list_contains (rel_table files, t_uchar * name);
 static void pfs_get_patch (int out_fd, struct arch_archive * a, t_uchar * revision);
 static void pfs_get_cached (int out_fd, struct arch_archive * a, t_uchar * revision);
 static void pfs_get_import (int out_fd, struct arch_archive * a, t_uchar * revision);
 static t_uchar * pfs_get_continuation (struct arch_archive * a, t_uchar * revision);
+static int pfs_set_meta_info (struct arch_archive * a, t_uchar * meta_info_name, t_uchar * meta_info_value);
 static t_uchar * pfs_get_meta_info (struct arch_archive * a, t_uchar * meta_info_name);
 static int pfs_make_category (t_uchar ** errstr, struct arch_archive * a, t_uchar * category);
 static int pfs_make_branch (t_uchar ** errstr, struct arch_archive * a, t_uchar * branch);
@@ -64,6 +67,7 @@ static int pfs_lock_revision (t_uchar ** errstr, struct arch_archive * a,
                               t_uchar * txn_id,
                               t_uchar * new_level);
 static int pfs_revision_ready (t_uchar ** errstr, struct arch_archive * a,
+                               struct arch_archive * from_archive, 
                                t_uchar * version,
                                t_uchar * prev_level,
                                t_uchar * uid,
@@ -113,27 +117,30 @@ static int pfs_put_import (t_uchar ** errstr, struct arch_archive * a,
                            t_uchar * txn_id,
                            t_uchar * level,
                            int in_fd);
-static int pfs_put_cached (t_uchar ** errstr, struct arch_archive * a, t_uchar * revision, int in_fd);
+static int pfs_put_cached (t_uchar ** errstr, struct arch_archive * a, struct arch_archive *from_archive, t_uchar * revision, int in_fd);
 static int pfs_delete_cached (t_uchar ** errstr, struct arch_archive * a, t_uchar * revision);
 static void pfs_repair_non_txnal (int chatter_fd, struct arch_archive * a);
 static void pfs_get_ancestry (int out_fd, struct arch_archive * a, t_uchar * revision);
-static int pfs_put_ancestry (t_uchar **errstr, struct arch_archive *a, t_uchar * revision, int in_fd);
+static int pfs_put_ancestry (t_uchar **errstr, struct arch_archive *a, struct arch_archive *from_archive, t_uchar * revision, int in_fd);
 static int pfs_delete_ancestry (t_uchar ** errstr, struct arch_archive * a, t_uchar * revision);
 static void sign_and_upload (struct arch_archive * a, t_uchar * tmp_file, t_uchar * path, int in_fd, t_uchar * description, int batteries_to_power);
 static void pfs_create_signature_file (struct arch_archive * a,
                                        t_uchar * revision);
 static int pfs_finish_signature_file (struct arch_archive *a, 
+                                      struct arch_archive *from_archive,
                                        t_uchar * version,
                                        t_uchar * prev_level,
                                        t_uchar * uid,
                                        t_uchar * txn_id,
                                        t_uchar * new_level);
-static int pfs_finish_signature_file_worker (struct arch_archive * a, t_uchar * revision, t_uchar * dir, t_uchar * checksumname);
+static int pfs_finish_signature_file_worker (struct arch_archive * a, struct arch_archive *from_archive, t_uchar * revision, t_uchar * dir, t_uchar * checksumname);
 static void insert_sums (struct arch_pfs_archive *arch,
                          t_uchar * description, 
                          t_uchar * data_file);
 static void dir_changed (struct arch_pfs_archive * arch, t_uchar const *dir);
 static int pfs_delete_file_pair (t_uchar ** errstr, struct arch_archive * a, t_uchar * file1, t_uchar *file2);
+static int pfs_set_mirror (t_uchar ** errstr, struct arch_archive * archive, int enabled);
+static void pfs_clear_cached (struct arch_pfs_archive * arch);
 
 
 
@@ -141,8 +148,6 @@ static struct arch_archive_vtable pfs_vtable =
 {
   "pfs",
 
-  pfs_close,
-
   pfs_archive_version,
 
   pfs_categories,
@@ -156,6 +161,7 @@ static struct arch_archive_vtable pfs_vtable =
   pfs_get_cached,
   pfs_get_import,
   pfs_get_continuation,
+  pfs_set_meta_info,
   pfs_get_meta_info,
 
   pfs_make_category,
@@ -180,7 +186,9 @@ static struct arch_archive_vtable pfs_vtable =
 
   pfs_get_ancestry,
   pfs_put_ancestry,
-  pfs_delete_ancestry
+  pfs_delete_ancestry,
+
+  pfs_set_mirror
 };
 
 
@@ -196,10 +204,14 @@ int
 arch_pfs_archive_connect (struct arch_archive ** a, int soft_errors)
 {
   struct arch_pfs_archive * arch = (struct arch_pfs_archive *)*a;
-
-  arch = lim_realloc (0, (t_uchar *)arch, sizeof (struct arch_pfs_archive));
+  arch = talloc_realloc_size (0, arch, sizeof (struct arch_pfs_archive));
+  talloc_set_destructor (arch, pfs_close);
   *a = (struct arch_archive *)arch;
   arch->arch.vtable = &pfs_vtable;
+  arch->from_arch = 0;
+  arch->txn_signature_file = NULL;
+  arch->txn_signature_fd = -1;
+  arch->root_dirs = NULL;
   arch->pfs = arch_pfs_connect (arch->arch.location, soft_errors);
   if (!arch->pfs)
     {
@@ -209,26 +221,33 @@ arch_pfs_archive_connect (struct arch_archive ** a, int soft_errors)
         }
       panic ("failed to connect to archive with soft errors zero");
     }
-  arch->from_arch = 0;
-  arch->txn_signature_file = NULL;
-  arch->txn_signature_fd = -1;
   return 0;
 }
 
-
-static void
-pfs_close (struct arch_archive * a)
+/**
+ * \brief see arch_archive_close
+ */
+int
+pfs_close (void * arch_archive_void)
 {
+  struct arch_archive *a = (struct arch_archive *) arch_archive_void;
+  struct arch_pfs_archive * arch = (struct arch_pfs_archive *)a;
+  if (arch->pfs)
+      arch_pfs_disconnect (&arch->pfs);
+  pfs_clear_cached (arch);
+  return arch_archive_finalise (a);
 }
 
-
+/**
+ * \brief see arch_archive.archive_version
+ */
 static t_uchar *
 pfs_archive_version (struct arch_archive * a)
 {
   struct arch_pfs_archive * arch = (struct arch_pfs_archive *)a;
   t_uchar * path = 0;
   t_uchar * meta_info_path = 0;
-  t_uchar * contents = 0;
+  t_uchar * contents = NULL;
   t_uchar * nl;
 
   path = arch_fs_archive_archive_version_path (0);
@@ -243,8 +262,7 @@ pfs_archive_version (struct arch_archive * a)
     }
   else
     {
-      safe_printfmt (2, "pfs_archive_version: unidentifiable archive (%s)\n", a->name);
-      exit (2);
+      goto done;
     }
 
   nl = str_chr_index (contents, '\n');
@@ -257,7 +275,7 @@ pfs_archive_version (struct arch_archive * a)
       contents = lim_realloc (0, contents, len);
       contents[len - 1] = 0;
     }
-
+done:
   lim_free (0, path);
   lim_free (0, meta_info_path);
   return contents;
@@ -270,20 +288,22 @@ pfs_baz_search (struct arch_archive *a,
 		enum arch_parse_package_name_type ret_type)
 {
   struct arch_pfs_archive * arch = (struct arch_pfs_archive *)a;
-  rel_table files = 0;
   rel_table answer = 0;
   int x;
 
-  files = arch_pfs_directory_files (arch->pfs, ".", 0);
-
-  for (x = 0; x < rel_n_records (files); ++x)
+  if (!arch->root_dirs)
     {
-      if (arch_valid_package_name (files[x][0], arch_no_archive, req_type, 1) && 
-          !str_cmp_prefix(prefix, files[x][0]))
-        rel_add_records (&answer, rel_make_record (arch_parse_package_name(ret_type, NULL, files[x][0]), 0), 0);
+      arch->root_dirs = arch_pfs_directory_files (arch->pfs, ".", 0);
+      rel_sort_table_by_field (0, arch->root_dirs, 0);
+    }
+
+  for (x = 0; x < rel_n_records (arch->root_dirs); ++x)
+    {
+      if (arch_valid_package_name (arch->root_dirs[x][0], arch_no_archive, req_type, 1) && 
+          !str_cmp_prefix(prefix, arch->root_dirs[x][0]))
+        rel_add_records (&answer, rel_make_record (arch_parse_package_name(ret_type, NULL, arch->root_dirs[x][0]), 0), 0);
     }
 
-  rel_free_table (files);
   rel_uniq_by_field (&answer, 0);
   return answer;
 }
@@ -326,6 +346,7 @@ pfs_branches (struct arch_archive * a, t_uchar * category)
 
   rel_free_table (files);
   lim_free (0, search_prefix);
+  lim_free (0, category_path);
   return answer;
 }
 
@@ -348,6 +369,7 @@ pfs_versions (struct arch_archive * a, t_uchar * package)
   answer = arch_pick_versions_by_field (files, 0);
 
   rel_free_table (files);
+  lim_free (0, branch_path);
   lim_free (0, search_prefix);
   return answer;
 }
@@ -368,6 +390,7 @@ pfs_revisions (struct arch_archive * a, t_uchar * version)
   answer = arch_pick_patch_levels_by_field (files, 0);
 
   rel_free_table (files);
+  lim_free (0, version_path);
   return answer;
 }
 
@@ -381,7 +404,7 @@ pfs_archive_log (struct arch_archive * a, t_uchar * revision)
 
   if (arch_pfs_ensure_checksum_data (arch, revision))
     {
-      safe_printfmt (2, "trouble reading checksum file for %s/%s\n", a->name, revision);
+      safe_printfmt (2, "trouble reading checksum file for %s/%s url: %s\n", a->official_name, revision, a->location);
       exit (2);
     }
 
@@ -403,7 +426,7 @@ pfs_file_path_to_tail (t_uchar *file_path)
 
 static int
 pfs_revision_type (enum arch_revision_type * type, int * archive_cached, int *has_ancestry,
-                   struct arch_archive * a, t_uchar * revision)
+                   struct arch_archive * a, t_uchar const * revision)
 {
   struct arch_pfs_archive * arch = (struct arch_pfs_archive *)a;
   t_uchar * continuation_path = 0;
@@ -446,7 +469,7 @@ pfs_revision_type (enum arch_revision_type * type, int * archive_cached, int *ha
 
   if (arch_pfs_ensure_checksum_data (arch, revision))
     {
-      safe_printfmt (2, "trouble reading checksum file for %s/%s\n", a->name, revision);
+      safe_printfmt (2, "trouble reading checksum file for %s/%s url: %s\n", a->official_name, revision, a->location);
       exit (2);
     }
 
@@ -457,13 +480,13 @@ pfs_revision_type (enum arch_revision_type * type, int * archive_cached, int *ha
   has_log_file = list_contains (files, log_tail);
   has_ancestry_file = list_contains (files, ancestry_tail);
 
-  if (arch_pfs_checksum_governs (a->name, revision)
-      && ((!!has_continuation_file != !!arch_pfs_checksum_anticipates_file (a->name, revision, continuation_tail))
-          || (!!has_changeset_file != !!arch_pfs_checksum_anticipates_file (a->name, revision, changeset_tail))
-          || (!!has_import_file != !!arch_pfs_checksum_anticipates_file (a->name, revision, import_tail))
-          || (!!has_cacherev_file != !!arch_pfs_checksum_anticipates_file (a->name, revision, cacherev_tail))
-          || (!!has_log_file != !!arch_pfs_checksum_anticipates_file (a->name, revision, log_tail))
-          || (!!has_ancestry_file != !!arch_pfs_checksum_anticipates_file (a->name, revision, ancestry_tail))
+  if (arch_pfs_checksum_governs (arch, revision)
+      && ((!!has_continuation_file != !!arch_pfs_checksum_anticipates_file (a, revision, continuation_tail))
+          || (!!has_changeset_file != !!arch_pfs_checksum_anticipates_file (a, revision, changeset_tail))
+          || (!!has_import_file != !!arch_pfs_checksum_anticipates_file (a, revision, import_tail))
+          || (!!has_cacherev_file != !!arch_pfs_checksum_anticipates_file (a, revision, cacherev_tail))
+          || (!!has_log_file != !!arch_pfs_checksum_anticipates_file (a, revision, log_tail))
+          || (!!has_ancestry_file != !!arch_pfs_checksum_anticipates_file (a, revision, ancestry_tail))
           ))
     {
       safe_printfmt (2, ("\n"
@@ -476,10 +499,11 @@ pfs_revision_type (enum arch_revision_type * type, int * archive_cached, int *ha
                          "\n"
                          "  archive: %s\n"
                          "  revision: %s\n"
+                         "  url: %s\n"
                          "\n"
                          "***********************************\n"
                          "\n"),
-                     a->name, revision);
+                     a->official_name, revision, a->location);
 
       if (arch_pfs_checksum_governs_strictly (arch))
         exit (2);
@@ -554,7 +578,7 @@ pfs_get_patch (int out_fd, struct arch_archive * a, t_uchar * revision)
 
   if (arch_pfs_ensure_checksum_data (arch, revision))
     {
-      safe_printfmt (2, "trouble reading checksum file for %s/%s\n", a->name, revision);
+      safe_printfmt (2, "trouble reading checksum file for %s/%s url: %s\n", a->official_name, revision, a->location);
       exit (2);
     }
 
@@ -572,7 +596,7 @@ pfs_get_cached (int out_fd, struct arch_archive * a, t_uchar * revision)
 
   if (arch_pfs_ensure_checksum_data (arch, revision))
     {
-      safe_printfmt (2, "trouble reading checksum file for %s/%s\n", a->name, revision);
+      safe_printfmt (2, "trouble reading checksum file for %s/%s url: %s\n", a->official_name, revision, a->location);
       exit (2);
     }
 
@@ -592,7 +616,7 @@ pfs_get_import (int out_fd, struct arch_archive * a, t_uchar * revision)
 
   if (arch_pfs_ensure_checksum_data (arch, revision))
     {
-      safe_printfmt (2, "trouble reading checksum file for %s/%s\n", a->name, revision);
+      safe_printfmt (2, "trouble reading checksum file for %s/%s url: %s\n", a->official_name, revision, a->location);
       exit (2);
     }
 
@@ -612,7 +636,7 @@ pfs_get_continuation (struct arch_archive * a, t_uchar * revision)
 
   if (arch_pfs_ensure_checksum_data (arch, revision))
     {
-      safe_printfmt (2, "trouble reading checksum file for %s/%s\n", a->name, revision);
+      safe_printfmt (2, "trouble reading checksum file for %s/%s url: %s\n", a->official_name, revision, a->location);
       exit (2);
     }
 
@@ -624,6 +648,28 @@ pfs_get_continuation (struct arch_archive * a, t_uchar * revision)
   return answer;
 }
 
+static int 
+pfs_set_meta_info (struct arch_archive * a, 
+                   t_uchar * meta_info_name, 
+                   t_uchar * meta_info_value)
+{
+  struct arch_pfs_archive * arch = (struct arch_pfs_archive *)a;
+  t_uchar * meta_info_path = arch_fs_archive_meta_info_item_path (0, meta_info_name);
+  int answer;
+  if (meta_info_value == NULL)
+    {
+      answer = arch_pfs_rm (arch->pfs, meta_info_path, 1);
+    }
+  else
+    {
+      answer = arch_pfs_set_file_contents (arch->pfs, meta_info_path, meta_info_value, 1);
+    }
+
+  lim_free (0, meta_info_path);
+
+  return answer;
+}
+
 static t_uchar *
 pfs_get_meta_info (struct arch_archive * a, t_uchar * meta_info_name)
 {
@@ -780,6 +826,7 @@ pfs_make_version (t_uchar ** errstr, struct arch_archive * a, t_uchar * version)
 
   version_path = arch_fs_archive_version_path (a, 0, version);
 
+  /* TODO check via root_dirs for speed ? */
   if (arch_pfs_file_exists (arch->pfs, version_path))
     {
       if (errstr)
@@ -825,6 +872,7 @@ pfs_make_version (t_uchar ** errstr, struct arch_archive * a, t_uchar * version)
       lim_free (0, version_path_dir);
       lim_free (0, lock_dir_path);
       lim_free (0, contents_dir_path);
+      pfs_clear_cached(arch);
     }
 
   lim_free (0, version_path);
@@ -1063,13 +1111,14 @@ pfs_lock_revision (t_uchar ** errstr, struct arch_archive * a,
 
 static int
 pfs_revision_ready (t_uchar ** errstr, struct arch_archive * a,
+                    struct arch_archive *from_archive,
                     t_uchar * version,
                     t_uchar * prev_level,
                     t_uchar * uid,
                     t_uchar * txn_id,
                     t_uchar * new_level)
 {
-  if (pfs_finish_signature_file (a, version, prev_level, uid, txn_id, new_level))
+  if (pfs_finish_signature_file (a, from_archive, version, prev_level, uid, txn_id, new_level))
     {
       return -1;
     }
@@ -1133,33 +1182,37 @@ pfs_finish_revision (t_uchar ** errstr, struct arch_archive * a,
 
           if (arch->arch.http_blows)
             {
-              t_uchar * prev_revision = 0;
+              arch_patch_id * prev_revision = 0;
               t_uchar * new_rev_dir_path = 0;
 
               new_rev_dir_path = file_name_directory_file (0, new_rev_path);
               arch_pfs_update_listing_file (arch->pfs, new_rev_path);
               arch_pfs_update_listing_file (arch->pfs, new_rev_dir_path);
 
-              prev_revision = arch_previous_revision (a, revision);
+              {
+                arch_patch_id * revision_patch = arch_patch_id_new_archive (a->official_name, revision);
+              prev_revision = arch_previous_revision (a, revision_patch);
+                talloc_free (revision_patch);
+              }
               if (prev_revision)
                 {
                   t_uchar * prev_rev_path = 0;
 
-                  prev_rev_path = arch_fs_archive_revision_path (a, 0, prev_revision);
+                  prev_rev_path = arch_fs_archive_revision_path (a, 0, arch_patch_id_revision (prev_revision));
                   arch_pfs_update_listing_file (arch->pfs, prev_rev_path);
 
                   lim_free (0, prev_rev_path);
                 }
 
               lim_free (0, new_rev_dir_path);
-              lim_free (0, prev_revision);
+              talloc_free (prev_revision);
             }
         }
       else
         {
           {
-            safe_printfmt (2, "i/o error modifying archive (%s)\n    archive: %s\n    path: %s\n",
-                           *errstr, arch->arch.name, new_contents_path);
+            safe_printfmt (2, "i/o error modifying archive (%s)\n    archive: %s\n    path: %s\n,    url: %s\n",
+                           *errstr, arch->arch.official_name, new_contents_path, arch->arch.location);
             exit (2);
           }
         }
@@ -1354,7 +1407,6 @@ pfs_break_revision_lock (t_uchar ** errstr, struct arch_archive * a,
       t_uchar * version_path = 0;
       t_uchar * broken_lock_path = 0;
       t_uchar * broken_lock_dir = 0;
-      t_uchar * junk_path = 0;
 
       remade_basename = str_alloc_cat (0, ",,remade-lock--", (prev_level ? prev_level : (t_uchar *)"absolute-0"));
       new_lock_path = file_name_in_vicinity (0, locked_contents_path, remade_basename);
@@ -1363,7 +1415,6 @@ pfs_break_revision_lock (t_uchar ** errstr, struct arch_archive * a,
       version_path = arch_fs_archive_version_path (a, 0, version);
       broken_lock_path = arch_fs_archive_revision_lock_broken_path (a, 0, version, prev_level);
       broken_lock_dir = file_name_directory_file (0, broken_lock_path);
-      junk_path = tmp_file_name (version_path, ",,junk");
 
       arch_pfs_mkdir (arch->pfs, new_lock_path, 0777, 1);
       arch_pfs_mkdir (arch->pfs, new_lock_contents_path, 0777, 1);
@@ -1498,7 +1549,7 @@ safely_remove_spent_lock (struct arch_pfs_archive * arch, t_uchar * spent_lock)
   invariant (!str_cmp_prefix ("++revision-lock-held--", spent_lock_tail));
 
   nested = arch_pfs_directory_files (arch->pfs, spent_lock, 1);
-  tmp_path = tmp_file_name (spent_lock, ",,junk");
+  tmp_path = talloc_tmp_file_name (talloc_context, spent_lock, ",,junk");
 
   arch_pfs_rmrf_file (arch->pfs, tmp_path);
 
@@ -1524,7 +1575,7 @@ safely_remove_spent_lock (struct arch_pfs_archive * arch, t_uchar * spent_lock)
 
   lim_free (0, spent_lock_tail);
   rel_free_table (nested);
-  lim_free (0, tmp_path);
+  talloc_free (tmp_path);
 }
 
 
@@ -1552,7 +1603,7 @@ pfs_put_log (t_uchar ** errstr, struct arch_archive * a,
   locked_contents_path = arch_fs_archive_revision_lock_locked_contents_path (a, 0, version, prev_level, uid, txn_id);
   log_path = file_name_in_vicinity (0, locked_contents_path, "log");
 
-  tmp_path = tmp_file_name ("/tmp", ",,pfs-dav-put-log");
+  tmp_path = tmp_file_name_in_tmp (",,pfs-dav-put-log");
   fd = safe_open (tmp_path, O_RDWR | O_CREAT | O_EXCL, 0400);
 
   safe_printfmt (fd, "%s", log_text);
@@ -1593,7 +1644,7 @@ pfs_put_continuation (t_uchar ** errstr, struct arch_archive * a,
   locked_contents_path = arch_fs_archive_revision_lock_locked_contents_path (a, 0, version, prev_level, uid, txn_id);
   continuation_path = file_name_in_vicinity (0, locked_contents_path, "CONTINUATION");
 
-  tmp_path = tmp_file_name ("/tmp", ",,pfs-dav-put-log");
+  tmp_path = tmp_file_name_in_tmp (",,pfs-dav-put-log");
   fd = safe_open (tmp_path, O_RDWR | O_CREAT | O_EXCL, 0400);
 
   safe_printfmt (fd, "%s", continuation);
@@ -1679,7 +1730,7 @@ pfs_put_import (t_uchar ** errstr, struct arch_archive * a,
 
 
 static int
-pfs_put_cached (t_uchar ** errstr, struct arch_archive * a, t_uchar * revision, int in_fd)
+pfs_put_cached (t_uchar ** errstr, struct arch_archive * a, struct arch_archive *from_archive, t_uchar * revision, int in_fd)
 {
   struct arch_pfs_archive * arch = (struct arch_pfs_archive *)a;
   t_uchar * cached_path = 0;
@@ -1692,7 +1743,7 @@ pfs_put_cached (t_uchar ** errstr, struct arch_archive * a, t_uchar * revision,
 
   pfs_create_signature_file (a, revision);
   sign_and_upload (a, 0, cached_path, in_fd, cached_tail, 1);
-  if (pfs_finish_signature_file_worker (a, revision, cached_dir, "checksum.cacherev"))
+  if (pfs_finish_signature_file_worker (a, from_archive, revision, cached_dir, "checksum.cacherev"))
     {
       /* FIXME: remove cacherev file. This is no worse than the fire and forget code this 
        * block replaces.*/
@@ -1757,7 +1808,7 @@ pfs_get_ancestry (int out_fd, struct arch_archive * a, t_uchar * revision)
 
   if (arch_pfs_ensure_checksum_data (arch, revision))
     {
-      safe_printfmt (2, "trouble reading checksum file for %s/%s\n", a->name, revision);
+      safe_printfmt (2, "trouble reading checksum file for %s/%s url: %s\n", a->official_name, revision, a->location);
       exit (2);
     }
 
@@ -1768,7 +1819,7 @@ pfs_get_ancestry (int out_fd, struct arch_archive * a, t_uchar * revision)
 }
 
 int
-pfs_put_ancestry (t_uchar **errstr, struct arch_archive *a, t_uchar * revision, int in_fd)
+pfs_put_ancestry (t_uchar **errstr, struct arch_archive *a, struct arch_archive *from_archive, t_uchar * revision, int in_fd)
 {
   struct arch_pfs_archive * arch = (struct arch_pfs_archive *)a;
   t_uchar * ancestry_path = 0;
@@ -1781,7 +1832,7 @@ pfs_put_ancestry (t_uchar **errstr, struct arch_archive *a, t_uchar * revision,
 
   pfs_create_signature_file (a, revision);
   sign_and_upload (a, 0, ancestry_path, in_fd, ancestry_tail, 1);
-  if (pfs_finish_signature_file_worker (a, revision, ancestry_dir, "ancestry.gz.checksum"))
+  if (pfs_finish_signature_file_worker (a, from_archive, revision, ancestry_dir, "ancestry.gz.checksum"))
     {
       /* FIXME: remove cacherev file. This is no worse than the fire and forget code this 
        * block replaces.*/
@@ -1944,7 +1995,7 @@ sign_and_upload (struct arch_archive * a, t_uchar * tmp_file, t_uchar * path, in
       t_uchar * tmp_contents = 0;
       int my_fd = -1;
 
-      tmp_contents = tmp_file_name ("/tmp", ",,arch-sign-and-upload"); 
+      tmp_contents = tmp_file_name_in_tmp (",,arch-sign-and-upload"); 
 
       my_fd = safe_open (tmp_contents, O_RDWR | O_CREAT | O_EXCL, 0400);
       copy_fd (in_fd, my_fd);
@@ -1970,7 +2021,7 @@ pfs_create_signature_file (struct arch_archive * a,
 
   invariant(arch->txn_signature_file == 0);
     
-  arch->txn_signature_file = tmp_file_name ("/tmp", ",,archive-signature");
+  arch->txn_signature_file = tmp_file_name_in_tmp (",,archive-signature");
   arch->txn_signature_fd = safe_open (arch->txn_signature_file, O_RDWR | O_CREAT | O_EXCL, 0400);
   arch->deferred_sha1_sums = 0;
   safe_printfmt (arch->txn_signature_fd, "Signature-for: %s/%s\n", a->official_name, revision);
@@ -1992,6 +2043,7 @@ pfs_create_signature_file (struct arch_archive * a,
 
 static int
 pfs_finish_signature_file (struct arch_archive *a, 
+                           struct arch_archive *from_archive, 
                            t_uchar * version,
                            t_uchar * prev_level,
                            t_uchar * uid,
@@ -2009,7 +2061,7 @@ pfs_finish_signature_file (struct arch_archive *a,
   revision = str_alloc_cat_many (0, version, "--", new_level, str_end);
 
   dir = arch_fs_archive_revision_lock_locked_contents_path (a, 0, version, prev_level, uid, txn_id);
-  if (pfs_finish_signature_file_worker (a, revision, dir, "checksum"))
+  if (pfs_finish_signature_file_worker (a, from_archive, revision, dir, "checksum"))
     {
       /* abort the transaction */
       result = -1;
@@ -2020,9 +2072,19 @@ pfs_finish_signature_file (struct arch_archive *a,
   return result;
 }
 
-
-static int
-pfs_finish_signature_file_worker (struct arch_archive * a, t_uchar * revision, t_uchar * dir, t_uchar * checksumname)
+/**
+ * \brief finish a checksum file
+ *
+ * sign if needed, and upload to the archive
+ * \param a the archive the checksum is for
+ * \param from_archive the archive that it should be copied from 
+ * \param revision name the checksum will belong to
+ * \param dir the directory the checksum is going into
+ * \param checksumname the filename for the checksum
+ * \return 0 on success
+ */
+int
+pfs_finish_signature_file_worker (struct arch_archive * a, struct arch_archive *from_archive, t_uchar * revision, t_uchar * dir, t_uchar * checksumname)
 {
   struct arch_pfs_archive *arch = (struct arch_pfs_archive *)a;
   t_uchar * signed_file = 0;
@@ -2056,7 +2118,7 @@ pfs_finish_signature_file_worker (struct arch_archive * a, t_uchar * revision, t
       signed_fd = safe_open (signed_file, O_RDWR | O_CREAT | O_EXCL, 0);
       safe_unlink (signed_file);
 
-      error = arch_pfs_sign_for_archive (a->name, a->official_name, revision, checksumname, in_fd, signed_fd);
+      error = arch_pfs_sign_for_archive (a, revision, checksumname, in_fd, signed_fd, from_archive);
 
       if (error)
 	{
@@ -2074,7 +2136,13 @@ pfs_finish_signature_file_worker (struct arch_archive * a, t_uchar * revision, t
     }
   else 
     {
-      if (arch_pfs_has_signing_rule (a->name))
+      /* TODO: check for has registry file, and if present
+       * check warn_unsigned instead
+       * RBC 20050309
+       * ->name cannot be removed here as its how the 
+       *  legacy archive registration works.
+       */
+      if (arch_pfs_has_signing_rule (a->registered_name))
         {
           safe_printfmt (2, ("\n"
                              "\n"
@@ -2089,7 +2157,7 @@ pfs_finish_signature_file_worker (struct arch_archive * a, t_uchar * revision, t
                              "********************************\n"
                              "\n"
                              "\n"),
-                         a->name, a->name);
+                         a->registered_name, a->official_name);
           /* FIXME: check for resource leaks */
           return -1;
         }
@@ -2168,6 +2236,40 @@ dir_changed (struct arch_pfs_archive * arch, t_uchar const *dir)
   arch_pfs_update_listing_file (arch->pfs, (t_uchar *)dir);
 }
 
+int 
+pfs_set_mirror (t_uchar ** errstr, struct arch_archive * archive, int enabled)
+{
+  struct arch_pfs_archive *arch = (struct arch_pfs_archive *)archive;
+    if (enabled)
+      {
+        t_uchar * metainfo_path = arch_fs_archive_meta_info_path (0);
+        int result = arch_pfs_set_file_contents_in_dir (arch->pfs, metainfo_path, "mirror", archive->official_name, 1);
+        if (result)
+          {
+            *errstr = str_save (0, "Failed to upload mirror file");
+          }
+        lim_free (0, metainfo_path);
+        return result;
+      }
+    else
+      {
+        t_uchar * metainfo_item_path = arch_fs_archive_meta_info_item_path (0, "mirror");
+        int result = arch_pfs_rm (arch->pfs, metainfo_item_path, 1);
+        if (result)
+          {
+            *errstr = str_save (0, "Failed to delete mirror file");
+          }
+        lim_free (0, metainfo_item_path);
+        return result;
+      }
+}
+
+void
+pfs_clear_cached (struct arch_pfs_archive * arch)
+{
+    rel_free_table (arch->root_dirs);
+    arch->root_dirs = NULL;
+}
 
 
 /* tag: Tom Lord Tue May 20 13:35:38 2003 (archive-pfs.c)
