@@ -13,7 +13,7 @@
 #include "hackerlab/vu/safe.h"
 #include "hackerlab/char/str.h"
 #include "hackerlab/bugs/panic.h"
-
+#include "hackerlab/mem/talloc.h"
 #include "libfsutils/copy-file.h"
 #include "libarch/archives.h"
 #include "libarch/cached-archive.h"
@@ -23,7 +23,7 @@
 
 static void 
 copy_fd_close(int in_fd, int out_fd);
-static void cache_close (struct arch_archive * a);
+static int  cache_close (void * arch_archive_void);
 static t_uchar * cache_archive_version (struct arch_archive * a);
 static rel_table cache_categories (struct arch_archive * a);
 static rel_table cache_branches (struct arch_archive * a, t_uchar * category);
@@ -32,11 +32,12 @@ static rel_table cache_revisions (struct arch_archive * a, t_uchar * version);
 static t_uchar * cache_archive_log (struct arch_archive * a, t_uchar * revision);
 static int cache_revision_type  (enum arch_revision_type * type, int * archive_cached,
 				 int *has_ancestry,
-                                struct arch_archive * a, t_uchar * revision);
+                                struct arch_archive * a, t_uchar const * revision);
 static void cache_get_patch (int out_fd, struct arch_archive * a, t_uchar * revision);
 static void cache_get_cached (int out_fd, struct arch_archive * a, t_uchar * revision);
 static void cache_get_import (int out_fd, struct arch_archive * a, t_uchar * revision);
 static t_uchar * cache_get_continuation (struct arch_archive * a, t_uchar * revision);
+static int cache_set_meta_info (struct arch_archive * a, t_uchar * meta_info_name, t_uchar * meta_info_value);
 static t_uchar * cache_get_meta_info (struct arch_archive * a, t_uchar * meta_info_name);
 static int cache_make_category (t_uchar ** errstr, struct arch_archive * a, t_uchar * category);
 static int cache_make_branch (t_uchar ** errstr, struct arch_archive * a, t_uchar * branch);
@@ -48,6 +49,7 @@ static int cache_lock_revision (t_uchar ** errstr, struct arch_archive * a,
                               t_uchar * txn_id,
                               t_uchar * new_level);
 static int cache_revision_ready (t_uchar ** errstr, struct arch_archive * a,
+				 struct arch_archive *from_archive,
                                t_uchar * version,
                                t_uchar * prev_level,
                                t_uchar * uid,
@@ -95,20 +97,21 @@ static int cache_put_import (t_uchar ** errstr, struct arch_archive * a,
                            t_uchar * txn_id,
                            t_uchar * level,
                            int in_fd);
-static int cache_put_cached (t_uchar ** errstr, struct arch_archive * a, t_uchar * revision, int in_fd);
+static int cache_put_cached (t_uchar ** errstr, struct arch_archive * a, struct arch_archive * from_archive,
+			     t_uchar * revision, int in_fd);
 static int cache_delete_cached (t_uchar ** errstr, struct arch_archive * a, t_uchar * revision);
 static void cache_repair_non_txnal (int chatter_fd, struct arch_archive * a);
 static void cache_get_ancestry (int out_fd, struct arch_archive * a, t_uchar * revision);
-static int cache_put_ancestry (t_uchar ** errstr, struct arch_archive * a, t_uchar * revision, int in_fd);
+static int cache_put_ancestry (t_uchar ** errstr, struct arch_archive * a, struct arch_archive * from_archive,
+			       t_uchar * revision, int in_fd);
 static int cache_delete_ancestry (t_uchar ** errstr, struct arch_archive * a, t_uchar * revision);
+static int cache_set_mirror (t_uchar ** errstr, struct arch_archive * archive, int enabled);
 
 
 static struct arch_archive_vtable cache_vtable =
 {
   "cache",
 
-  cache_close,
-
   cache_archive_version,
 
   cache_categories,
@@ -122,6 +125,7 @@ static struct arch_archive_vtable cache_vtable =
   cache_get_cached,
   cache_get_import,
   cache_get_continuation,
+  cache_set_meta_info,
   cache_get_meta_info,
 
   cache_make_category,
@@ -146,7 +150,9 @@ static struct arch_archive_vtable cache_vtable =
 
   cache_get_ancestry,
   cache_put_ancestry,
-  cache_delete_ancestry
+  cache_delete_ancestry,
+  
+  cache_set_mirror
 };
 
 
@@ -155,7 +161,7 @@ static struct arch_archive_vtable cache_vtable =
  * Returns a newly-allocated string
  */
 extern t_uchar * 
-arch_revision_query (struct arch_archive * a, t_uchar * revision, 
+arch_revision_query (struct arch_archive * a, t_uchar const * revision, 
                      t_uchar const * extension)
 {
   arch_patch_id patch;
@@ -184,7 +190,7 @@ copy_fd_close(int in_fd, int out_fd)
  * Determines whether the provided location is a cache URL.
  */
 int
-arch_cached_archive_protocol (t_uchar * location)
+arch_cached_archive_protocol (t_uchar const * location)
 {
   if (str_cmp_prefix ("cached:", location) == 0)
     return 1;
@@ -199,9 +205,9 @@ arch_cached_archive_protocol (t_uchar * location)
  * \return a new string, containing the modified location
  */
 extern t_uchar * 
-arch_uncached_location (t_uchar *location)
+arch_uncached_location (t_uchar const *location)
 {
-  t_uchar * new_location = location;
+  t_uchar const * new_location = location;
   while (1)
     {
       if (str_cmp_prefix("cached:", new_location) == 0)
@@ -232,7 +238,7 @@ arch_cached_archive_connect (struct arch_archive ** answer, int soft_errors)
   invariant (arch_cached_archive_protocol ((*answer)->location));
   if (! arch_cache_active ())
     {
-      safe_printfmt (2, "No cache attempting to connect to cached archive\n archive: %s\nlocation: %s\n", (*answer)->name, (*answer)->location);
+      safe_printfmt (2, "No cache attempting to connect to cached archive\n archive: %s\nlocation: %s\n", (*answer)->official_name, (*answer)->location);
       if (soft_errors)
 	  return -1;
       else
@@ -240,20 +246,31 @@ arch_cached_archive_connect (struct arch_archive ** answer, int soft_errors)
     }
 
   
-  arch = lim_realloc (0, (t_uchar *) *answer, sizeof (struct arch_cached_archive));
+  arch = talloc_realloc_size (0, *answer, sizeof (struct arch_cached_archive));
+  talloc_set_destructor (arch, cache_close);
   *answer = (struct arch_archive *)arch;
   arch->arch.vtable = &cache_vtable;
   
   wrap_location = arch_uncached_location ((*answer)->location);
-  arch->wrapped_arch = arch_archive_connect_location ((*answer)->name, wrap_location, 0, soft_errors);
+  arch->wrapped_arch = arch_archive_connect_location_ext ((*answer)->official_name, wrap_location, 0, soft_errors, 0);
   lim_free (0, wrap_location);
   if (arch->wrapped_arch)
     {
       /* this code section is copied from archive.c's .._connect, which this
        * API is meant to duplicate
        */
-      lim_free (0, arch->wrapped_arch->name);
-      arch->wrapped_arch->name = str_save (0, (*answer)->name);
+      /* what this does is set the name field in the wrapped archive
+       * to the name that the wrapping archive is connecting as
+       * this may be different from the official name.
+       * FIXME: remove this in baz 1.4 when we remove support for the old 
+       * registered-names concept
+       * RBC 200503012
+       */
+      lim_free (0, arch->wrapped_arch->registered_name);
+      arch->wrapped_arch->registered_name = str_save (0, (*answer)->registered_name);
+      /* save the policy details */
+      arch->arch.when_unsigned = arch->wrapped_arch->when_unsigned;
+      arch->arch.in_registry = arch->wrapped_arch->in_registry;
       /* not sure about these, from the same block.. 
       answer->refs = 1;
       answer->next = connected_by_name;
@@ -273,16 +290,20 @@ arch_cached_make_archive (t_uchar * name, t_uchar * uri, t_uchar * version, t_uc
 {
   t_uchar * wrap_location = arch_uncached_location (uri);
   invariant (arch_cached_archive_protocol (uri));
-  arch_make_archive (name, wrap_location, mirror_of, dot_listing_lossage, signed_archive, arch_archive_type(version) == arch_archive_tla);
-  arch_set_archive_location (name, uri, 1, ARCH_REG_FAIL_NORM);
+  arch_make_archive (name, wrap_location, mirror_of, dot_listing_lossage, signed_archive, arch_archive_type(version) == arch_archive_tla, 0);
   lim_free (0, wrap_location);
 }
 
 
-static void cache_close (struct arch_archive * a)
+int 
+cache_close (void * arch_archive_void)
 {
+  struct arch_archive * a = (struct arch_archive *) arch_archive_void;
   struct arch_archive *wrapped_arch = ((struct arch_cached_archive *) a)->wrapped_arch;
-  wrapped_arch->vtable->close (wrapped_arch);
+  if (wrapped_arch)
+      arch_archive_close (wrapped_arch);
+  wrapped_arch = NULL;
+  return arch_archive_finalise (a);
 }
 
 
@@ -390,7 +411,7 @@ arch_revision_type_enum (t_uchar * type)
 static int 
 cache_revision_type  (enum arch_revision_type * type, int * archive_cached, 
 		      int *has_ancestry, 
-                      struct arch_archive * a, t_uchar * revision)
+                      struct arch_archive * a, t_uchar const * revision)
 {
   struct arch_archive *wrapped_arch = ((struct arch_cached_archive *) a)->wrapped_arch;
   enum arch_revision_type tmp_type;
@@ -524,6 +545,15 @@ cache_get_continuation (struct arch_archive * a, t_uchar * revision)
 }
 
 
+int 
+cache_set_meta_info (struct arch_archive * a, 
+                     t_uchar * meta_info_name, 
+                     t_uchar * meta_info_value)
+{
+  struct arch_archive *wrapped_arch = ((struct arch_cached_archive *) a)->wrapped_arch;
+  return wrapped_arch->vtable->set_meta_info (wrapped_arch, meta_info_name, meta_info_value); 
+}
+
 static t_uchar * 
 cache_get_meta_info (struct arch_archive * a, t_uchar * meta_info_name)
 {
@@ -569,6 +599,7 @@ cache_lock_revision (t_uchar ** errstr, struct arch_archive * a,
 
 static int 
 cache_revision_ready (t_uchar ** errstr, struct arch_archive * a,
+		      struct arch_archive * from_archive,
                       t_uchar * version,
                       t_uchar * prev_level,
                       t_uchar * uid,
@@ -576,7 +607,7 @@ cache_revision_ready (t_uchar ** errstr, struct arch_archive * a,
                       t_uchar * new_level)
 {
   struct arch_archive *wrapped_arch = ((struct arch_cached_archive *) a)->wrapped_arch;
-  return wrapped_arch->vtable->revision_ready (errstr, wrapped_arch, version, prev_level, uid, txn_id, new_level);
+  return wrapped_arch->vtable->revision_ready (errstr, wrapped_arch, from_archive, version, prev_level, uid, txn_id, new_level);
 }
 
 
@@ -670,10 +701,11 @@ cache_put_import (t_uchar ** errstr, struct arch_archive * a,
 
 
 static int 
-cache_put_cached (t_uchar ** errstr, struct arch_archive * a, t_uchar * revision, int in_fd)
+cache_put_cached (t_uchar ** errstr, struct arch_archive * a, struct arch_archive *from_archive,
+		  t_uchar * revision, int in_fd)
 {
   struct arch_archive *wrapped_arch = ((struct arch_cached_archive *) a)->wrapped_arch;
-  return wrapped_arch->vtable->put_cached (errstr, wrapped_arch, revision, in_fd);
+  return wrapped_arch->vtable->put_cached (errstr, wrapped_arch, from_archive, revision, in_fd);
 }
 
 
@@ -701,10 +733,11 @@ cache_get_ancestry (int out_fd, struct arch_archive * a, t_uchar * revision)
 }
 
 static int 
-cache_put_ancestry (t_uchar ** errstr, struct arch_archive * a, t_uchar * revision, int in_fd)
+cache_put_ancestry (t_uchar ** errstr, struct arch_archive * a, struct arch_archive *from_archive,
+		    t_uchar * revision, int in_fd)
 {
   struct arch_archive *wrapped_arch = ((struct arch_cached_archive *) a)->wrapped_arch;
-  return wrapped_arch->vtable->put_ancestry (errstr, wrapped_arch, revision, in_fd);
+  return wrapped_arch->vtable->put_ancestry (errstr, wrapped_arch, from_archive, revision, in_fd);
 }
 
 
@@ -725,5 +758,12 @@ arch_cached_archive_as_pfs(struct arch_archive * arch)
     return result;
 }
 
+int
+cache_set_mirror (t_uchar ** errstr, struct arch_archive * archive, int enabled)
+{
+  struct arch_archive *wrapped_arch = ((struct arch_cached_archive *) archive)->wrapped_arch;
+  return arch_archive_set_mirror (errstr, wrapped_arch, enabled);
+}
+
 /* tag: b6a409d5-c4b2-48ee-bdb4-5c2eeebb2bc6 
  */
