@@ -16,7 +16,7 @@
 
 extern int arch_cached_archive_connect (struct arch_archive ** a, int soft_errors);
 
-extern t_uchar * arch_uncached_location (t_uchar *location);
+extern t_uchar * arch_uncached_location (t_uchar const *location);
 
 extern void arch_cached_make_archive (t_uchar * name, t_uchar * uri, t_uchar * version, t_uchar *mirror_of, int dot_listing_lossage, int signed_archive);
 
@@ -26,10 +26,11 @@ struct arch_cached_archive
   struct arch_archive *wrapped_arch;
 };
 
-extern int arch_cached_archive_protocol (t_uchar * location);
+extern int arch_cached_archive_protocol (t_uchar const * location);
 
-extern t_uchar * arch_revision_query (struct arch_archive * a, t_uchar *
-                                      revision, t_uchar const * extension);
+extern t_uchar * arch_revision_query (struct arch_archive * a,
+				      t_uchar const * revision, 
+				      t_uchar const * extension);
 extern struct arch_archive * arch_cached_archive_as_pfs(struct arch_archive * arch);
 #endif
 
