@@ -11,6 +11,7 @@
 #include "hackerlab/bugs/panic.h"
 #include "hackerlab/char/str.h"
 #include "hackerlab/fs/file-names.h"
+#include "hackerlab/mem/mem.h"
 #include "hackerlab/vu/safe.h"
 #include "hackerlab/arrays/ar.h"
 #include "hackerlab/char/pika-escaping-utils.h"
@@ -26,18 +27,28 @@
 
 /* __STDC__ prototypes for static functions */
 static void fix_mod_locs (assoc_table loc_of, rel_table table);
+
 static void arch_print_file_metadata_diffs (int out_fd,
                                             struct arch_changeset_report * report,
                                             int escape_classes);
 static void arch_print_dir_metadata_diffs (int out_fd,
                                            struct arch_changeset_report * report,
                                            int escape_classes);
-static void arch_print_file_diffs (int out_fd,
-                                   struct arch_changeset_report * report,
-                                   int escape_classes);
-static void arch_print_added_file_diffs (int out_fd,
-                                         struct arch_changeset_report * report,
-                                         int escape_classes);
+static void print_custom_diffs(struct arch_changeset_report * report, 
+                               t_uchar * orig_dir, t_uchar * mod_dir, 
+                               t_uchar ** opts);
+static void arch_print_file_diffs (struct arch_changeset_report * report);
+static void arch_print_added_file_diffs (struct arch_changeset_report * report);
+static void
+print_removed_file_diffs (struct arch_changeset_report * report, 
+                          t_uchar ** opts);
+static void print_file_metadata (int out_fd, 
+                                 struct arch_changeset_report * report, 
+                                 int escape_classes);
+static void print_dir_metadata (int out_fd, 
+                                struct arch_changeset_report * report, 
+                                int escape_classes);
+static void print_diffs_new(struct arch_changeset_report * report);
 static void report_unique_files_and_symlinks (rel_table * files, rel_table * symlinks,
                                               rel_table index,
                                               t_uchar * archive_dir, t_uchar * archive_basename);
@@ -377,19 +388,19 @@ arch_reverse_changeset (struct arch_changeset_report * report)
 
   for (x = 0; x < rel_n_records (report->renamed_files); ++x)
     {
-      t_uchar * s;
-
-      s = report->renamed_files[x][0];
-      report->renamed_files[x][0] = report->renamed_files[x][1];
-      report->renamed_files[x][1] = s;
+      rel_replace_record (report->renamed_files, x, 
+	      rel_make_record (report->renamed_files[x][1],
+			       report->renamed_files[x][0],
+			       report->renamed_files[x][2],
+			       NULL));
     }
   for (x = 0; x < rel_n_records (report->renamed_dirs); ++x)
     {
-      t_uchar * s;
-
-      s = report->renamed_dirs[x][0];
-      report->renamed_dirs[x][0] = report->renamed_dirs[x][1];
-      report->renamed_dirs[x][1] = s;
+      rel_replace_record (report->renamed_dirs, x, 
+	      rel_make_record (report->renamed_dirs[x][1],
+			       report->renamed_dirs[x][0],
+			       report->renamed_dirs[x][2],
+			       NULL));
     }
 
   new_mod_file_loc_of = rel_to_assoc (report->mod_files_index, 1, 0);
@@ -412,16 +423,17 @@ fix_mod_locs (assoc_table loc_of, rel_table table)
   for (x = 0; x < rel_n_records (table); ++x)
     {
       t_uchar * new_loc = 0;
-      t_uchar * old_loc = 0;
       t_uchar * id;
 
       id = table[x][1];
-      old_loc = table[x][0];
       new_loc = assoc_ref (loc_of, id);
       if (new_loc)
         new_loc = str_save (0, new_loc);
-      table[x][0] = new_loc;
-      lim_free (0, old_loc);
+      rel_replace_record (table, x, rel_make_record
+			  (new_loc,
+			   table[x][1],
+			   table[x][2],
+			   NULL));
     }
 }
 
@@ -494,9 +506,7 @@ arch_print_dir_metadata_diffs (int out_fd,
 }
 
 static void
-arch_print_file_diffs (int out_fd,
-                       struct arch_changeset_report * report,
-                       int escape_classes)
+arch_print_file_diffs (struct arch_changeset_report * report)
 {
   int x;
 
@@ -507,28 +517,25 @@ arch_print_file_diffs (int out_fd,
 
       patch_file = str_alloc_cat (0, report->patched_regular_files[x][2], ".patch");
       in_fd = safe_open (patch_file, O_RDONLY, 0);
-      copy_fd (in_fd, out_fd);
-      safe_printfmt (out_fd, "\n\n");
+      copy_fd (in_fd, 1);
+      safe_printfmt (1, "\n\n");
       safe_close (in_fd);
       lim_free (0, patch_file);
     }
 }
 
 static void
-arch_print_added_file_diffs (int out_fd,
-                             struct arch_changeset_report * report,
-                             int escape_classes)
+arch_print_added_file_diffs (struct arch_changeset_report * report)
 {
   int x;
-
   for (x = 0; x < rel_n_records (report->added_files); ++x)
     {
-      arch_really_invoke_diff (1, "/dev/null", NULL, report->added_files[x][2], report->added_files[x][0]);
+      arch_really_invoke_diff (1, "/dev/null", NULL, report->added_files[x][2], report->added_files[x][0], NULL);
     }
 }
 
-void
-arch_print_changeset_diffs (int out_fd, struct arch_changeset_report * report, int escape_classes)
+static void print_file_metadata (int out_fd, struct arch_changeset_report *
+                                 report, int escape_classes)
 {
   if (report->file_metadata_changed)
     {
@@ -536,58 +543,165 @@ arch_print_changeset_diffs (int out_fd, struct arch_changeset_report * report, i
       arch_print_file_metadata_diffs (out_fd, report, escape_classes);
       safe_printfmt (out_fd, "\n");
     }
+}
 
+
+static void print_dir_metadata (int out_fd, 
+                                struct arch_changeset_report * report, 
+                                int escape_classes)
+{
   if (report->dir_metadata_changed)
     {
       safe_printfmt (out_fd, "* dir metadata changed\n\n");
       arch_print_dir_metadata_diffs (out_fd, report, escape_classes);
       safe_printfmt (out_fd, "\n");
     }
+}
 
+static void print_diffs_new(struct arch_changeset_report * report)
+{
   if (report->patched_regular_files)
     {
-      safe_printfmt (out_fd, "* modified files\n\n");
-      arch_print_file_diffs (out_fd, report, escape_classes);
-      safe_printfmt (out_fd, "\n");
+      safe_printfmt (1, "* modified files\n\n");
+      arch_print_file_diffs (report);
+      safe_printfmt (1, "\n");
     }
   if (report->added_files)
     {
-      safe_printfmt (out_fd, "* added files\n\n");
-      arch_print_added_file_diffs (out_fd, report, escape_classes);
-      safe_printfmt (out_fd, "\n");
+      safe_printfmt (1, "* added files\n\n");
+      arch_print_added_file_diffs (report);
+      safe_printfmt (1, "\n");
+    }
+  print_removed_file_diffs(report, NULL);
+}
+
+
+/**
+ * \brief Write custom diffs for to a file descriptor
+ *
+ * The changeset report is used to determine which files to diff, and the
+ * partial file paths.  The orig_dir and mod_dir are used to produce full
+ * paths.  diff is invoked to produce the output.
+ *
+ * \param report The report to use for output
+ * \param orig_dir The path to a copy of the ORIG tree
+ * \param mod_dir The path to a copy of the MOD tree
+ */
+static void
+print_custom_diffs (struct arch_changeset_report * report, t_uchar * orig_dir, t_uchar * mod_dir, t_uchar ** opts)
+{
+  if (report->patched_regular_files)
+    {
+      int x;
+      assoc_table orig_paths = rel_to_assoc (report->orig_files_index, 1, 0);
+
+      safe_printfmt (1, "* modified files\n\n");
+      for (x = 0; x < rel_n_records (report->patched_regular_files); ++x)
+        {
+          t_uchar * id = report->patched_regular_files[x][1];
+          t_uchar * orig_part_path = assoc_ref(orig_paths, id);
+          t_uchar * orig_path = file_name_in_vicinity(0, orig_dir, orig_part_path);
+          t_uchar * mod_path = file_name_in_vicinity(0, mod_dir, report->patched_regular_files[x][0]);
+
+          arch_really_invoke_diff (1, orig_path,  orig_part_path, mod_path, report->patched_regular_files[x][0], (char **)opts);
+
+          lim_free(0, mod_path);
+          lim_free(0, orig_path);
+        }
+      safe_printfmt (1, "\n");
+      free_assoc_table (orig_paths);
+    }
+  if (report->added_files)
+    {
+      int x;
+      safe_printfmt (1, "* added files\n\n");
+
+      for (x = 0; x < rel_n_records (report->added_files); ++x)
+        {
+          arch_really_invoke_diff (1, "/dev/null", NULL, report->added_files[x][2], report->added_files[x][0], (char**)opts);
+        }
+      safe_printfmt (1, "\n");
+    }
+  print_removed_file_diffs (report, opts);
+}
+
+
+/**
+ * \brief Print the diffs for any removed files to output.
+ *
+ * If no files were removed, no output is produced
+ * \param report The changeset report to generate results for
+ * \param opts Additional parameters to pass to diff, or NULL
+ */
+static void
+print_removed_file_diffs (struct arch_changeset_report * report, t_uchar ** opts)
+{
+  if (report->removed_files)
+    {
+      int x;
+      safe_printfmt (1, "* removed files\n\n");
+
+      for (x = 0; x < rel_n_records (report->removed_files); ++x)
+        {
+          arch_really_invoke_diff (1, report->removed_files[x][2], report->removed_files[x][0], "/dev/null", NULL, (char**)opts);
+        }
+      safe_printfmt (1, "\n");
     }
 }
 
+
+/**
+ * \brief Write a changeset report with custom diffs to a file descriptor
+ *
+ * \param report The report to use for output
+ * \param orig_dir The path to a copy of the ORIG tree
+ * \param mod_dir The path to a copy of the MOD tree
+ * \param escape_classes The types of characters that should be escaped
+ * \todo Consider reimplementing so that a custom changeset is generated, not
+ * custom diffs.
+ */
 void
-arch_print_changeset  (int out_fd, struct arch_changeset_report * report, int diffs, int escape_classes)
+arch_print_changeset_custom_diffs (struct arch_changeset_report * report, t_uchar * orig_dir, t_uchar * mod_dir, t_uchar ** opts, int escape_classes)
+{
+  print_file_metadata (1, report, escape_classes);
+  print_dir_metadata (1, report, escape_classes);
+  if (opts == NULL)
+    print_diffs_new (report);
+  else
+    print_custom_diffs (report, orig_dir, mod_dir, opts);
+}
+
+void
+arch_print_changeset  (struct arch_changeset_report * report, int diffs, int escape_classes)
 {
   int x;
+  int out_fd = 1;
 
   if (report->removed_dirs)
     {
-      safe_printfmt (out_fd, "* removed directories\n\n");
+      safe_printfmt (1, "* removed directories\n\n");
       for (x = 0; x < rel_n_records (report->removed_dirs); ++x)
         {
           t_uchar * item;
           item = pika_save_escape_iso8859_1 (0, 0, escape_classes,
                                              report->removed_dirs[x][0]);
-          safe_printfmt (out_fd, "    %s\n", no_dot (item));
+          safe_printfmt (1, "    %s\n", no_dot (item));
           lim_free (0, item);
         }
-      safe_printfmt (out_fd, "\n");
+      safe_printfmt (1, "\n");
     }
   if (report->added_dirs)
     {
-      safe_printfmt (out_fd, "* added directories\n\n");
+      safe_printfmt (1, "* added directories\n\n");
       for (x = 0; x < rel_n_records (report->added_dirs); ++x)
         {
           t_uchar * item;
           item = pika_save_escape_iso8859_1 (0, 0, escape_classes,
                                              report->added_dirs[x][0]);
-          safe_printfmt (out_fd, "    %s\n", no_dot (item));
+          safe_printfmt (1, "    %s\n", no_dot (item));
           lim_free (0, item);
         }
-      safe_printfmt (out_fd, "\n");
+      safe_printfmt (1, "\n");
     }
 
   if (report->removed_files)
@@ -771,7 +885,7 @@ arch_print_changeset  (int out_fd, struct arch_changeset_report * report, int di
           safe_printfmt (out_fd, "    %s\n", no_dot (item));
           if (diffs)
             {
-              arch_print_added_file_diffs (out_fd, report, escape_classes);
+              arch_print_added_file_diffs (report);
             }
           lim_free (0, item);
         }
@@ -784,7 +898,7 @@ arch_print_changeset  (int out_fd, struct arch_changeset_report * report, int di
 
       if (diffs)
         {
-          arch_print_file_diffs (out_fd, report, escape_classes);
+          arch_print_file_diffs (report);
         }
       else
         {
@@ -1118,6 +1232,11 @@ find_renames (rel_table * out, rel_table common, assoc_table orig_dir_id_of, ass
     }
 }
 
+void
+arch_changeset_report_init (struct arch_changeset_report * changeset)
+{
+  mem_set0 ((t_uchar *)changeset, sizeof (*changeset));
+}
 
 
 
