@@ -9,6 +9,7 @@
  */
 
 
+#include "hackerlab/bugs/exception.h"
 #include "hackerlab/bugs/panic.h"
 #include "hackerlab/os/time.h"
 #include "hackerlab/char/str.h"
@@ -47,7 +48,7 @@
 static t_uchar * arch_prepare_commit_changeset (t_uchar ** changelog_loc_ret,
                                                 int chatter_fd,
                                                 t_uchar ** cooked_log_ret,
-                                                t_uchar * tree_root,
+                                                arch_project_tree_t * tree,
                                                 t_uchar * raw_log,
                                                 struct arch_archive * arch,
                                                 t_uchar * revision,
@@ -55,9 +56,8 @@ static t_uchar * arch_prepare_commit_changeset (t_uchar ** changelog_loc_ret,
                                                 rel_table file_list,
                                                 int escape_classes);
 static rel_table pick_patch_logs (rel_table table);
-static void arch_commit_mid_commit (t_uchar * tree_root, t_uchar * cooked_log);
-static void arch_finish_commit (int chatter_fd, t_uchar * tree_root, t_uchar * archive, t_uchar * prev_revision, t_uchar * revision, t_uchar * changeset, t_uchar * changelog_loc, int escape_classes);
-static void arch_commit_failed (t_uchar * tree_root, t_uchar * archive, t_uchar * revision, t_uchar * changeset);
+static void arch_recycle_pristine (int chatter_fd, arch_project_tree_t * tree, t_uchar * archive, t_uchar * prev_revision, t_uchar * revision, t_uchar * changeset, t_uchar * changelog_loc, int escape_classes);
+static void arch_commit_failed (arch_project_tree_t * tree, t_uchar * archive, t_uchar * revision, t_uchar * changeset);
 static void commit_make_changeset_callback (void * vfd, char * fmt, va_list ap);
 
 
@@ -66,7 +66,7 @@ int
 arch_commit (int chatter_fd,
              struct arch_archive * arch,
              t_uchar * revision,
-             t_uchar * tree_root,
+             arch_project_tree_t * tree,
              t_uchar * raw_log,
              rel_table file_list,
              int just_commit,
@@ -75,8 +75,7 @@ arch_commit (int chatter_fd,
   t_uchar * errstr;
   t_uchar * version = 0;
   t_uchar * this_level = 0;
-  t_uchar * prev_revision = 0;
-  t_uchar * prev_level = 0;
+  arch_patch_id * prev_revision;
   t_uchar * changelog_loc = 0;
   t_uchar * changeset_path = 0;
   t_uchar * cooked_log = 0;
@@ -85,21 +84,21 @@ arch_commit (int chatter_fd,
   int error;
   int result = 0;
   int do_cacherev = 0;
-
+  arch_patch_id * patch_id = arch_patch_id_new_archive (arch->official_name, revision);
+  
   invariant (!!raw_log);
 
   version = arch_parse_package_name (arch_ret_package_version, 0, revision);
   this_level = arch_parse_package_name (arch_ret_patch_level, 0, revision);
-  prev_revision = arch_previous_revision (arch, revision);
-  prev_level = arch_parse_package_name (arch_ret_patch_level, 0, prev_revision);
+  prev_revision = arch_previous_revision (arch, patch_id);
 
-  changeset_path = arch_prepare_commit_changeset (&changelog_loc, chatter_fd, &cooked_log, tree_root, raw_log, arch, revision, prev_level, file_list, escape_classes);
+  changeset_path = arch_prepare_commit_changeset (&changelog_loc, chatter_fd, &cooked_log, tree, raw_log, arch, revision, arch_patch_id_patchlevel (prev_revision), file_list, escape_classes);
 
-  invariant (!!prev_level);
+  invariant (!!arch_patch_id_patchlevel (prev_revision));
 
   /* Check the error return code for the "precommit" hook and exit if non-zero.
    */
-  error = arch_run_hook ("precommit", "ARCH_ARCHIVE", arch->name, "ARCH_REVISION", revision, "ARCH_TREE_ROOT", tree_root, 0) ;
+  error = arch_run_hook ("precommit", "ARCH_ARCHIVE", arch->official_name, "ARCH_REVISION", revision, "ARCH_TREE_ROOT", tree->root, 0) ;
   if (error)
     {
       safe_printfmt (2, "arch_commit: precommit hook function failed with error (%d)\n commit cancelled.\n", error);
@@ -109,93 +108,98 @@ arch_commit (int chatter_fd,
   my_uid = arch_my_id_uid ();
   txn_id = arch_generate_txn_id ();
 
-  if (arch_archive_lock_revision (&errstr, arch, version, prev_level, my_uid, txn_id, this_level))
+  if (arch_archive_lock_revision (&errstr, arch, version, arch_patch_id_patchlevel (prev_revision), my_uid, txn_id, this_level))
     {
-      safe_printfmt (2, "arch_commit: unable to acquire revision lock (%s)\n    tree: %s\n    revision: %s/%s\n",
-                     errstr, tree_root, arch->name, revision);
+      safe_printfmt (2, "arch_commit: unable to acquire revision lock (%s)\n    tree: %s\n    revision: %s/%s\n    url: %s\n",
+                     errstr, tree->root, arch->official_name, revision, arch->location);
       exit (2);
     }
 
-  if (arch_archive_put_log (&errstr, arch, version, prev_level, my_uid, txn_id, cooked_log))
+  if (arch_archive_put_log (&errstr, arch, version, arch_patch_id_patchlevel (prev_revision), my_uid, txn_id, cooked_log))
     {
-      safe_printfmt (2, "arch_commit: unable to send log message to archive (%s)\n    tree: %s\n    revision: %s/%s\n",
-                     errstr, tree_root, arch->name, version);
+      safe_printfmt (2, "arch_commit: unable to send log message to archive (%s)\n    tree: %s\n    revision: %s/%s\n    url: %s\n",
+                     errstr, tree->root, arch->official_name, version, arch->location);
       exit (2);
     }
 
-  if (arch_archive_put_changeset (&errstr, arch, version, prev_level, my_uid, txn_id, this_level, changeset_path))
+  if (arch_archive_put_changeset (&errstr, arch, version, arch_patch_id_patchlevel (prev_revision), my_uid, txn_id, this_level, changeset_path))
     {
-      safe_printfmt (2, "arch_commit: unable to send changeset to archive (%s)\n    tree: %s\n    revision: %s/%s\n",
-                     errstr, tree_root, arch->name, revision);
+      safe_printfmt (2, "arch_commit: unable to send changeset to archive (%s)\n    tree: %s\n    revision: %s/%s\n    url: %s\n",
+                     errstr, tree->root, arch->official_name, revision, arch->location);
       exit (2);
     }
 
-  if (arch_revision_ready (&errstr, arch, version, prev_level, my_uid, txn_id, this_level))
+  if (arch_revision_ready (&errstr, arch, version, arch_patch_id_patchlevel (prev_revision), my_uid, txn_id, this_level))
     {
-      safe_printfmt (2, "arch_commit: error sending revision to archive (%s)\n    tree: %s\n    revision: %s/%s\n",
-                     errstr, tree_root, arch->name, revision);
+      safe_printfmt (2, "arch_commit: error sending revision to archive (%s)\n    tree: %s\n    revision: %s/%s\n    url: %s\n",
+                     errstr, tree->root, arch->official_name, revision, arch->location);
       /* FIXME: detail unlock failures to the user gracefully */
-      invariant (!arch_archive_break_revision_lock (&errstr, arch, version, prev_level, my_uid, txn_id));
+      invariant (!arch_archive_break_revision_lock (&errstr, arch, version, arch_patch_id_patchlevel (prev_revision), my_uid, txn_id));
       result = -1;
       goto error_exit;
     }
 
+  arch_start_tree_commit (tree, cooked_log);
 
-  arch_commit_mid_commit (tree_root, cooked_log);
-
-  if (arch_archive_finish_revision (&errstr, arch, version, prev_level, my_uid, txn_id, this_level))
+  if (arch_archive_finish_revision (&errstr, arch, version, arch_patch_id_patchlevel (prev_revision), my_uid, txn_id, this_level))
     {
-      arch_commit_failed (tree_root, arch->name, version, changeset_path);
-      safe_printfmt (2, "arch_commit: unable to complete commit transaction (%s)\n    tree: %s\n    revision: %s/%s\n",
-                     errstr, tree_root, arch->name, version);
+      arch_commit_failed (tree, arch->official_name, version, changeset_path);
+      safe_printfmt (2, "arch_commit: unable to complete commit transaction (%s)\n    tree: %s\n    revision: %s/%s\n    url: %s\n",
+                     errstr, tree->root, arch->official_name, version, arch->location);
       exit (2);
     }
 
   if (!just_commit)
     {
-      t_uchar *fqrevision = arch_fully_qualify (arch->official_name, revision);
+      struct exception * e;
 
-      arch_project_tree_t tree;
-      rel_table ancestry;
-      arch_project_tree_init (&tree, tree_root);
-      
       /* FIXME upload limited history and update a current dir in the archive */
-      ancestry_upload_patch (arch, fqrevision, -1);
+      ancestry_upload_patch (arch, patch_id, -1);
 
-      /* pull in the tree ancestry */
-      /* bah again with the needing policy stuff */
-      ancestry = patch_ancestry (&tree, fqrevision, -1, 1);
-      do_cacherev =  rel_n_records (ancestry) % MAGIC_CACHEREV_INTERVAL == 2;
-      lim_free (0, fqrevision);
-      rel_free_table (ancestry);
-      arch_project_tree_finalise (&tree);
+      Try
+        {
+          /* pull in the tree ancestry */
+          /* bah again with the needing policy stuff */
+          rel_table ancestry = patch_ancestry (talloc_context, tree, patch_id, -1);
+          do_cacherev =  rel_n_records (ancestry) % MAGIC_CACHEREV_INTERVAL == 2;
+          rel_free_table (ancestry);
+        }
+      Catch (e)
+        {
+          /* invariants */
+          if (e->code < 0)
+              Throw (e);
+          talloc_free (e);
+        }
     }
 
-  arch_maybe_cache_commit (arch, revision, arch->name, prev_revision, changeset_path);
-  arch_finish_commit (chatter_fd, tree_root, arch->name, prev_revision, revision, changeset_path, changelog_loc, escape_classes);
+  arch_maybe_cache_commit (arch, revision, arch->official_name, arch_patch_id_revision (prev_revision), changeset_path);
+  arch_finish_tree_commit (tree, arch->official_name, revision, changelog_loc);
+  arch_recycle_pristine (chatter_fd, tree, arch->official_name, arch_patch_id_revision (prev_revision), revision, changeset_path, changelog_loc, escape_classes);
+  
+  arch_run_hook ("commit", "ARCH_ARCHIVE", arch->official_name, "ARCH_REVISION", revision, "ARCH_TREE_ROOT", tree->root, 0);
 
   if (rel_n_records (file_list) == 0) 
-    arch_snap_inode_sig (tree_root, arch->name, revision);
+    arch_snap_inode_sig_not_reference (tree, arch->official_name, revision);
   else
   {
-    arch_snap_inode_sig_files(tree_root, arch->name, prev_revision, revision, file_list);
+    arch_snap_inode_sig_not_reference_files(tree, arch->official_name, arch_patch_id_revision (prev_revision), revision, file_list);
   }
   
   if (do_cacherev)
-      arch_archive_cache (2, arch, arch->official_name, revision, 0);
-
-  arch_run_hook ("commit", "ARCH_ARCHIVE", arch->name, "ARCH_REVISION", revision, "ARCH_TREE_ROOT", tree_root, 0);
+      arch_archive_cache (2, arch, arch->official_name, revision, tree);
+  
 
 error_exit:
   lim_free (0, version);
   lim_free (0, this_level);
-  lim_free (0, prev_revision);
-  lim_free (0, prev_level);
+  talloc_free (prev_revision);
   lim_free (0, changeset_path);
   lim_free (0, cooked_log);
   lim_free (0, changelog_loc);
   lim_free (0, my_uid);
   lim_free (0, txn_id);
+  talloc_free (patch_id);
   return result;
 }
 
@@ -206,7 +210,7 @@ static t_uchar *
 arch_prepare_commit_changeset (t_uchar ** changelog_loc_ret,
                                int chatter_fd,
                                t_uchar ** cooked_log_ret,
-                               t_uchar * tree_root,
+                               arch_project_tree_t * tree,
                                t_uchar * raw_log,
                                struct arch_archive * arch,
                                t_uchar * revision,
@@ -214,7 +218,7 @@ arch_prepare_commit_changeset (t_uchar ** changelog_loc_ret,
                                rel_table file_list,
                                int escape_classes)
 {
-  t_uchar * archive = arch->name;
+  t_uchar * archive = arch->official_name;
   t_uchar * tmp_stem = 0;
   t_uchar * tmp_path = 0;
   t_uchar * changeset_basename = 0;
@@ -223,6 +227,7 @@ arch_prepare_commit_changeset (t_uchar ** changelog_loc_ret,
   t_uchar * level = 0;
   t_uchar * prev_revision = 0;
   t_uchar * prev_rev_path = 0;
+  arch_project_tree_t * prev_rev_tree;
   struct arch_make_changeset_report make_report = {0, }; /* should have a chatter callback here */
   t_uchar * changelog_id_suffix = 0;
   t_uchar * changelog_x_id = 0;
@@ -252,7 +257,7 @@ arch_prepare_commit_changeset (t_uchar ** changelog_loc_ret,
    * make a temp dir for the changeset
    */
   tmp_stem = str_alloc_cat_many (0, ",,commit.", revision, "--", archive, str_end);
-  tmp_path = tmp_file_name (tree_root, tmp_stem);
+  tmp_path = talloc_tmp_file_name (talloc_context, tree->root, tmp_stem);
 
   rmrf_file (tmp_path);
   safe_mkdir (tmp_path, 0777);
@@ -271,7 +276,8 @@ arch_prepare_commit_changeset (t_uchar ** changelog_loc_ret,
   version = arch_parse_package_name (arch_ret_package_version, 0, revision);
   level = arch_parse_package_name (arch_ret_patch_level, 0, revision);
   prev_revision = str_alloc_cat_many (0, version, "--", prev_level, str_end);
-  prev_rev_path = arch_find_or_make_local_copy (chatter_fd, tree_root, 0, arch, archive, prev_revision);
+  prev_rev_path = arch_find_or_make_local_copy (chatter_fd, tree, 0, arch, archive, prev_revision);
+  prev_rev_tree = arch_project_tree_new (talloc_context, prev_rev_path);
 
   if (chatter_fd >= 0)
     {
@@ -283,14 +289,14 @@ arch_prepare_commit_changeset (t_uchar ** changelog_loc_ret,
     {
       assoc_table inode_shortcut = 0;
 
-      arch_read_inode_sig_ids (0, &inode_shortcut, tree_root, archive, prev_revision);
+      arch_read_inode_sig_ids (0, &inode_shortcut, tree->root, archive, prev_revision);
 
-      arch_make_changeset (&make_report, prev_rev_path, tree_root, changeset_path, arch_unspecified_id_tagging, arch_inventory_unrecognized, 0, inode_shortcut, 0, escape_classes);
+      arch_make_changeset (&make_report, prev_rev_tree, tree, changeset_path, arch_unspecified_id_tagging, arch_inventory_unrecognized, 0, inode_shortcut, 0, escape_classes);
 
       free_assoc_table (inode_shortcut);
     }
   else
-    arch_make_files_changeset (&make_report, changeset_path, file_list, prev_rev_path, tree_root, arch_unspecified_id_tagging, arch_inventory_unrecognized, escape_classes);
+    arch_make_files_changeset (&make_report, changeset_path, file_list, prev_rev_tree, tree, arch_unspecified_id_tagging, arch_inventory_unrecognized, escape_classes);
 
   /****************************************************************
    * Look for a Changelog for the version we're committing to.
@@ -525,7 +531,7 @@ arch_prepare_commit_changeset (t_uchar ** changelog_loc_ret,
       invariant (changelog_add_fd >= 0);
       invariant (changelog_diffs_fd < 0);
 
-      arch_generate_changelog (changelog_add_fd, tree_root, 0, 0, level, new_log_path, archive, version);
+      arch_generate_changelog (changelog_add_fd, tree, 0, 0, level, new_log_path, archive, version);
       /* new_changelog_path already set. */
       safe_close (changelog_add_fd);
     }
@@ -536,10 +542,10 @@ arch_prepare_commit_changeset (t_uchar ** changelog_loc_ret,
 
       new_changelog_path = file_name_in_vicinity (0, changeset_path, ",,changelog");
       fd = safe_open (new_changelog_path, O_WRONLY | O_CREAT | O_EXCL, 0444);
-      arch_generate_changelog (fd, tree_root, 0, 0, level, new_log_path, archive, version);
+      arch_generate_changelog (fd, tree, 0, 0, level, new_log_path, archive, version);
       safe_close (fd);
 
-      changelog_orig_path = file_name_in_vicinity (0, prev_rev_path, changelog_orig_loc);
+      changelog_orig_path = file_name_in_vicinity (0, prev_rev_tree->root, changelog_orig_loc);
 
       invariant (changelog_diffs_fd >= 0);
       invariant (changelog_add_fd < 0);
@@ -554,12 +560,13 @@ arch_prepare_commit_changeset (t_uchar ** changelog_loc_ret,
     }
 
   lim_free (0, tmp_stem);
-  lim_free (0, tmp_path);
+  talloc_free (tmp_path);
   lim_free (0, changeset_basename);
   lim_free (0, version);
   lim_free (0, level);
   lim_free (0, prev_revision);
   lim_free (0, prev_rev_path);
+  arch_project_tree_delete (prev_rev_tree);
   arch_free_make_changeset_report_data (&make_report);
   lim_free (0, changelog_id_suffix);
   lim_free (0, changelog_x_id);
@@ -677,59 +684,58 @@ pick_patch_logs (rel_table table)
 
 
 static void
-arch_commit_mid_commit (t_uchar * tree_root, t_uchar * cooked_log)
-{
-  arch_start_tree_commit (tree_root, cooked_log);
-}
-
-
-static void
-arch_finish_commit (int chatter_fd, t_uchar * tree_root, t_uchar * archive, t_uchar * prev_revision, t_uchar * revision, t_uchar * changeset, t_uchar * changelog_loc, int escape_classes)
+arch_recycle_pristine (int chatter_fd, arch_project_tree_t * tree, t_uchar * archive, t_uchar * prev_revision, t_uchar * revision, t_uchar * changeset, t_uchar * changelog_loc, int escape_classes)
 {
   t_uchar * dir = 0;
   t_uchar * dir_tail = 0;
-  t_uchar * recycled_pristine_src = 0;
-  t_uchar * recycled_pristine_tmp = 0;
+  arch_project_tree_t * recycled_pristine_src = 0;
 
   dir = file_name_directory_file (0, changeset);
   dir_tail = file_name_tail (0, dir);
   invariant (!str_cmp_prefix (",,", dir_tail));
 
-  arch_finish_tree_commit (tree_root, archive, revision, changelog_loc);
-
-  recycled_pristine_src = arch_find_pristine (0, tree_root, archive, prev_revision, arch_unlocked_pristine, arch_tree_pristine_search);
+  {
+    arch_patch_id * prev = arch_patch_id_new_archive (archive, prev_revision);
+    recycled_pristine_src = arch_find_pristine (tree, prev, arch_tree_pristine_search);
+    talloc_free (prev);
+  }
   if (recycled_pristine_src)
     {
       t_uchar * inventory_path = 0;
-      struct arch_apply_changeset_report report = {0, };
+      struct arch_apply_changeset_report * report;
+      arch_project_tree_t * recycled_tree;
+      t_uchar * recycled_pristine_tmp = 0;
 
       arch_chatter (chatter_fd, "* update pristine tree (%s/%s => %s)\n", archive, prev_revision, revision);
 
-      recycled_pristine_tmp = tmp_file_name (dir, ",,pristine");
-      safe_rename (recycled_pristine_src, recycled_pristine_tmp);
+      recycled_pristine_tmp = talloc_tmp_file_name (talloc_context, dir, ",,pristine");
+      safe_rename (recycled_pristine_src->root, recycled_pristine_tmp);
 
       inventory_path = file_name_in_vicinity (0, recycled_pristine_tmp, ",,index");
       rmrf_file (inventory_path);
+      recycled_tree = arch_project_tree_new (talloc_context, recycled_pristine_tmp);
 
-      arch_apply_changeset (&report, changeset, recycled_pristine_tmp, arch_unspecified_id_tagging, arch_inventory_unrecognized, 0, 0, escape_classes);
-      invariant (!arch_conflicts_occured (&report));
+      report = arch_apply_changeset (changeset, talloc_context, recycled_tree, arch_unspecified_id_tagging, arch_inventory_unrecognized, 0, 0, escape_classes, NULL, NULL);
+      invariant (!arch_conflicts_occurred (report));
+      talloc_free (report);
 
-      arch_install_pristine (tree_root, archive, revision, recycled_pristine_tmp);
+      arch_install_pristine (tree, archive, revision, recycled_pristine_tmp);
 
+      arch_project_tree_delete (recycled_tree);
       lim_free (0, inventory_path);
+      talloc_free (recycled_pristine_tmp);
     }
 
   rmrf_file (dir);
 
   lim_free (0, dir);
   lim_free (0, dir_tail);
-  lim_free (0, recycled_pristine_src);
-  lim_free (0, recycled_pristine_tmp);
+  arch_project_tree_delete (recycled_pristine_src);
 }
 
 
 static void
-arch_commit_failed (t_uchar * tree_root, t_uchar * archive, t_uchar * revision, t_uchar * changeset)
+arch_commit_failed (arch_project_tree_t * tree, t_uchar * archive, t_uchar * revision, t_uchar * changeset)
 {
   t_uchar * dir = 0;
   t_uchar * dir_tail = 0;
@@ -738,8 +744,7 @@ arch_commit_failed (t_uchar * tree_root, t_uchar * archive, t_uchar * revision,
   dir_tail = file_name_tail (0, dir);
   invariant (!str_cmp_prefix (",,", dir_tail));
 
-  arch_abort_tree_commit (tree_root, archive, revision);
-
+  arch_abort_tree_commit (tree, archive, revision);
 
   rmrf_file (dir);
 
