@@ -12,13 +12,14 @@
 
 
 #include "libarch/archive.h"
+#include "libarch/project-tree.h"
 
 
 #define MAGIC_CACHEREV_INTERVAL 50
 extern int arch_commit (int chatter_fd,
                          struct arch_archive * arch,
                          t_uchar * revision,
-                         t_uchar * tree_root,
+                         arch_project_tree_t * tree_root,
                          t_uchar * raw_log,
                          rel_table file_list,
 			 int just_commit, 
