@@ -16,7 +16,7 @@
 
 
 void
-arch_copy_project_tree (t_uchar * from, t_uchar * to, int precious, int nested)
+arch_copy_project_tree (arch_project_tree_t * from, t_uchar * to, int precious, int nested)
 {
   rel_table inventory = 0;
 
@@ -24,7 +24,7 @@ arch_copy_project_tree (t_uchar * from, t_uchar * to, int precious, int nested)
 
   safe_mkdir (to, 0777);
 
-  copy_file_list (to, from, inventory);
+  copy_file_list (to, from->root, inventory);
 
   rel_free_table (inventory);
 }
