@@ -19,6 +19,7 @@
 #include "libawk/relational.h"
 #include "libawk/relassoc.h"
 #include "libarch/arch.h"
+#include "libarch/debug.h"
 #include "libarch/namespace.h"
 #include "libarch/invent.h"
 #include "libarch/inode-sig.h"
@@ -33,25 +34,26 @@ static t_uchar const * INODE_PATHS = "{arch}/,,inode-sig-paths";
 
 /* __STDC__ prototypes for static functions */
 static void arch_update_inode_sig (rel_table sig, assoc_table newsig, assoc_table id_list);
-static void inode_sig_callback (t_uchar * path, struct stat * stat_buf,
-                    enum arch_inventory_category category,
-                    t_uchar * id, int has_source_name,
-                    void * closure,
-                    int escape_classes);
+static void inode_sig_callback (void * closure, invent_callback_data_t const * const data);
+static void inode_sig_callback_not_reference (void * closure, invent_callback_data_t const * const data);
 static void arch_update_inode_sig (rel_table sig, assoc_table newsig, assoc_table id_list);
-static t_uchar * arch_inode_sig_dir (t_uchar * tree_root, t_uchar const * sig_type);
-static t_uchar * arch_inode_sig_file (t_uchar * tree_root, t_uchar * archive, t_uchar * revision, t_uchar const * sig_type);
-static t_uchar * arch_inode_sig_tmp_file (t_uchar * tree_root, t_uchar * archive, t_uchar * revision, t_uchar const * sig_type);
-static int arch_creat_inode_sig_file (t_uchar * tree_root, t_uchar * archive, t_uchar * revision, t_uchar const * sig_type);
-static void arch_finish_inode_sig_file (t_uchar * tree_root, t_uchar * archive, t_uchar * revision, t_uchar const * sig_type, int fd);
-static void arch_prune_inode_sig_dir (t_uchar * tree_root, t_uchar * current_inode_sig_file, t_uchar const * sig_type);
+static t_uchar * arch_inode_sig_dir (t_uchar const * const tree_root, t_uchar const * sig_type);
+static t_uchar * arch_inode_sig_file (t_uchar const * const tree_root, t_uchar const * const archive, t_uchar const * const revision, t_uchar const * sig_type);
+static t_uchar * arch_inode_sig_tmp_file (t_uchar const * const tree_root, t_uchar const * const archive, t_uchar const * const revision, t_uchar const * sig_type);
+static int arch_creat_inode_sig_file (t_uchar const * const tree_root, t_uchar const * const archive, t_uchar const * const revision, t_uchar const * sig_type);
+static void arch_finish_inode_sig_file (t_uchar const * const tree_root, t_uchar const * const archive, t_uchar const * const revision, t_uchar const * sig_type, int fd);
+static void arch_prune_inode_sig_dir (t_uchar const * const tree_root, t_uchar const * const current_inode_sig_file, t_uchar const * sig_type, int max_sigs);
 static void prune_old_ctime_cruft (rel_table the_table);
 static void prune_old_dev_cruft (rel_table the_table);
 static void arch_read_inode_sig (inode_sig * sig, t_uchar * tree_root, t_uchar * archive, t_uchar * revision);
 static void arch_read_inode_sig_paths (rel_table * as_table, assoc_table * as_assoc,
-                                       t_uchar * tree_root, t_uchar * archive, t_uchar * revision);
-static void arch_write_inode_sig_file (inode_sig sig, t_uchar * tree_root, t_uchar * archive, t_uchar * revision);
+                                       t_uchar const * const tree_root, t_uchar const * const archive, t_uchar const * const revision);
+static void arch_write_inode_sig_file (inode_sig sig, t_uchar const * const tree_root, t_uchar const * const archive, t_uchar const * const revision);
 static int arch_inode_sig_equal (int current_result, rel_table existing_sig, rel_table current_sig, int chatter_fd);
+static void arch_snap_common (arch_project_tree_t * tree, t_uchar const * const archive, t_uchar const * const revision, inode_sig sig);
+static void arch_snap_common_files (arch_project_tree_t * tree, t_uchar * archive, t_uchar * prev_revision, t_uchar * revision, rel_table file_list, inode_sig sig);
+static inode_sig arch_tree_inode_sig_not_reference (arch_project_tree_t * tree);
+static inode_sig arch_tree_inode_sig_common (arch_project_tree_t * tree, void (*sig_callback )(void * closure, invent_callback_data_t const * const data));
 
 
 
@@ -68,9 +70,20 @@ arch_statb_inode_sig (struct stat * statb)
   return string_fd_close (fd);
 }
 
+inode_sig
+arch_tree_inode_sig (arch_project_tree_t * tree)
+{
+  return arch_tree_inode_sig_common (tree, inode_sig_callback);
+}
 
 inode_sig
-arch_tree_inode_sig (t_uchar * tree_root)
+arch_tree_inode_sig_not_reference (arch_project_tree_t * tree)
+{
+  return arch_tree_inode_sig_common (tree, inode_sig_callback_not_reference);
+}
+
+inode_sig
+arch_tree_inode_sig_common (arch_project_tree_t * tree, void (*sig_callback )(void * closure, invent_callback_data_t const * const data))
 {
   struct arch_inventory_options options = {0, };
   int here_fd;
@@ -78,17 +91,21 @@ arch_tree_inode_sig (t_uchar * tree_root)
   
   answer.ids = 0;
   answer.paths = 0;
+  answer.start_time = time(NULL);
+  if (answer.start_time == (time_t)-1)
+      panic ("Could not get current time");
 
   here_fd = safe_open (".", O_RDONLY, 0);
-  safe_chdir (tree_root);
+  safe_chdir (tree->root);
 
   options.categories = arch_inventory_source;
   options.want_ids = 1;
   options.include_excluded = 1;
+  options.treat_unrecognized_source_as_source = 1;
 
-  arch_get_inventory_naming_conventions (&options, ".");
+  arch_get_inventory_naming_conventions (&options, tree);
 
-  arch_inventory_traversal (&options, ".", inode_sig_callback, (void *)&answer, 0);
+  arch_project_tree_changeset_inventory_traversal (tree, &options, sig_callback, &answer);
 
   arch_free_inventory_naming_conventions (&options);
 
@@ -101,15 +118,29 @@ arch_tree_inode_sig (t_uchar * tree_root)
 }
 
 void
-arch_snap_inode_sig (t_uchar * tree_root, t_uchar * archive, t_uchar * revision)
+arch_snap_inode_sig (arch_project_tree_t * tree, arch_patch_id * patch_id)
 {
-  inode_sig sig = arch_tree_inode_sig (tree_root);
-  arch_write_inode_sig_file (sig, tree_root, archive, revision);
+  inode_sig sig = arch_tree_inode_sig (tree);
+  arch_snap_common (tree, arch_patch_id_archive (patch_id), arch_patch_id_revision (patch_id), sig);
+}
+
+void
+arch_snap_inode_sig_not_reference (arch_project_tree_t * tree, t_uchar const * const archive, t_uchar const * const revision)
+{
+  inode_sig sig = arch_tree_inode_sig_not_reference (tree);
+  arch_snap_common (tree, archive, revision, sig);
+}
+
+void
+arch_snap_common (arch_project_tree_t * tree, t_uchar const * const archive, t_uchar const * const revision, inode_sig sig)
+{
+  debug (dbg_invent, 6, "snapped inode sig for %s\n", tree->root);
+  arch_write_inode_sig_file (sig, tree->root, archive, revision);
   arch_inode_sig_free (sig);
 }
 
 static void
-arch_write_inode_sig_file (inode_sig sig, t_uchar * tree_root, t_uchar * archive, t_uchar * revision)
+arch_write_inode_sig_file (inode_sig sig, t_uchar const * const tree_root, t_uchar const * const archive, t_uchar const * const revision)
 {
   int fd;
   
@@ -123,25 +154,39 @@ arch_write_inode_sig_file (inode_sig sig, t_uchar * tree_root, t_uchar * archive
 }
  
 void
-arch_snap_inode_sig_files (t_uchar * tree_root, t_uchar * archive, t_uchar * prev_revision, t_uchar * revision, rel_table file_list)
+arch_snap_inode_sig_files (arch_project_tree_t * tree, t_uchar * archive, t_uchar * prev_revision, t_uchar * revision, rel_table file_list)
+{
+  inode_sig newsig = arch_tree_inode_sig (tree);
+
+  arch_snap_common_files (tree, archive, prev_revision, revision, file_list, newsig);
+}
+ 
+void
+arch_snap_inode_sig_not_reference_files (arch_project_tree_t * tree, t_uchar * archive, t_uchar * prev_revision, t_uchar * revision, rel_table file_list)
+{
+  inode_sig newsig = arch_tree_inode_sig_not_reference (tree);
+
+  arch_snap_common_files (tree, archive, prev_revision, revision, file_list, newsig);
+}
+
+void
+arch_snap_common_files (arch_project_tree_t * tree, t_uchar * archive, t_uchar * prev_revision, t_uchar * revision, rel_table file_list, inode_sig newsig)
 {
   inode_sig oldsig;
   assoc_table newsig_as;
   assoc_table newsig_paths;
-  inode_sig newsig;
   assoc_table id_list;
   
-  arch_read_inode_sig (&oldsig, tree_root, archive, prev_revision);
-  newsig = arch_tree_inode_sig (tree_root);
+  arch_read_inode_sig (&oldsig, tree->root, archive, prev_revision);
   newsig_as = rel_to_assoc (newsig.ids, 0, 1);
   newsig_paths = rel_to_assoc (newsig.paths, 0, 1);
 
-  id_list = arch_filenames_ids ( &file_list, tree_root);
+  id_list = arch_filenames_ids ( &file_list, tree);
 
   arch_update_inode_sig (oldsig.ids, newsig_as, id_list);
   arch_update_inode_sig (oldsig.paths, newsig_paths, id_list);
 
-  arch_write_inode_sig_file (oldsig, tree_root, archive, revision);
+  arch_write_inode_sig_file (oldsig, tree->root, archive, revision);
 
   arch_inode_sig_free (oldsig);
   arch_inode_sig_free (newsig);
@@ -161,51 +206,61 @@ arch_update_inode_sig (rel_table sig, assoc_table newsig, assoc_table id_list)
     {
       if (assoc_ref (id_list, sig [i] [0] ))
 	{
-	  rel_record new_record = 0;
 	  t_uchar *vals = assoc_ref (newsig, sig [i] [0]);
-	  invariant (vals != 0);
-	  new_record = rel_make_record (sig [i] [0], vals, 0);
-	  rel_free_record (sig [i]);
-	  sig [i] = new_record;
+	  if (!vals)
+	    {
+	      /* stripped by not_reference logic to avoid
+	       * race conditions
+	       */
+	    }
+	  else
+	    {
+	      rel_replace_record (sig, i, rel_make_record (sig [i] [0], vals, 0));
+	    }
 	}
     }
 }
 
 
 int
-arch_valid_inode_sig (t_uchar * tree_root, t_uchar * archive, t_uchar * revision)
+arch_valid_inode_sig (arch_project_tree_t * tree, t_uchar const * const archive, t_uchar const * const revision)
 {
   int result = 1;
-  inode_sig current_sig = arch_tree_inode_sig (tree_root);
-  
-  if (safe_access (arch_inode_sig_file(tree_root, archive, revision, INODE_IDS), F_OK))
+  inode_sig current_sig = arch_tree_inode_sig (tree);
+  t_uchar *sig_file;
+    
+  sig_file = arch_inode_sig_file(tree->root, archive, revision, INODE_IDS);
+  if (safe_access (sig_file, F_OK))
     {
       safe_printfmt (2, "Missing inode signature. Cannot verify reference tree integrity: you should remove and recreate this reference tree.\n    tree: %s\n    archive: %s\n    revision: %s\n",
-      tree_root, archive, revision);
+      tree->root, archive, revision);
     }
-   else 
+  else 
     {
       rel_table existing_sig = 0;
 
-      arch_read_inode_sig_ids (&existing_sig, NULL, tree_root, archive, revision);
+      arch_read_inode_sig_ids (&existing_sig, NULL, tree->root, archive, revision);
       result = arch_inode_sig_equal (result, existing_sig, current_sig.ids, 2);
       rel_free_table (existing_sig);
     }
+  lim_free (0, sig_file);
    
-  if (safe_access (arch_inode_sig_file(tree_root, archive, revision, INODE_PATHS), F_OK))
+  sig_file = arch_inode_sig_file(tree->root, archive, revision, INODE_PATHS);
+  if (safe_access (sig_file, F_OK))
     {
       /* don't spew warnings, most folk have perfectly ok revlibs. this is a 
        * correctness problem that using baz will just address over time.
        */
     }
-   else 
+  else 
     {
       rel_table existing_paths = 0;
 
-      arch_read_inode_sig_paths (&existing_paths, NULL, tree_root, archive, revision);
+      arch_read_inode_sig_paths (&existing_paths, NULL, tree->root, archive, revision);
       result = arch_inode_sig_equal (result, existing_paths, current_sig.paths, 2);
       rel_free_table (existing_paths);
     }
+  lim_free (0, sig_file);
 
   arch_inode_sig_free (current_sig);
   return result;
@@ -239,7 +294,7 @@ arch_read_inode_sig (inode_sig * sig, t_uchar * tree_root, t_uchar * archive, t_
 }
 
 static void arch_read_inode_sig_paths (rel_table * as_table, assoc_table * as_assoc,
-                                       t_uchar * tree_root, t_uchar * archive, t_uchar * revision)
+                                       t_uchar const * const tree_root, t_uchar const * const archive, t_uchar const * const revision)
 {
   t_uchar * sig_file = arch_inode_sig_file (tree_root, archive, revision, INODE_PATHS);
 
@@ -266,11 +321,12 @@ static void arch_read_inode_sig_paths (rel_table * as_table, assoc_table * as_as
 
       safe_close (fd);
     }
+  lim_free (0, sig_file);
 }
 
 void
 arch_read_inode_sig_ids (rel_table * as_table, assoc_table * as_assoc,
-                     t_uchar * tree_root, t_uchar * archive, t_uchar * revision)
+                     t_uchar const * const tree_root, t_uchar const * const archive, t_uchar const * const revision)
 {
   t_uchar * sig_file = arch_inode_sig_file (tree_root, archive, revision, INODE_IDS);
 
@@ -300,6 +356,7 @@ arch_read_inode_sig_ids (rel_table * as_table, assoc_table * as_assoc,
 
       safe_close (fd);
     }
+  lim_free (0, sig_file);
 }
 
 
@@ -310,7 +367,7 @@ arch_read_id_shortcut (assoc_table * ids_shortcut, t_uchar * tree_root)
   rel_table dir_listing = maybe_directory_files (inode_sig_dir);
   assoc_table answer = 0;
   assoc_table rejected = 0;
-  int x;
+  int x, y;
   time_t newest_time = 0;
   rel_table sig_table = 0;
 
@@ -334,6 +391,7 @@ arch_read_id_shortcut (assoc_table * ids_shortcut, t_uchar * tree_root)
 		{
 		  newest_time = statb.st_mtime;
                   rel_free_table (sig_table);
+		  debug (dbg_invent, 6, "reading inode sig for %s/%s in %s\n", maybe_archive, maybe_revision, tree_root);
                   arch_read_inode_sig_ids (&sig_table, 0, tree_root, maybe_archive, maybe_revision);
 		}
 	    }
@@ -343,7 +401,6 @@ arch_read_id_shortcut (assoc_table * ids_shortcut, t_uchar * tree_root)
       lim_free (0, path);
     }
               
-  int y;
   for (y = 0; y < rel_n_records (sig_table); ++y)
     {
       t_uchar * this_id = sig_table[y][0];
@@ -381,35 +438,38 @@ arch_read_id_shortcut (assoc_table * ids_shortcut, t_uchar * tree_root)
 
 
 
-
-static void
-inode_sig_callback (t_uchar * path, struct stat * stat_buf,
-                    enum arch_inventory_category category,
-                    t_uchar * id, int has_source_name,
-                    void * closure,
-                    int escape_classes)
+void
+inode_sig_callback (void * closure, invent_callback_data_t const * const data)
 {
   inode_sig * answer = (inode_sig *)closure;
-  t_uchar * signature = arch_statb_inode_sig (stat_buf);
+  t_uchar * signature = arch_statb_inode_sig ((struct stat *)&data->stat_buf);
 
-  if (!S_ISDIR (stat_buf->st_mode) && !S_ISLNK (stat_buf->st_mode))
+  if (!S_ISDIR (data->stat_buf.st_mode) && !S_ISLNK (data->stat_buf.st_mode))
     {
-      rel_add_records (&answer->ids, rel_make_record (id, signature, 0), 0);
-      rel_add_records (&answer->paths, rel_make_record (id, path, 0), 0);
+      rel_add_records (&answer->ids, rel_make_record (data->id, signature, 0), 0);
+      rel_add_records (&answer->paths, rel_make_record (data->id, data->path, 0), 0);
     }
   lim_free(0, signature);
 }
 
+void
+inode_sig_callback_not_reference (void * closure, invent_callback_data_t const * const data)
+{
+  inode_sig * answer = (inode_sig *)closure;
+  if (data->stat_buf.st_mtime < answer->start_time && data->stat_buf.st_atime < answer->start_time)
+      inode_sig_callback (closure, data);
+}
+
 
 static t_uchar *
-arch_inode_sig_dir (t_uchar * tree_root, t_uchar const * sig_type)
+arch_inode_sig_dir (t_uchar const * const tree_root, t_uchar const * sig_type)
 {
   /* FIXME: file_name_in_vicinity should be checked for const correctness */
   return file_name_in_vicinity (0, tree_root, (t_uchar *)sig_type);
 }
 
 static t_uchar *
-arch_inode_sig_file (t_uchar * tree_root, t_uchar * archive, t_uchar * revision, t_uchar const * sig_type)
+arch_inode_sig_file (t_uchar const * const tree_root, t_uchar const * const archive, t_uchar const * const revision, t_uchar const * sig_type)
 {
   t_uchar * inode_sig_dir = arch_inode_sig_dir (tree_root, sig_type);
   t_uchar * basename = str_alloc_cat_many (0, archive, "%", revision, str_end);
@@ -422,7 +482,7 @@ arch_inode_sig_file (t_uchar * tree_root, t_uchar * archive, t_uchar * revision,
 
 
 static t_uchar *
-arch_inode_sig_tmp_file (t_uchar * tree_root, t_uchar * archive, t_uchar * revision, t_uchar const * sig_type)
+arch_inode_sig_tmp_file (t_uchar const * const tree_root, t_uchar const * const archive, t_uchar const * const revision, t_uchar const * sig_type)
 {
   t_uchar * inode_sig_dir = arch_inode_sig_dir (tree_root, sig_type);
   t_uchar * basename = str_alloc_cat_many (0, ",,", archive, "%", revision, str_end);
@@ -435,7 +495,7 @@ arch_inode_sig_tmp_file (t_uchar * tree_root, t_uchar * archive, t_uchar * revis
 
 
 static int
-arch_creat_inode_sig_file (t_uchar * tree_root, t_uchar * archive, t_uchar * revision, t_uchar const * sig_type)
+arch_creat_inode_sig_file (t_uchar const * const tree_root, t_uchar const * const archive, t_uchar const * const revision, t_uchar const * sig_type)
 {
   t_uchar * inode_sig_dir = arch_inode_sig_dir (tree_root, sig_type);
   t_uchar * inode_sig_tmp_file = arch_inode_sig_tmp_file (tree_root, archive, revision, sig_type);
@@ -453,14 +513,14 @@ arch_creat_inode_sig_file (t_uchar * tree_root, t_uchar * archive, t_uchar * rev
 }
 
 static void
-arch_finish_inode_sig_file (t_uchar * tree_root, t_uchar * archive, t_uchar * revision, t_uchar const * sig_type, int fd)
+arch_finish_inode_sig_file (t_uchar const * const tree_root, t_uchar const * const archive, t_uchar const * const revision, t_uchar const * sig_type, int fd)
 {
   t_uchar * inode_sig_tmp_file = arch_inode_sig_tmp_file (tree_root, archive, revision, sig_type);
   t_uchar * inode_sig_file = arch_inode_sig_file (tree_root, archive, revision, sig_type);
 
   safe_close (fd);
   safe_rename (inode_sig_tmp_file, inode_sig_file);
-  arch_prune_inode_sig_dir (tree_root, inode_sig_file, sig_type);
+  arch_prune_inode_sig_dir (tree_root, inode_sig_file, sig_type, MAX_INODE_SIG_FILES);
 
   lim_free (0, inode_sig_tmp_file);
   lim_free (0, inode_sig_file);
@@ -468,7 +528,7 @@ arch_finish_inode_sig_file (t_uchar * tree_root, t_uchar * archive, t_uchar * re
 
 
 static void
-arch_prune_inode_sig_dir (t_uchar * tree_root, t_uchar * current_inode_sig_file, t_uchar const * sig_type)
+arch_prune_inode_sig_dir (t_uchar const * const tree_root, t_uchar const * const current_inode_sig_file, t_uchar const * sig_type, int max_sigs)
 {
   t_uchar * inode_sig_dir = arch_inode_sig_dir (tree_root, sig_type);
   rel_table dir_listing = maybe_directory_files (inode_sig_dir);
@@ -513,7 +573,7 @@ arch_prune_inode_sig_dir (t_uchar * tree_root, t_uchar * current_inode_sig_file,
           lim_free (0, path);
         }
 
-      if (n_inode_sig_files > MAX_INODE_SIG_FILES)
+      if (n_inode_sig_files > max_sigs)
         {
           t_uchar * dead_path = file_name_in_vicinity (0, inode_sig_dir, dir_listing[dead_one][0]);
 
@@ -521,10 +581,12 @@ arch_prune_inode_sig_dir (t_uchar * tree_root, t_uchar * current_inode_sig_file,
           --n_inode_sig_files;
 
           lim_free (0, dead_path);
+	  rel_remove_records (&dir_listing, dead_one, dead_one);
+	  dead_one = -1;
         }
 
 
-    } while (n_inode_sig_files > MAX_INODE_SIG_FILES); /* usually only one iteration expected */
+    } while (n_inode_sig_files > max_sigs); /* usually only one iteration expected */
 
   lim_free (0, inode_sig_dir);
   rel_free_table (dir_listing);
@@ -599,7 +661,19 @@ prune_old_dev_cruft (rel_table the_table)
     }
 }
 
-
+/**
+ * \brief remove all inode sigs in tree_root
+ *
+ * this is extremely useful if dealing with full copy trees that will have inappropriate
+ * inode sigs stored in them
+ * \param tree_root the tree root.
+ */
+void
+arch_clear_inode_sigs (t_uchar const * const tree_root)
+{
+  arch_prune_inode_sig_dir (tree_root, NULL, INODE_IDS, 0);
+  arch_prune_inode_sig_dir (tree_root, NULL, INODE_PATHS, 0);
+}
 
 
 
