@@ -9,12 +9,14 @@
 
 
 #include "hackerlab/bugs/panic.h"
+#include "hackerlab/bugs/exception.h"
 #include "hackerlab/os/errno.h"
 #include "hackerlab/os/errno-to-string.h"
 #include "hackerlab/os/time.h"
 #include "hackerlab/os/sys/types.h"
 #include "hackerlab/os/unistd.h"
 #include "hackerlab/mem/mem.h"
+#include "hackerlab/mem/talloc.h"
 #include "hackerlab/char/char-class.h"
 #include "hackerlab/char/str.h"
 #include "hackerlab/fmt/cvt.h"
@@ -30,41 +32,29 @@
 #include "libarch/inv-ids.h"
 #include "libarch/pfs.h"
 #include "hackerlab/fs/cwd.h"
+#include "po/gettext.h"
 #include "libawk/relational.h"
 #include "libarch/changelogs.h"
+#include "libarch/debug.h"
 
 
 
-enum ftag_method
-{
-  ftag_names,
-  ftag_implicit,
-  ftag_tagline,
-  ftag_explicit
-};
-
 
 /* __STDC__ prototypes for static functions */
-static t_uchar * file_id (int * errn,
-                          struct alloc_limits * limits,
-                          enum ftag_method method,
-                          int untagged_is_source,
-                          t_uchar * path,
-                          assoc_table id_tagging_shortcut,
-                          struct stat * known_lstat,
-                          assoc_table * explicit_skips);
 static int is_at_or_underneath_archdir (char * rel_file);
 static int filename_matches (regex_t * pattern, char * filename);
 static t_uchar * explicit_id (int * errn,
                               assoc_table * skips,
                               struct alloc_limits * limits,
-                              t_uchar * arg_file,
+                              t_uchar const * const arg_file,
                               t_uchar * id_file,
                               t_uchar * prefix,
-                              t_uchar * postfix);
+                              t_uchar * postfix,
+			      struct stat * statb,
+			      assoc_table id_tagging_shortcut);
 static t_uchar * implicit_id (int * errn,
                               struct alloc_limits * limits,
-                              t_uchar * file,
+                              t_uchar const * const file,
                               t_uchar * basename,
                               t_uchar * prefix,
                               struct stat * statb,
@@ -88,7 +78,7 @@ arch_log_file_id (t_uchar * archive, t_uchar * revision)
 
 
 assoc_table
-arch_filenames_ids (rel_table * file_list, t_uchar * tree_root)
+arch_filenames_ids (rel_table * file_list, arch_project_tree_t * tree)
 {
   struct arch_inventory_options options = {0, };
   int i = 0;
@@ -98,11 +88,18 @@ arch_filenames_ids (rel_table * file_list, t_uchar * tree_root)
   options.categories = arch_inventory_source;
   options.want_ids = 1;
   options.include_excluded = 1;
-  arch_get_inventory_naming_conventions (&options, tree_root);
+  arch_get_inventory_naming_conventions (&options, tree);
+  if (tree->untagged_is_source)
+    {
+      /* FIXME if this is annoying remove it. but its probably a bug to do this at all
+       */
+      safe_printfmt (2, "overriding untagged is source for arch_filenames_ids\n");
+      tree->untagged_is_source = 0;
+    }
 
   for (i = 0; i != file_size; ++ i)
     {
-      t_uchar * id = arch_inventory_id (options.method, 0, (* file_list) [i] [0], 0, 0, 0);
+      t_uchar * id = arch_inventory_id (tree, (* file_list) [i] [0], 0);
       assoc_set (&id_list, id, (* file_list) [i] [0]);
       lim_free (0, id);
     }
@@ -112,82 +109,34 @@ arch_filenames_ids (rel_table * file_list, t_uchar * tree_root)
 }
 
 
+/**
+ * \brief retrieve a file id for an arbitrary file on disk
+ * \param tree the tree to get the id for. 
+ * \param method override the inventory logic, use arch_unspecified_id_tagging to autodetect.
+ * \param untagged_is_source give source file magic ids
+ * \param path the filename to inventory
+ * \param id_tagging_shortcut cached id tags
+ * \param known_lstat prior stat result for this file
+ * \param explicit_skips files to not id
+ */
 t_uchar *
-arch_inventory_id (enum arch_id_tagging_method method,
-                    int untagged_is_source,
-                    t_uchar * path,
-                    assoc_table id_tagging_shortcut,
-                    struct stat * known_lstat,
-                    assoc_table * explicit_skips)
+arch_inventory_id (arch_project_tree_t * tree,
+                    t_uchar const * const path,
+                    struct stat * known_lstat)
 {
-  int errn;
   t_uchar * answer;
-  enum ftag_method m;
+  inventory_entry_t *entry = arch_project_tree_path_inventory (tree, path, known_lstat);
 
-  if (method == arch_unspecified_id_tagging)
-    {
-      t_uchar * dir = 0;
-      t_uchar * root = 0;
+  if (entry)
+      answer = str_save (0, entry->id);
+  else
+      answer = NULL;
 
-      dir = file_name_directory_file (0, path);
-      root = arch_tree_root (0, dir, 0);
-
-      if (!root)
-        {
-          method = arch_names_id_tagging;
-          untagged_is_source = 1;
-        }
-      else
-        {
-          enum arch_inventory_category untagged_category;
-
-          method = arch_tree_id_tagging_method (&untagged_category, root, 0);
-          untagged_is_source = (untagged_category == arch_inventory_source);
-        }
-
-      lim_free (0, dir);
-      lim_free (0, root);
-    }
-
-  switch (method)
-    {
-    default:
-      panic ("unrecognized method in arch_inventory_id");
-      break;
-
-    case arch_names_id_tagging:
-      {
-        m = ftag_names;
-        break;
-      }
-    case arch_explicit_id_tagging:
-      {
-        m = ftag_explicit;
-        break;
-      }
-    case arch_implicit_id_tagging:
-      {
-        m = ftag_implicit;
-        break;
-      }
-    case arch_tagline_id_tagging:
-      {
-        m = ftag_tagline;
-        break;
-      }
-    }
-
-
-
-  errn = 0;
-
-  answer = file_id (&errn, 0, m, untagged_is_source, path, id_tagging_shortcut, known_lstat, explicit_skips);
-
-  if (!answer && errn)
-    {
-      safe_printfmt (2, "error finding file id (%d: %s)\n path: %s\n", errn, errno_to_string(errn), path);
-      panic ("arch_inventory_id");
-    }
+  /* the tree gives us a reference ... for transition purposes.
+   * FIXME RBC 20050331 the tree should just give us a pointer, not a reference locked
+   * object, and cache ids sanely. not possible for hackerlab trees yet.
+   */
+  talloc_free (entry);
 
   return answer;
 }
@@ -434,41 +383,33 @@ arch_default_id_tagging_method_contents (enum arch_id_tagging_method method)
   return answer;
 }
 
-
-
 t_uchar *
-arch_tree_id_tagging_method_file (t_uchar * tree_root)
+arch_tree_id_tagging_method_file (arch_project_tree_t * tree)
 {
   t_uchar * ctl_dir;
   t_uchar * answer;
 
-  ctl_dir = arch_tree_ctl_dir (tree_root);
+  ctl_dir = arch_tree_ctl_dir (tree->root);
   answer = file_name_in_vicinity (0, ctl_dir, "=tagging-method");
   lim_free (0, ctl_dir);
   return answer;
 }
 
-
+/**
+ * \brief determing the id tagging method for a tree
+ *
+ * this is the only function allowed to call vtable->id_tagging_method directly,
+ * all other functions should call this wrapper
+ */
 enum arch_id_tagging_method
-arch_tree_id_tagging_method (enum arch_inventory_category * cat_var, t_uchar * tree_root, int strict)
+arch_tree_id_tagging_method (arch_project_tree_t * tree, enum arch_inventory_category * cat_var, int strict)
 {
-  struct arch_inventory_options options;
-  enum arch_id_tagging_method answer;
-
-  mem_set0 ((t_uchar *)&options, sizeof (options));
-  arch_get_inventory_naming_conventions (&options, tree_root);
-  if (cat_var)
-    {
-      *cat_var = options.untagged_source_category;
-    }
-  answer = options.method;
-  arch_free_inventory_naming_conventions (&options);
-  return answer;
+  return tree->vtable->id_tagging_method (tree, cat_var, strict);
 }
 
 
 void
-arch_set_tree_id_tagging_method (t_uchar * tree_root,
+arch_set_tree_id_tagging_method (arch_project_tree_t * tree,
                               enum arch_id_tagging_method method)
 {
   int errn;
@@ -479,7 +420,7 @@ arch_set_tree_id_tagging_method (t_uchar * tree_root,
   int out_fd;
 
   method_name = arch_id_tagging_method_name (method);
-  method_file = arch_tree_id_tagging_method_file (tree_root);
+  method_file = arch_tree_id_tagging_method_file (tree);
   method_dir = file_name_directory_file (0, method_file);
   method_tmp = file_name_in_vicinity (0, method_dir, ",,tagging-method");
 
@@ -670,12 +611,27 @@ arch_generate_id (void)
 
 /* precondition: path must be accessible */
 t_uchar *
-arch_choose_explicit_id (t_uchar * path, t_uchar * id)
+arch_choose_explicit_id (arch_project_tree_t * tree, t_uchar * path, t_uchar * id)
 {
   t_uchar * tagline_id;
   t_uchar * new_explicit_id;
+  int old_untagged_source;
+  enum arch_id_tagging_method old_tagging;
   
-  tagline_id = arch_inventory_id (arch_tagline_id_tagging, 0, path, 0, 0, 0);
+  /* this sets untagged_is_source 0 so that we don't get names ids
+   * however, this should not be needed - we should ask for the implicit
+   * id directly
+   * RBC 20050321
+   */
+  old_untagged_source = tree->untagged_is_source;
+  tree->untagged_is_source = 0;
+  /* FIXME RBC 20050321 overriding the tagging method is not necesarily possible
+   * fix this leaky abstraction (perhaps 'has_real_id ?') */
+  old_tagging = tree->tag_method;
+  tree->tag_method = arch_tagline_id_tagging;
+  tagline_id = arch_inventory_id (tree, path, 0);
+  tree->untagged_is_source = old_untagged_source;
+  tree->tag_method = old_tagging;
 
   if (!tagline_id || !arch_id_indicates_changelog (tagline_id))
     if (!id)
@@ -715,7 +671,7 @@ t_uchar *
 arch_add_id (t_uchar * path, t_uchar * id)
 {
   t_uchar * new_explicit_id;
-  t_uchar * tree_root;
+  arch_project_tree_t * tree;
   t_uchar * rel_path;
   t_uchar * full_path;
   
@@ -727,20 +683,24 @@ arch_add_id (t_uchar * path, t_uchar * id)
     }
 
   /* don't add ids for control paths */
-  tree_root = arch_tree_root(0, path, 0);
+  tree = arch_project_tree_new (talloc_context, path);
   full_path = arch_abs_path (path);
-  rel_path = full_path + str_length (tree_root) + 1;
-  lim_free (0, tree_root);
+  rel_path = arch_project_tree_rel_path_from_abs (tree, full_path);
   if (arch_is_control_file (rel_path, NULL) && str_cmp_suffix_n(path, ".arch-inventory", 15))
     {
       lim_free (0, full_path);
+      lim_free (0, rel_path);
+      arch_project_tree_delete (tree);
       return NULL;
     }
   lim_free (0, full_path);
 
-  new_explicit_id = arch_choose_explicit_id (path, id);
+  debug (dbg_invent, 3, _("getting id for (%s)\n"), rel_path);
+  new_explicit_id = arch_choose_explicit_id (tree, rel_path, id); /* FIXME path here is full not rel */
   arch_add_explicit_id (path, new_explicit_id);
 
+  arch_project_tree_delete (tree);
+  lim_free (0, rel_path);
   return new_explicit_id;
 }
 
@@ -989,37 +949,40 @@ arch_set_dont_care_explicit_default (t_uchar * dir)
 
 
 
-/*(c file_id)
- * static t_uchar * file_id (int * errn,
- *                           struct alloc_limits * limits,
- *                           enum ftag_method method,
- *                           t_uchar * path);
- *
- * Return a newly allocated string containing the inventory id
- * of the file `path' using inventory method `method'.
+/**
+ * \brief return a newly allocated string with the inventory id of the file / directory.
  *
+ * WARNING. DO NOT CALL THIS FUNCTION. It is private to tla-project-tree.c, but not
+ * relocated until more of the object model clears up.
  * Return 0 and set `*errn' if the id can not be computed.
  * If no I/O error occurs, the file does not have an explicit
  * id, but `method' is `ftag_explicit', `*errn' is set to 0.
+ * \param errn put error details here
+ * \param tree the tree to get the id for
+ * \param limits unused ?
+ * \param method override the tagging behaviour of the tree
+ * \param untagged_is_source override the untagged-source bahviour of the tree
+ * \param path the relative to cwd, or abs path to the file.
+ * \param id_tagging_shortcut cached ids
+ * \param known_lstat cached stat result
+ * \param explicit_skips files to skip
  */
-static t_uchar *
-file_id (int * errn,
-         struct alloc_limits * limits,
-         enum ftag_method method,
-         int untagged_is_source,
-         t_uchar * path,
-         assoc_table id_tagging_shortcut,
-         struct stat * known_lstat,
-         assoc_table * explicit_skips)
+inventory_entry_t *
+tla_path_id (arch_project_tree_t * tree,
+	     int * errn,
+	     t_uchar const * const path,
+	     struct stat * known_lstat)
 {
-  t_uchar * answer = 0;
+  t_uchar * answer_id = 0;
   t_uchar * as_file = 0;
   t_uchar * basename = 0;
   t_uchar * dir = 0;
   t_uchar * dir_as_file = 0;
   t_uchar * dir_basename = 0;
   t_uchar * id_file = 0;
-  struct stat stat_buf;
+  t_uchar * fq_path = NULL;
+  struct stat local_stat_buf;
+  struct stat *stat_buf = NULL;
   int is_dir;
   int is_symlink;
 
@@ -1027,48 +990,81 @@ file_id (int * errn,
     {
       *errn = EINVAL;
 
-    return_answer:
+    return_answer_id:
 
-      lim_free (limits, as_file);
-      lim_free (limits, basename);
-      lim_free (limits, dir);
-      lim_free (limits, dir_as_file);
-      lim_free (limits, dir_basename);
-      lim_free (limits, id_file);
+      lim_free (0, as_file);
+      lim_free (0, dir);
+      lim_free (0, dir_as_file);
+      lim_free (0, dir_basename);
+      lim_free (0, id_file);
+	{
+	  /* FIXME RBC 20050331 typed returns - dir link etc */
+	  inventory_entry_t *answer = NULL;
+	  /* FIXME RBC 20050331 make this cleaner and only-called once */
+	  if (answer_id)
+	    {
+	      if (!stat_buf)
+		{
+		  if (known_lstat)
+		      stat_buf = known_lstat;
+		  else
+		    {
+		      stat_buf = &local_stat_buf;
+		      
+		      if (vu_lstat (errn, fq_path, stat_buf))
+			{
+			  lim_free (0, basename);
+			  lim_free (0, fq_path);
+			  return NULL;
+			}
+		    }
+		}
 
-      return answer;
+	      if (S_ISDIR (stat_buf->st_mode))
+		  answer = inventory_dir_new (NULL, basename, answer_id);
+	      else if (S_ISLNK (stat_buf->st_mode))
+		  answer = inventory_link_new (NULL, basename, answer_id);
+	      else
+		  answer = inventory_file_new (NULL, basename, answer_id);
+	    }
+	  lim_free (0, basename);
+	  lim_free (0, fq_path);
+	  return answer;
+	}
     }
 
-  as_file = file_name_from_directory (limits, path);
+  as_file = file_name_from_directory (0, (t_uchar *)path);
   if (!as_file)
     {
     enomem_error:
       *errn = ENOMEM;
-      goto return_answer;
+      goto return_answer_id;
     }
 
-  if (method == ftag_names)
+  fq_path = file_name_in_vicinity (0, tree->root, as_file);
+  basename = file_name_tail (0, as_file);
+
+  if (tree->tag_method == arch_names_id_tagging)
     {
-      answer = str_alloc_cat (limits, "?", as_file);
-      if (!answer)
+      answer_id = str_alloc_cat (0, "?", as_file);
+      if (!answer_id)
         goto enomem_error;
       else
-        goto return_answer;
+        goto return_answer_id;
     }
 
 
-  basename = file_name_tail (limits, as_file);
-  dir = file_name_directory (limits, as_file);
+  dir = file_name_directory (0, as_file);
   if (!dir)
-    dir = str_save (limits, ".");
+    dir = str_save (0, ".");
   if (!(basename && dir))
     goto enomem_error;
 
-  dir_as_file = file_name_from_directory (limits, dir);
+  dir_as_file = file_name_from_directory (0, dir);
   if (!dir_as_file)
     goto enomem_error;
 
-  dir_basename = file_name_tail (limits, dir_as_file);
+  dir_basename = file_name_tail (0, dir_as_file);
   if (!dir_basename)
     goto enomem_error;
 
@@ -1077,8 +1073,8 @@ file_id (int * errn,
    */
   if (!str_cmp (dir_basename, ".arch-ids"))
     {
-      answer = explicit_id (errn, 0, limits, path, as_file, "E_", 0);
-      goto return_answer;
+      answer_id = explicit_id (errn, 0, 0, fq_path, fq_path, "E_", 0, known_lstat, tree->id_tagging_shortcut);
+      goto return_answer_id;
     }
 
   /* Explicit id file directories:
@@ -1086,12 +1082,12 @@ file_id (int * errn,
   if (!str_cmp (basename, ".arch-ids"))
     {
       long amt;
-      answer = str_alloc_cat (limits, "D_", as_file);
-      if (!answer)
+      answer_id = str_alloc_cat (0, "D_", as_file);
+      if (!answer_id)
         goto enomem_error;
-      amt = smash_non_graphical (answer, str_length (answer));
-      answer[amt] = 0;
-      goto return_answer;
+      amt = smash_non_graphical (answer_id, str_length (answer_id));
+      answer_id[amt] = 0;
+      goto return_answer_id;
     }
 
   /* Paths beginning with "./{arch}" are tagged with their own
@@ -1102,145 +1098,151 @@ file_id (int * errn,
   if (is_at_or_underneath_archdir (as_file))
     {
       long amt;
-      answer = str_alloc_cat (limits, "A_", as_file);
-      if (!answer)
+      answer_id = str_alloc_cat (0, "A_", as_file);
+      if (!answer_id)
         goto enomem_error;
-      amt = smash_non_graphical (answer, str_length (answer));
-      answer[amt] = 0;
-      goto return_answer;
+      amt = smash_non_graphical (answer_id, str_length (answer_id));
+      answer_id[amt] = 0;
+      goto return_answer_id;
     }
 
 
   /* Try for an explicit id:
    */
   if (known_lstat)
-    stat_buf = *known_lstat;
-  else if (0 > vu_lstat (errn, as_file, &stat_buf))
-    goto return_answer;
+    stat_buf = known_lstat;
+  else if (0 > vu_lstat (errn, fq_path, &local_stat_buf))
+    goto return_answer_id;
+  else
+      stat_buf = &local_stat_buf;
 
-  if (S_ISDIR (stat_buf.st_mode))
+  if (S_ISDIR (stat_buf->st_mode))
     {
       is_dir = 1;
       is_symlink = 0;
-      id_file = file_name_in_vicinity (limits, as_file, ".arch-ids/=id");
+      id_file = file_name_in_vicinity (0, fq_path, ".arch-ids/=id");
       if (!id_file)
         goto enomem_error;
     }
   else
     {
       is_dir = 0;
-      is_symlink = S_ISLNK (stat_buf.st_mode);
+      is_symlink = S_ISLNK (stat_buf->st_mode);
 
-      id_file = file_name_in_vicinity (limits, dir, ".arch-ids/");
+      id_file = file_name_in_vicinity (0, dir, ".arch-ids/");
       if (!id_file)
         goto enomem_error;
-      id_file = str_realloc_cat (limits, id_file, basename);
+      id_file = str_realloc_cat (0, id_file, basename);
       if (!id_file)
         goto enomem_error;
-      id_file = str_realloc_cat (limits, id_file, ".id");
+      id_file = str_realloc_cat (0, id_file, ".id");
+      if (!id_file)
+        goto enomem_error;
+      id_file = str_replace (id_file, file_name_in_vicinity (0, tree->root, id_file));
       if (!id_file)
         goto enomem_error;
     }
 
   *errn = 0;
-  answer = explicit_id (errn, 0, limits, path, id_file, "x_", 0);
-  if (answer || (*errn != ENOENT))
-    goto return_answer;
+  answer_id = explicit_id (errn, 0, 0, path, id_file, "x_", 0, stat_buf, tree->id_tagging_shortcut);
+  if (answer_id || (*errn != ENOENT))
+    goto return_answer_id;
   else
     {
       /* Is there a .arch-ids/=all file here?
        */
-      lim_free (limits, id_file);
+      lim_free (0, id_file);
       if (is_dir)
-        id_file = file_name_in_vicinity (limits, as_file, ".arch-ids/=all");
+        id_file = file_name_in_vicinity (0, fq_path, ".arch-ids/=all");
       else
-        id_file = file_name_in_vicinity (limits, dir, ".arch-ids/=all");
+        id_file = file_name_in_vicinity (0, dir, ".arch-ids/=all");
 
       if (!id_file)
         goto enomem_error;
 
       *errn = 0;
-      answer = explicit_id (errn, explicit_skips, limits, path, id_file, "a_", (is_dir ? (t_uchar *)"./." : basename));
-      if (answer || (*errn != ENOENT))
-        goto return_answer;
+      answer_id = explicit_id (errn, &tree->explicit_skips, 0, fq_path, id_file, "a_", (is_dir ? (t_uchar *)"./." : basename), stat_buf, NULL);
+      if (answer_id || (*errn != ENOENT))
+        goto return_answer_id;
 
-      if ((method == ftag_implicit) && !is_dir && !is_symlink)
+      if ((tree->tag_method == arch_implicit_id_tagging) && !is_dir && !is_symlink)
         {
           *errn = 0;
-          answer = implicit_id (errn, limits, path, basename, "i_", &stat_buf, id_tagging_shortcut);
-          if (answer || *errn)
-            goto return_answer;
+          answer_id = implicit_id (errn, 0, fq_path, basename, "i_", stat_buf, tree->id_tagging_shortcut);
+          if (answer_id || *errn)
+            goto return_answer_id;
         }
 
-      if ((method == ftag_tagline) && !is_dir && !is_symlink)
+      if ((tree->tag_method == arch_tagline_id_tagging) && !is_dir && !is_symlink)
         {
           *errn = 0;
-          answer = implicit_id (errn, limits, path, 0, "i_", &stat_buf, id_tagging_shortcut);
-          if (answer || *errn)
-            goto return_answer;
+          answer_id = implicit_id (errn, 0, fq_path, 0, "i_", stat_buf, tree->id_tagging_shortcut);
+          if (answer_id || *errn)
+            goto return_answer_id;
         }
 
 
       /* is there an "=default" id?
        */
-      lim_free (limits, id_file);
+      lim_free (0, id_file);
       if (is_dir)
-        id_file = file_name_in_vicinity (limits, as_file, ".arch-ids/=default");
+        id_file = file_name_in_vicinity (0, fq_path, ".arch-ids/=default");
       else
-        id_file = file_name_in_vicinity (limits, dir, ".arch-ids/=default");
+        id_file = file_name_in_vicinity (0, dir, ".arch-ids/=default");
 
       if (!id_file)
         goto enomem_error;
 
       *errn = 0;
-      answer = explicit_id (errn, explicit_skips, limits, path, id_file, "w_", (is_dir ?  (t_uchar *)"./." : basename));
-      if (answer || (*errn != ENOENT))
-        goto return_answer;
+      answer_id = explicit_id (errn, &tree->explicit_skips, 0, fq_path, id_file, "w_", (is_dir ?  (t_uchar *)"./." : basename), stat_buf, NULL);
+      if (answer_id || (*errn != ENOENT))
+        goto return_answer_id;
 
       /* no explicit, =all, implicit, tagline, or =default id.
        */
-      if (untagged_is_source)
+      if (tree->untagged_is_source)
         {
           int skipped;
+	  struct stat dont_care_stat_buf;
 
-          lim_free (limits, id_file);
-          id_file = file_name_in_vicinity (limits, dir, ".arch-ids/=dont-care");
+          lim_free (0, id_file);
+          id_file = file_name_in_vicinity (0, dir, ".arch-ids/=dont-care");
           if (!id_file)
             goto enomem_error;
 
-          skipped = (explicit_skips && !!assoc_ref (*explicit_skips, id_file));
+          skipped = (!!assoc_ref (tree->explicit_skips, id_file));
 
-          if (!skipped && (0 <= vu_lstat (errn, id_file, &stat_buf)))
+          if (!skipped && (0 <= vu_lstat (errn, id_file, &dont_care_stat_buf)))
             {
               long amt;
-              answer = str_alloc_cat (limits, "k_", as_file);
-              if (!answer)
+              answer_id = str_alloc_cat (0, "k_", as_file);
+              if (!answer_id)
                 goto enomem_error;
-              amt = smash_non_graphical (answer, str_length (answer));
-              answer[amt] = 0;
-              goto return_answer;
+              amt = smash_non_graphical (answer_id, str_length (answer_id));
+              answer_id[amt] = 0;
+              goto return_answer_id;
             }
 
-          if (!skipped && explicit_skips)
-            assoc_set (explicit_skips, id_file, "yes");
+          if (!skipped)
+            assoc_set (&tree->explicit_skips, id_file, "yes");
 
           if (*errn == ENOENT)
             {
               long amt;
-              answer = str_alloc_cat (limits, "?_", as_file);
-              if (!answer)
+              answer_id = str_alloc_cat (0, "?_", as_file);
+              if (!answer_id)
                 goto enomem_error;
-              amt = smash_non_graphical (answer, str_length (answer));
-              answer[amt] = 0;
-              goto return_answer;
+              amt = smash_non_graphical (answer_id, str_length (answer_id));
+              answer_id[amt] = 0;
+              goto return_answer_id;
             }
           else
-            goto return_answer;
+            goto return_answer_id;
         }
       else
         {
           *errn = 0;
-          goto return_answer;
+          goto return_answer_id;
         }
     }
 
@@ -1251,7 +1253,7 @@ file_id (int * errn,
 t_uchar *
 arch_id_from_explicit_file (int *errn, t_uchar * path)
 {
-  return explicit_id (errn, 0, 0, 0, path, "x_", 0);
+  return explicit_id (errn, 0, 0, 0, path, "x_", 0, NULL, NULL);
 }
 
 static int
@@ -1264,7 +1266,7 @@ is_at_or_underneath_archdir (char * rel_file)
     {
       int re_error;
 
-      re_error = regcomp (&pattern, "^(.*/)?(\\{arch\\}(/[a-zA-Z0-9=][^/~]*)*|\\{arch\\}/\\.arch-project-tree)$", REG_EXTENDED);
+      re_error = regcomp (&pattern, "^(.*/)?(\\{arch\\}(/[a-zA-Z=][^/]*)*|\\{arch\\}/\\.arch-project-tree)$", REG_EXTENDED);
       invariant (!re_error);
       compiled = 1;
     }
@@ -1290,14 +1292,16 @@ filename_matches (regex_t * pattern, char * filename)
 }
 
 
-static t_uchar *
+t_uchar *
 explicit_id (int * errn,
               assoc_table * skips,
               struct alloc_limits * limits,
-              t_uchar * arg_file,
+              t_uchar const * const arg_file,
               t_uchar * id_file,
               t_uchar * prefix,
-              t_uchar * postfix)
+              t_uchar * postfix,
+	      struct stat * statb,
+	      assoc_table id_tagging_shortcut)
 {
   int id_fd;
   t_uchar * answer;
@@ -1312,13 +1316,13 @@ explicit_id (int * errn,
       return 0;
     }
 
-#if DISABLED_NEW_CODE
   if (id_tagging_shortcut)
     {
       struct stat stat_buf;
       /* prefix determines action..
        * E_ - it's a .arch-ids file we are examinig
        * x_ - it's a normal file that /may/ have an id file
+       * w_ a_ unsupported
        */
       /* id_file is always the (prospective) id file */
       /* TODO?: Cache the inode sig of the id file */
@@ -1327,38 +1331,49 @@ explicit_id (int * errn,
       
       if (statb)
 	stat_buf = *statb;
-      else if (0 > vu_lstat (errn, arg_file, &stat_buf))
+      else if (0 > vu_lstat (errn, (t_uchar *)arg_file, &stat_buf))
 	return 0;
       
       signature = arch_statb_inode_sig (&stat_buf);
       cached_id = assoc_ref (id_tagging_shortcut, signature);
+      if (!cached_id)
+	  debug (dbg_invent, 8, "no cached id for signature '%s'\n", signature);
       lim_free (0, signature);
 
       if (cached_id) 
+	/* cache hit */
 	{
 	  if (!str_cmp (arg_file, id_file))
+	    /* id of the .id file itself */
 	    {
 	      if ((cached_id[0] == 'E') && (cached_id[1] == '_'))
-		return str_save (0, cached_id);
+		{
+		  debug (dbg_invent, 8, "Cached signature : '%s' on '%s'\n", cached_id, id_file);
+		  return str_save (0, cached_id);
+		}
+	      /* unexpected cached result, ignore */
 	    }
 	  else if ((cached_id[0] == 'x') && (cached_id[1] == '_'))
 	    {
 	      /* TODO: we could save 1 stat per .id file if we cached the results of the lookups somewhere - 
 	       * .arch-ids is read before the files in the dir.
 	       */
+	      /* this call here should get a cache-hit on the id shortcut file */
 	      t_uchar *id_tag = explicit_id (errn, skips, limits, id_file, id_file, "E_", 0, 0, id_tagging_shortcut);
 	      if (id_tag && (!str_cmp (id_tag + 1, cached_id + 1)))
 		{
 		  lim_free (0, id_tag);
+		  debug (dbg_invent, 8, "Cached signature on : '%s'\n", cached_id);
 		  return str_save (0, cached_id);
 		}
 	      else
 		lim_free (0, id_tag);
 	    }
+	  debug (dbg_invent, 8, "Cached signature '%s' not usable\n", cached_id);
 	}
     }
-#endif
 
+  debug (dbg_invent, 8, "Reading '%s' for explicit id on '%s'\n", id_file, arg_file);
   id_fd = vu_open (errn, id_file, O_RDONLY, 0);
 
   if (id_fd < 0)
@@ -1417,10 +1432,12 @@ explicit_id (int * errn,
         }
     }
 
-  answer = str_realloc_cat (limits, answer, (postfix ? postfix : (t_uchar *)""));
+  if (postfix)
+      answer = str_realloc_cat (limits, answer, postfix);
   if (!answer)
     goto enomem_error;
   vu_close (&ign, id_fd);
+  debug (dbg_invent, 8, "Created signature : '%s' on '%s' (from '%s')\n", answer, arg_file, id_file);
   return answer;
 }
 
@@ -1428,7 +1445,7 @@ explicit_id (int * errn,
 static t_uchar *
 implicit_id (int * errn,
               struct alloc_limits * limits,
-              t_uchar * file,
+              t_uchar const * const file,
               t_uchar * basename,
               t_uchar * prefix,
               struct stat * statb,
@@ -1436,7 +1453,7 @@ implicit_id (int * errn,
 {
   int file_fd;
   struct stat file_stat_buf;
-  char buf[1025];
+  char buf[1026]; /* one byte slack for \0 termination */
   int amt;
   int line;
   int bottom;
@@ -1480,7 +1497,7 @@ implicit_id (int * errn,
    * after the dash, skip any blanks -- the rest is the id.
    */
 
-  file_fd = vu_open (errn, file, O_RDONLY, 0);
+  file_fd = vu_open (errn, (t_uchar *)file, O_RDONLY, 0);
   if (file_fd < 0)
     return 0;
   
@@ -1498,7 +1515,7 @@ implicit_id (int * errn,
               vu_close (&ign, file_fd);
               return 0;
             }
-          amt = vu_read_retry (errn, file_fd, buf, sizeof (buf) - 1);
+          amt = vu_read_retry (errn, file_fd, buf, sizeof (buf) - 2);
           if (amt < 0)
             goto error_return;
         }
@@ -1506,8 +1523,8 @@ implicit_id (int * errn,
         {
           char * x;
 
-          if (file_stat_buf.st_size > sizeof (buf))
-            amt = sizeof (buf);
+          if (file_stat_buf.st_size > sizeof (buf) - 1)
+            amt = sizeof (buf) - 1;
           else
             continue;
           /* Yes, this is a off by one error. However changing it
@@ -1515,7 +1532,7 @@ implicit_id (int * errn,
            */
           if (0 > vu_lseek (errn, file_fd, -1026, SEEK_END))
             goto error_return;
-          amt = vu_read_retry (errn, file_fd, buf, sizeof (buf));
+          amt = vu_read_retry (errn, file_fd, buf, sizeof (buf) - 1);
           if (amt < 0)
             goto error_return;
           x = str_chr_index_n (buf, amt, '\n');
@@ -1596,7 +1613,7 @@ implicit_id (int * errn,
                 {
                   eol = buf + amt;
                   if (!bottom)
-                    if (file_stat_buf.st_size > sizeof (buf))
+                    if (file_stat_buf.st_size > sizeof (buf) - 1)
                       safe_printfmt (2, "Warning: top-of-file truncated tag in: %s\n", file);
                 }
 
@@ -1653,6 +1670,47 @@ smash_non_graphical (t_uchar * buf, long amt)
   return amt;
 }
 
+/**
+ * \brief convenience function for arbitrary id retrieval.
+ *
+ * please us arch_inventory_id if you have a tree object
+ * \param some_path the filepath relative or absolute
+ * \param method set to other than arch_unspecified_id_tagging to override for trees that support overriding.
+ */
+t_uchar *
+arch_id_for_path(t_uchar const * const some_path, enum arch_id_tagging_method method)
+{
+    arch_project_tree_t * tree;
+    t_uchar * dir;
+    t_uchar * answer = NULL;
+    t_uchar * rel_path;
+    t_uchar * abs_path = arch_abs_path (some_path);
+    
+    if (safe_file_is_directory (some_path))
+	dir = str_save (0, some_path);
+    else
+      {
+	dir = file_name_directory_file (0, abs_path);
+      }
+
+    tree = arch_project_tree_new_ext (talloc_context, dir, 1, 0);
+
+    /* load the inventory rules */
+    arch_tree_id_tagging_method (tree, NULL, 0);
+    /* override them if needed */
+    if (method != arch_unspecified_id_tagging)
+	tree->tag_method = method;
+
+    rel_path = arch_project_tree_rel_path_from_abs (tree, abs_path);
+    answer = arch_inventory_id (tree, rel_path, 0);
+
+    arch_project_tree_delete (tree);
+    lim_free (0, dir);
+    lim_free (0, abs_path);
+    lim_free (0, rel_path);
+    return answer;
+}
+
 
 /* tag: Tom Lord Wed May 14 07:20:26 2003 (inv-tags.c)
  */
