@@ -8,6 +8,7 @@
  */
 
 
+#include "hackerlab/bugs/exception.h"
 #include "hackerlab/bugs/panic.h"
 #include "hackerlab/char/str.h"
 #include "hackerlab/fs/file-names.h"
@@ -31,7 +32,7 @@
 
 rel_table
 arch_library_archive_dirs (rel_table opt_lib_path,
-                           t_uchar * archive,
+                           t_uchar const * const archive,
                            int for_add)
 {
   rel_table lib_path = 0;
@@ -86,8 +87,8 @@ arch_library_archive_dirs (rel_table opt_lib_path,
 
 rel_table
 arch_library_category_dirs (rel_table opt_lib_path,
-                            t_uchar * archive,
-                            t_uchar * category,
+                            t_uchar const * const archive,
+                            t_uchar const * const category,
                             int for_add)
 {
   rel_table answer = 0;
@@ -136,8 +137,8 @@ arch_library_category_dirs (rel_table opt_lib_path,
 
 rel_table
 arch_library_branch_dirs (rel_table opt_lib_path,
-                          t_uchar * archive,
-                          t_uchar * branch,
+                          t_uchar const * const archive,
+                          t_uchar const * const branch,
                           int for_add)
 {
   rel_table answer = 0;
@@ -190,8 +191,8 @@ arch_library_branch_dirs (rel_table opt_lib_path,
 
 rel_table
 arch_library_version_dirs (rel_table opt_lib_path,
-                           t_uchar * archive,
-                           t_uchar * version,
+                           t_uchar const * const archive,
+                           t_uchar const * const version,
                            int for_add)
 {
   rel_table answer = 0;
@@ -297,40 +298,42 @@ arch_library_revision_dirs (rel_table opt_lib_path,
 
 
 
-t_uchar *
+arch_project_tree_t *
 arch_library_find (rel_table opt_lib_path,
 		   arch_patch_id *patch_id,
                    int check_inode_sigs)
 {
   static assoc_table validated_trees = NULL;
   rel_table revision_dirs = 0;
-  t_uchar * answer = 0;
-  t_uchar * tree_version = 0;
+  arch_project_tree_t * answer = NULL;
 
   revision_dirs = arch_library_revision_dirs (opt_lib_path, arch_patch_id_archive(patch_id), arch_patch_id_revision(patch_id), 0);
   if (rel_n_records(revision_dirs) == 0)
     return 0;
 
   invariant (!safe_access (revision_dirs[0][0], F_OK));
-
-  if (check_inode_sigs && !assoc_ref (validated_trees, revision_dirs[0][0]))
+  
     {
-      if (!arch_valid_inode_sig (revision_dirs[0][0], arch_patch_id_archive(patch_id), arch_patch_id_revision(patch_id)))
+      t_uchar * tree_version = arch_tree_version (revision_dirs[0][0]);
+      if (str_cmp (tree_version, arch_patch_id_branch (patch_id)))
+	  arch_set_tree_version (revision_dirs[0][0], patch_id);
+      lim_free (0, tree_version);
+    }
+
+  answer = arch_project_tree_new_exact (talloc_context, revision_dirs[0][0]);
+  if (check_inode_sigs && !assoc_ref (validated_trees, answer->root))
+    {
+      if (!arch_valid_inode_sig (answer, arch_patch_id_archive(patch_id), arch_patch_id_revision(patch_id)))
 	{
 	  safe_printfmt (2, "corrupt library (failed inode signature validation)\n    revision: %s\nYou should remove this revision from your library.\n",
 			 arch_patch_id_patch_id(patch_id));
 	  exit (2);
 	}
-      assoc_set (&validated_trees, revision_dirs[0][0], arch_patch_id_patch_id(patch_id));
+      assoc_set (&validated_trees, answer->root, arch_patch_id_patch_id(patch_id));
     }
 
-  answer = str_save (0, revision_dirs[0][0]);
-  tree_version = arch_tree_version (answer);
-  if (str_cmp (tree_version, arch_patch_id_branch (patch_id)))
-    arch_set_tree_version (answer, arch_patch_id_branch (patch_id));
   rel_free_table (revision_dirs);
 
-  lim_free (0, tree_version);
   return answer;
 }
 
@@ -339,17 +342,17 @@ arch_library_find (rel_table opt_lib_path,
 t_uchar *
 arch_library_find_file (t_uchar * archive, t_uchar * revision, t_uchar * loc)
 {
-  t_uchar * revision_dir = 0;
+  arch_project_tree_t * revision_tree;
   t_uchar * file_path = 0;
   arch_patch_id patch_id;
 
   arch_patch_id_init_archive (&patch_id, archive, revision);
-  revision_dir = arch_library_find (0, &patch_id, 1);
+  revision_tree = arch_library_find (0, &patch_id, 1);
   arch_patch_id_finalise (&patch_id);
 
-  if (revision_dir)
+  if (revision_tree)
     {
-      file_path = file_name_in_vicinity (0, revision_dir, loc);
+      file_path = file_name_in_vicinity (0, revision_tree->root, loc);
 
       if (safe_access (file_path, F_OK))
         {
@@ -358,7 +361,7 @@ arch_library_find_file (t_uchar * archive, t_uchar * revision, t_uchar * loc)
         }
     }
 
-  lim_free (0, revision_dir);
+  arch_project_tree_delete (revision_tree);
 
   return file_path;
 }
@@ -367,15 +370,15 @@ arch_library_find_file (t_uchar * archive, t_uchar * revision, t_uchar * loc)
 t_uchar *
 arch_library_find_file_by_id (t_uchar * archive, t_uchar * revision, t_uchar * id)
 {
-  t_uchar * revision_dir = 0;
+  arch_project_tree_t * revision_tree;
   t_uchar * file_path = 0;
   arch_patch_id patch_id;
 
   arch_patch_id_init_archive (&patch_id, archive, revision);
-  revision_dir = arch_library_find (0, &patch_id, 1);
+  revision_tree = arch_library_find (0, &patch_id, 1);
   arch_patch_id_finalise (&patch_id);
 
-  if (revision_dir)
+  if (revision_tree)
     {
       t_uchar * index_file = 0;
       rel_table index = 0;
@@ -388,7 +391,7 @@ arch_library_find_file_by_id (t_uchar * archive, t_uchar * revision, t_uchar * i
         {
           if (!str_cmp (id, index[x][1]))
             {
-              file_path = file_name_in_vicinity (0, revision_dir, index[x][0]);
+              file_path = file_name_in_vicinity (0, revision_tree->root, index[x][0]);
               break;
             }
         }
@@ -397,7 +400,7 @@ arch_library_find_file_by_id (t_uchar * archive, t_uchar * revision, t_uchar * i
       rel_free_table (index);
     }
 
-  lim_free (0, revision_dir);
+  arch_project_tree_delete (revision_tree);
 
   return file_path;
 }
@@ -529,8 +532,8 @@ arch_library_versions (t_uchar * archive, t_uchar * branch)
 
 
 rel_table
-arch_library_revisions (t_uchar * archive,
-                        t_uchar * version,
+arch_library_revisions (t_uchar const * const archive,
+                        t_uchar const * const version,
                         int full)
 {
   rel_table answer = 0;
@@ -550,20 +553,18 @@ arch_library_revisions (t_uchar * archive,
       
       for (y = 0; y < rel_n_records (tmp); ++y)
         {
-          t_uchar * t = 0;
-
-          t = tmp[y][0];
-
           if (full)
             {
-              tmp[y][0] = arch_fully_qualify (archive, t);
+	      rel_replace_record (tmp, y, rel_make_record 
+				  (arch_fully_qualify (archive, tmp[y][0]),
+				   NULL));
             }
           else
             {
-              tmp[y][0] = arch_parse_package_name (arch_ret_patch_level, 0, t);
+	      rel_replace_record (tmp, y, rel_make_record 
+				  (arch_parse_package_name (arch_ret_patch_level, 0, tmp[y][0]),
+				   NULL));
             }
-
-          lim_free (0, t);
         }
 
       rel_append_x (&answer, tmp);
@@ -592,7 +593,7 @@ arch_library_revisions (t_uchar * archive,
 t_uchar *
 arch_library_log (t_uchar * archive, t_uchar * revision)
 {
-  t_uchar * rev_dir = 0;
+  arch_project_tree_t * rev_tree = 0;
   t_uchar * log_path = 0;
   int in_fd;
   t_uchar * answer = 0;
@@ -600,17 +601,17 @@ arch_library_log (t_uchar * archive, t_uchar * revision)
   arch_patch_id patch_id;
 
   arch_patch_id_init_archive (&patch_id, archive, revision);
-  rev_dir = arch_library_find (0, &patch_id, 1);
+  rev_tree = arch_library_find (0, &patch_id, 1);
   arch_patch_id_finalise (&patch_id);
 
-  log_path = arch_log_file (rev_dir, archive, revision);
+  log_path = arch_log_file (rev_tree->root, archive, revision);
   in_fd = safe_open (log_path, O_RDONLY, 0);
   safe_file_to_string (&answer, &len, in_fd);
   answer = lim_realloc (0, answer, len + 1);
   answer[len] = 0;
   safe_close (in_fd);
 
-  lim_free (0, rev_dir);
+  arch_project_tree_delete (rev_tree);
   lim_free (0, log_path);
   return answer;
 }
@@ -620,32 +621,32 @@ arch_library_log (t_uchar * archive, t_uchar * revision)
  * already have found the library dir 
  */
 t_uchar *
-arch_library_index_file (t_uchar * archive, t_uchar * revision)
+arch_library_index_file (t_uchar const * const archive, t_uchar const * const revision)
 {
-  t_uchar * revision_dir = 0;
+  arch_project_tree_t * rev_tree;
   t_uchar * answer = 0;
   arch_patch_id patch_id;
 
   arch_patch_id_init_archive (&patch_id, archive, revision);
-  revision_dir = arch_library_find (0, &patch_id, 1);
+  rev_tree = arch_library_find (0, &patch_id, 1);
   arch_patch_id_finalise (&patch_id);
 
-  invariant (!!revision_dir);
+  invariant (!!rev_tree);
 
-  answer = file_name_in_vicinity (0, revision_dir, ",,index");
+  answer = file_name_in_vicinity (0, rev_tree->root, ",,index");
 
-  lim_free (0, revision_dir);
+  arch_project_tree_delete (rev_tree);
   return answer;
 }
 
 
 rel_table
-arch_library_index (t_uchar * archive, t_uchar * revision)
+arch_library_index (arch_patch_id * revision)
 {
   t_uchar * index_path = 0;
   rel_table answer = 0;
 
-  index_path = arch_library_index_file (archive, revision);
+  index_path = arch_library_index_file (arch_patch_id_archive (revision), arch_patch_id_revision (revision));
   answer = arch_read_changeset_index (index_path);
 
   lim_free (0, index_path);
@@ -655,7 +656,7 @@ arch_library_index (t_uchar * archive, t_uchar * revision)
 
 
 int
-arch_library_has_archive (t_uchar * lib, t_uchar * archive)
+arch_library_has_archive (t_uchar const * const lib, t_uchar const * const archive)
 {
   t_uchar * path = file_name_in_vicinity (0, lib, archive);
   int answer;
@@ -668,7 +669,7 @@ arch_library_has_archive (t_uchar * lib, t_uchar * archive)
 
 
 int
-arch_library_has_category (t_uchar * lib, t_uchar * archive, t_uchar * category)
+arch_library_has_category (t_uchar const * const lib, t_uchar const * const archive, t_uchar const * const category)
 {
   t_uchar * archive_path = file_name_in_vicinity (0, lib, archive);
   t_uchar * category_path = file_name_in_vicinity (0, archive_path, category);
@@ -683,7 +684,7 @@ arch_library_has_category (t_uchar * lib, t_uchar * archive, t_uchar * category)
 
 
 int
-arch_library_has_branch (t_uchar * lib, t_uchar * archive, t_uchar * branch)
+arch_library_has_branch (t_uchar const * const lib, t_uchar const * const archive, t_uchar const * const branch)
 {
   t_uchar * category = arch_parse_package_name (arch_ret_category, 0, branch);
   t_uchar * archive_path = file_name_in_vicinity (0, lib, archive);
@@ -703,7 +704,7 @@ arch_library_has_branch (t_uchar * lib, t_uchar * archive, t_uchar * branch)
 
 
 int
-arch_library_has_version (t_uchar * lib, t_uchar * archive, t_uchar * version)
+arch_library_has_version (t_uchar const * const lib, t_uchar const * const archive, t_uchar const * const version)
 {
   t_uchar * category = arch_parse_package_name (arch_ret_category, 0, version);
   t_uchar * branch = arch_parse_package_name (arch_ret_package, 0, version);
@@ -727,20 +728,18 @@ arch_library_has_version (t_uchar * lib, t_uchar * archive, t_uchar * version)
 
 
 t_uchar *
-arch_library_revision_dir_in_lib (t_uchar * lib, t_uchar * archive, t_uchar * revision)
+arch_library_revision_dir_in_lib (t_uchar const * const lib, arch_patch_id * patch_id)
 {
-  t_uchar * category = arch_parse_package_name (arch_ret_category, 0, revision);
-  t_uchar * branch = arch_parse_package_name (arch_ret_package, 0, revision);
-  t_uchar * version = arch_parse_package_name (arch_ret_package_version, 0, revision);
-  t_uchar * archive_path = file_name_in_vicinity (0, lib, archive);
+  t_uchar * branch = arch_parse_package_name (arch_ret_package, 0, arch_patch_id_revision (patch_id));
+  t_uchar * category = arch_parse_package_name (arch_ret_category, 0, arch_patch_id_revision (patch_id));
+  t_uchar * archive_path = file_name_in_vicinity (0, lib, arch_patch_id_archive (patch_id));
   t_uchar * category_path = file_name_in_vicinity (0, archive_path, category);
   t_uchar * branch_path = file_name_in_vicinity (0, category_path, branch);
-  t_uchar * version_path = file_name_in_vicinity (0, branch_path, version);
-  t_uchar * revision_path = file_name_in_vicinity (0, version_path, revision);
+  t_uchar * version_path = file_name_in_vicinity (0, branch_path, arch_patch_id_version (patch_id));
+  t_uchar * revision_path = file_name_in_vicinity (0, version_path, arch_patch_id_revision (patch_id));
 
-  lim_free (0, category);
   lim_free (0, branch);
-  lim_free (0, version);
+  lim_free (0, category);
 
   lim_free (0, archive_path);
   lim_free (0, category_path);
@@ -840,7 +839,7 @@ arch_verify_is_library (t_uchar * lib)
         return;
     }
 
-  safe_printfmt (2, "tla: indicated library is not on library path\n  dir %s\n", lib);
+  safe_printfmt (2, "baz: indicated library is not on library path\n  dir %s\n", lib);
   exit (2);
 }
 
