@@ -14,6 +14,7 @@
 #include "config-options.h"
 #include "po/gettext.h"
 #include "hackerlab/bugs/panic.h"
+#include "hackerlab/bugs/exception.h"
 #include "hackerlab/arrays/ar.h"
 #include "hackerlab/mem/mem.h"
 #include "hackerlab/char/str.h"
@@ -22,6 +23,7 @@
 #include "libfsutils/file-contents.h"
 #include "libfsutils/tmp-files.h"
 #include "libfsutils/rmrf.h"
+#include "libawk/relassoc.h"
 #include "libarch/ancestry.h"
 #include "libarch/namespace.h"
 #include "libarch/patch-logs.h"
@@ -136,35 +138,13 @@ arch_common_ancestor_search_prune_warning(struct arch_common_ancestor_search *se
     search->pruned = NULL;
 }
 
-static t_uchar *
-log_text (t_uchar * tree_root, t_uchar const * fqrevision)
-{
-    t_uchar *archive=arch_parse_package_name (arch_ret_archive, 0, fqrevision);
-    t_uchar *version=arch_parse_package_name (arch_ret_package_version, 0, fqrevision);
-    rel_table source_logs = arch_logs (tree_root, archive, version, 1);
-    t_uchar * log_text = 0;
-    int x;
-
-    for (x = 0; x < rel_n_records (source_logs); ++x)
-	if (!str_cmp (fqrevision, source_logs[x][0]))
-	  {
-	    log_text = file_contents (source_logs[x][1]);
-	    break;
-	  }
-
-    lim_free (0, archive);
-    lim_free (0, version);
-    rel_free_table (source_logs);
-    return log_text;
-}
-
 /* append return the directly reachable patches from fqrevision to out
  * , in no particular order
  */
 static void
-arch_included_from_patch (struct arch_common_ancestor_search *search, rel_table * out, t_uchar *tree_root, t_uchar * fqrevision)
+arch_included_from_patch (struct arch_common_ancestor_search *search, rel_table * out, arch_project_tree_t *tree, t_uchar * fqrevision)
 {
-  t_uchar * log = log_text (tree_root, fqrevision);
+  t_uchar * log = arch_project_tree_patch_log (tree, fqrevision);
   assoc_table headers = 0;
   t_uchar * new_patches_header;
   rel_table new_patches = 0;
@@ -188,7 +168,7 @@ arch_included_from_patch (struct arch_common_ancestor_search *search, rel_table
        * a subtraction to get the real added headers.
        * then remove fqrevision and *boom* we are done.
        */
-      rel_add_records (out, rel_make_record (continuation_header, tree_root, 0), 0);
+      rel_add_records (out, rel_make_record (continuation_header, 0), 0);
     }
   else
     {
@@ -197,15 +177,53 @@ arch_included_from_patch (struct arch_common_ancestor_search *search, rel_table
       
       for (x = 0; x < rel_n_records (new_patches); ++x)
 	  if (str_cmp( fqrevision, new_patches[x][0]))
-	      rel_add_records (out, rel_make_record (new_patches[x][0], tree_root, 0), 0);
+	      rel_add_records (out, rel_make_record (new_patches[x][0], 0), 0);
 
-      /* add the namespace predecessor */
+      /* add the patch ancestor predecessor */
+      if (!str_cmp("-0", fqrevision + str_length(fqrevision) - 2))
+        /* version-0, cannot determine the ancestor from the tree */
 	{
-	  t_uchar * predecessor = arch_highest_patch_level_before (tree_root, fqrevision);
-	  if (predecessor)
-	      rel_add_records (out, rel_make_record (predecessor, tree_root, 0), 0);
-	  lim_free (0, predecessor);
+          struct exception * e;
+          Try 
+            {
+              arch_patch_id * patch_id = talloc_steal (talloc_context, 
+                                                       arch_patch_id_new (fqrevision));
+              arch_patch_id * predecessor = arch_patch_ancestor (talloc_context, patch_id, 0);
+              if (predecessor)
+                {
+                  if (arch_project_tree_has_patch (tree, arch_patch_id_patch_id (predecessor)))
+                      rel_add_records (out, rel_make_record (arch_patch_id_patch_id (predecessor), 0), 0);
+                }
+              /* if the patch ancestry isn't available, we definately cannot build a 
+               * reference tree, so we just don't record it as reachable
+               */
+              talloc_free (patch_id);
+            }
+          Catch (e)
+            {
+              /* dont mask critical errors */
+              if (e->code < 0)
+                  Throw (e);
+              talloc_free (e);
+            }
 	}
+      else
+        /* tree ancestry is sufficient for our purposes */
+        {
+          t_uchar *predecessor = assoc_ref (tree->local_ancestry, fqrevision);
+          if (predecessor)
+            {
+              if (str_length (predecessor) && str_cmp (predecessor, ARCH_ANCESTRY_PATCH_PARTIAL))
+                  rel_add_records (out, rel_make_record (predecessor, 0), 0);
+            }
+          else
+            {
+              predecessor = arch_highest_patch_level_before (tree, fqrevision);
+              if (predecessor)
+                  rel_add_records (out, rel_make_record (predecessor, 0), 0);
+              lim_free (0, predecessor);
+            }
+        }
     }
 
   free_assoc_table (headers);
@@ -217,12 +235,12 @@ arch_included_from_patch (struct arch_common_ancestor_search *search, rel_table
 /* FIXME: this should be a simply adjacency list djikstra implementation. */
 /* topologically sort patches according to ancestry */
 static rel_table
-arch_topological_sort (struct arch_common_ancestor_search *search, rel_table fqrevisions, t_uchar *tree_root, int trace)
+arch_topological_sort (struct arch_common_ancestor_search *search, rel_table fqrevisions, arch_project_tree_t *tree, int trace)
 {
     /* generate predecessors */
     /* col 0 is the node. cols beyond that are the reachables */
-    rel_table graph = 0;
-    rel_table answer = 0;
+    rel_table graph = NULL;
+    rel_table answer = NULL;
     assoc_table found = 0;
     int found_count = 0;
     int loop;
@@ -233,7 +251,7 @@ arch_topological_sort (struct arch_common_ancestor_search *search, rel_table fqr
 	int graph_row;
 	
 	rel_add_records (&graph, rel_make_record (fqrevisions[loop][0], 0), 0);
-	arch_included_from_patch (search, &predecessors, tree_root, fqrevisions[loop][0]);
+	arch_included_from_patch (search, &predecessors, tree, fqrevisions[loop][0]);
 	graph_row = rel_n_records (graph) -1;
 	
 	for (fields=0; fields < rel_n_records (predecessors); ++fields)
@@ -270,7 +288,7 @@ arch_topological_sort (struct arch_common_ancestor_search *search, rel_table fqr
 	        //safe_printfmt(2, "unreachable %s\n", graph[loop][0]);
 		/* unreachable */
 		++found_count;
-		rel_add_records (&answer, rel_make_record (graph[loop][0], tree_root, 0), 0);
+		rel_add_records (&answer, rel_make_record (graph[loop][0], 0), 0);
 		assoc_set (&found, graph[loop][0], graph[loop][0]);
 		
 	      }
@@ -283,7 +301,7 @@ arch_topological_sort (struct arch_common_ancestor_search *search, rel_table fqr
     //safe_printfmt (2, "sorted table\n");
     //rel_print_table (2, answer);
     //safe_printfmt (2, "============\n");
-
+    rel_free_table (graph);
     return answer;
 }
 
@@ -300,12 +318,12 @@ arch_topological_sort (struct arch_common_ancestor_search *search, rel_table fqr
    * A1 even though A1 is a 'new patch'
    */
 static void
-arch_common_ancestor_search_reachable_from_patch (struct arch_common_ancestor_search * search, t_uchar *tree_root, t_uchar * fqrevision)
+arch_common_ancestor_search_reachable_from_patch (struct arch_common_ancestor_search * search, arch_project_tree_t *tree, t_uchar * fqrevision)
 {
   rel_table reachable = 0;
   if (search->trace)
       safe_printfmt (2, "Adding to todo queue as reachable from %s:\n", fqrevision);
-  arch_included_from_patch (search, &reachable, tree_root, fqrevision);
+  arch_included_from_patch (search, &reachable, tree, fqrevision);
   if (search->trace)
     {
       int loop;
@@ -325,7 +343,7 @@ arch_common_ancestor_search_reachable_from_patch (struct arch_common_ancestor_se
  */
 static void
 arch_reachable_from_tree (struct arch_common_ancestor_search *search, arch_project_tree_t *tree, t_uchar * fqrevision,
-			  t_uchar * tmp_dir, t_uchar *cache_dir)
+			  t_uchar * tmp_dir, arch_project_tree_t * cache)
 {
   t_uchar * reference_tree_root;
   rel_table reachable = 0;
@@ -333,15 +351,15 @@ arch_reachable_from_tree (struct arch_common_ancestor_search *search, arch_proje
   if (search->trace)
       safe_printfmt (2, "Adding to todo queue as reachable (from tree) %s:\n", fqrevision);
   /* patches in the tree from fqrevision */
-  arch_included_from_patch (search, &reachable, tree->root, fqrevision);
-  reference_tree_root = arch_find_or_make_tmp_local_copy (-1, tmp_dir, 0, cache_dir, 0, tree->archive, tree->revision);
+  arch_included_from_patch (search, &reachable, tree, fqrevision);
+  reference_tree_root = arch_find_or_make_tmp_local_copy (-1, tmp_dir, tree, cache, 0, tree->archive, tree->revision);
   new_in_tree = arch_new_in_tree (tree, reference_tree_root);
   /* I'm sure there is a function to do this trivially */
     {
       int loop;
       for (loop=0; loop < rel_n_records (new_in_tree); ++loop)
         {
-	  rel_add_records(&reachable, rel_make_record (new_in_tree[loop][0], tree->root, 0), 0);
+	  rel_add_records(&reachable, rel_make_record (new_in_tree[loop][0], 0), 0);
         }
     }
   if (search->trace)
@@ -391,7 +409,7 @@ arch_common_ancestor_search_finalise(struct arch_common_ancestor_search *search)
 }
 
 static void
-arch_common_ancestor_search_reach (struct arch_common_ancestor_search *search, int trace)
+arch_common_ancestor_search_reach (struct arch_common_ancestor_search *search, arch_project_tree_t *search_tree, int trace)
 {
     rel_table current=search->todo;
     int position;
@@ -411,7 +429,7 @@ arch_common_ancestor_search_reach (struct arch_common_ancestor_search *search, i
       }
     for (position=0; position < rel_n_records (filtered); ++position)
       {
-	arch_common_ancestor_search_reachable_from_patch (search, filtered[position][1], filtered[position][0]);
+	arch_common_ancestor_search_reachable_from_patch (search, search_tree, filtered[position][0]);
       }
   rel_free_table (current);
   rel_free_table (filtered);
@@ -427,15 +445,26 @@ typedef struct arch_common_ancestor_state
 static t_uchar *
 arch_project_tree_patch_ancestor (arch_project_tree_t *tree, t_uchar const *patch_id)
 {
-  t_uchar * log = log_text (tree->root, patch_id);
+  t_uchar * log;
   assoc_table headers = 0;
   t_uchar * new_patches_header;
   rel_table new_patches = 0;
   t_uchar * continuation_header;
   t_uchar *result;
 
+  result = assoc_ref (tree->local_ancestry, patch_id);
+  if (result)
+    {
+      if (str_length (result))
+          return str_save (0, result);
+      else
+          return NULL;
+    }
+  
+  log = arch_project_tree_patch_log (tree, patch_id);
   if (!log)
     {
+      assoc_set (&tree->local_ancestry, patch_id, ARCH_ANCESTRY_PATCH_PARTIAL);
       return str_save (0, ARCH_ANCESTRY_PATCH_PARTIAL);
     }
 
@@ -453,51 +482,79 @@ arch_project_tree_patch_ancestor (arch_project_tree_t *tree, t_uchar const *patc
      * were removed
      */
     {
-      t_uchar * predecessor = arch_highest_patch_level_before (tree->root, patch_id);
-      if (predecessor)
-          result = str_save (0, predecessor);
+      /* add the patch ancestor predecessor */
+      if (!str_cmp("-0", patch_id + str_length(patch_id) - 2))
+        {
+          /* version-0, cannot determine the ancestor from the tree */
+          struct exception * e;
+          Try 
+            {
+              arch_patch_id * patch_id_ = talloc_steal (talloc_context, 
+                                                       arch_patch_id_new (patch_id));
+              arch_patch_id * patch_result = arch_patch_ancestor (talloc_context,
+                                                                  patch_id_, 0);
+              if (patch_result)
+                  result = str_save (0, arch_patch_id_patch_id (patch_result));
+              else
+                  result = NULL;
+            }
+          Catch (e)
+            {
+              /* dont mask critical errors */
+              if (e->code < 0)
+                  Throw (e);
+              talloc_free (e);
+              result = str_save (0, ARCH_ANCESTRY_PATCH_PARTIAL);
+            }
+        }
       else
-          result = NULL;
-      lim_free (0, predecessor);
+        /* tree ancestry is sufficient for our purposes */
+	  result = arch_highest_patch_level_before (tree, patch_id);
     }
+  if (!result)
+      assoc_set (&tree->local_ancestry, patch_id, "");
+  else
+      assoc_set (&tree->local_ancestry, patch_id, result);
   free_assoc_table (headers);
   rel_free_table (new_patches);
   lim_free (0, log);
   return result;
 }
 
-/* find value in the set set 
- * currently does not depend on any sort order
- * but maybe somewhat inefficient 
- */
-static int
-rel_set_find (rel_table set, t_uchar const *value)
-{
-    int position;
-    for (position = 0; position < rel_n_records (set); ++position)
-        if (!str_cmp (set[position][0], value))
-            return -1;
-    return 0;
-}
-
 /* determine the ancestry of patch_id that is distinct from the ancestry for tree)
  */
 static rel_table
 arch_ancestry_unique_to_branch (t_uchar * patch_id, arch_project_tree_t *tree)
 {
-    rel_table base_ancestry = arch_project_tree_ancestry (tree);
+    assoc_table base_ancestry = rel_to_assoc (arch_project_tree_ancestry(tree), 0, 0);
     t_uchar *current = str_save (0, patch_id);
     rel_table result = NULL;
     while (1)
       {
         /* may hit the archive, thats ok */
-        t_uchar *next = arch_patch_ancestor (current, 1, 0);
-        if (!str_cmp (next, ARCH_ANCESTRY_PATCH_PARTIAL))
-          /* couldn't find it */
+        t_uchar * next = NULL;
+        struct exception * e;
+        Try 
           {
-            lim_free (0, next);
+            arch_patch_id * patch_id = talloc_steal (talloc_context, 
+                                                     arch_patch_id_new (current));
+            arch_patch_id * patch_result = arch_patch_ancestor (talloc_context,
+                                                                patch_id, 0);
+            if (patch_result)
+                next = str_save (0, arch_patch_id_patch_id (patch_result));
+            else
+                next = NULL;
+          }
+        Catch (e)
+          {
+            /* dont mask critical errors */
+            if (e->code < 0)
+                Throw (e);
+            talloc_free (e);
+            /* no ancestor from ancestry, use local cache */
             next = arch_project_tree_patch_ancestor (tree, current);
           }
+
         if (!str_cmp (next, ARCH_ANCESTRY_PATCH_PARTIAL))
           /* still couldn't */
           {
@@ -510,13 +567,13 @@ arch_ancestry_unique_to_branch (t_uchar * patch_id, arch_project_tree_t *tree)
             break;
         /* ok have an ancestor */
         rel_add_records (&result, rel_make_record (next, 0), 0);
-        lim_free (0, current);
-        current = next;
-        if (rel_set_find (base_ancestry, next))
+        current = str_replace (current, next);
+        if (assoc_ref (base_ancestry, next))
             /* we've met the branch */
             break;
       }
     lim_free (0, current);
+    free_assoc_table (base_ancestry);
     return result;
 }
 
@@ -535,13 +592,20 @@ arch_patch_is_cherrypicked (t_uchar * patch_id, arch_project_tree_t *tree, int t
     unchecked = arch_ancestry_unique_to_branch (patch_id, tree);
     for (position=0; position < rel_n_records (unchecked); ++position)
       {
-        if (!arch_project_tree_has_patch (tree, unchecked[position][0]))
+        if (!str_cmp (unchecked[position][0], ARCH_ANCESTRY_PATCH_PARTIAL))
+          {
+            if (trace)
+                safe_printfmt (2, "full ancestry to the branch was not available from tree %s\n", tree->root);
+            result = -1;
+          }
+        else if (!arch_project_tree_has_patch (tree, unchecked[position][0]))
           {
             if (trace)
                 safe_printfmt (2, "missing patch %s from tree %s\n", unchecked[position][0], tree->root);
             result = -1;
           }
       }
+    rel_free_table (unchecked);
     return result;
     
 }
@@ -577,19 +641,19 @@ rel_remove_common (rel_table *me, rel_table other)
 static t_uchar *
 arch_nearest_common_ancestor (arch_project_tree_t * from_tree, t_uchar * from_archive, t_uchar * from_revision,
                               arch_project_tree_t * to_tree, t_uchar * to_archive, t_uchar * to_version, int trace,
-			      t_uchar *tmp_dir, t_uchar *cache_dir)
+			      t_uchar *tmp_dir, arch_project_tree_t * cache)
 {
   common_ancestor state;
   t_uchar * from_fqrevision = 0;
   t_uchar * to_fqrevision = 0;
   rel_table intersection = 0;
+  t_uchar * result = NULL;
   arch_common_ancestor_search_init (&state.from);
   arch_common_ancestor_search_init (&state.to);
   /* TODO refactor this */
   state.trace = trace;
   state.from.trace = trace;
   state.to.trace = trace;
-  t_uchar * result = NULL;
 
     {
       t_uchar * patch_level = arch_highest_patch_level (to_tree->root, to_archive, to_version);
@@ -599,11 +663,11 @@ arch_nearest_common_ancestor (arch_project_tree_t * from_tree, t_uchar * from_ar
   from_fqrevision = arch_fully_qualify (from_archive, from_revision);
 
   /* FIXME pass in the chatter_fd */
-  safe_printfmt (1, "Searching for best merge point\n");
+  safe_printfmt (1, "* Searching for best merge point: ");
   
   /* from_tree is always a pristine */
-  arch_common_ancestor_search_reachable_from_patch (&state.from, from_tree->root, from_fqrevision);
-  arch_reachable_from_tree (&state.to, to_tree, to_fqrevision, tmp_dir, cache_dir);
+  arch_common_ancestor_search_reachable_from_patch (&state.from, from_tree, from_fqrevision);
+  arch_reachable_from_tree (&state.to, to_tree, to_fqrevision, tmp_dir, cache);
   arch_visited (&state.from.reached, from_fqrevision, trace);
   arch_visited (&state.to.reached, to_fqrevision, trace);
   while (!result && ( rel_n_records (state.to.todo) || rel_n_records (state.from.todo)))
@@ -611,9 +675,10 @@ arch_nearest_common_ancestor (arch_project_tree_t * from_tree, t_uchar * from_ar
       if (state.trace)
 	  safe_printfmt (2, "Checking next reachable nodes\n");
       safe_printfmt (1, ".");
+      safe_flush (1);
       
-      arch_common_ancestor_search_reach (&state.from, state.trace);
-      arch_common_ancestor_search_reach (&state.to, state.trace);
+      arch_common_ancestor_search_reach (&state.from, from_tree, state.trace);
+      arch_common_ancestor_search_reach (&state.to,  to_tree, state.trace);
       intersection = assoc_intersection (state.from.reached, state.to.reached);
       if (!intersection)
 	{
@@ -639,7 +704,12 @@ arch_nearest_common_ancestor (arch_project_tree_t * from_tree, t_uchar * from_ar
           rel_remove_common (&intersection, from_cherries);
           rel_remove_common (&intersection, to_cherries);
           
-          intersection = arch_topological_sort (&state.from, intersection, from_tree->root, trace);
+            {
+              rel_table temp_table = arch_topological_sort (&state.from, intersection, from_tree, trace);
+              rel_free_table (intersection);
+              intersection = temp_table;
+            }
+
           /* don't warn on pruned from this sort, they are irrelevant */
           if (state.trace)
             {
@@ -654,6 +724,8 @@ arch_nearest_common_ancestor (arch_project_tree_t * from_tree, t_uchar * from_ar
           rel_free_table (to_cherries);
         }
     }
+  lim_free (0, from_fqrevision);
+  lim_free (0, to_fqrevision);
   arch_common_ancestor_search_finalise (&state.from);
   arch_common_ancestor_search_finalise (&state.to);
   safe_printfmt (1, " done\n");
@@ -684,8 +756,8 @@ arch_nearest_common_ancestor (arch_project_tree_t * from_tree, t_uchar * from_ar
 int
 arch_star_merge (int chatter_fd,
                  struct arch_archive * from_arch, t_uchar * from_archive, t_uchar * from_revision,
-                 t_uchar * to_tree_root, struct arch_archive * to_arch, t_uchar * to_archive, t_uchar * to_version,
-                 t_uchar * cache_dir, t_uchar * changeset_output, int use_diff3, int forward,
+                 arch_project_tree_t * to_tree, struct arch_archive * to_arch, t_uchar * to_archive, t_uchar * to_version,
+                 arch_project_tree_t * cache, t_uchar * changeset_output, int use_diff3, int forward,
                  int escape_classes, int show_points_only, int graph_merge, int trace)
 {
   int answer = 2;
@@ -695,12 +767,10 @@ arch_star_merge (int chatter_fd,
   t_uchar * fq_orig_revision = 0;
   t_uchar * orig_archive = 0;
   t_uchar * orig_revision = 0;
-  arch_project_tree_t from_tree;
-  arch_project_tree_t to_tree;
+  arch_project_tree_t * from_tree;
+  arch_project_tree_t * orig_tree = NULL;
 
-  arch_project_tree_init (&to_tree, to_tree_root);
-
-  tmp_dir = tmp_file_name (to_tree_root, ",,merge");
+  tmp_dir = talloc_tmp_file_name (talloc_context, to_tree->root, ",,merge");
   if (!changeset_output)
     changeset = file_name_in_vicinity (0, tmp_dir, ",,changeset");
   else
@@ -709,13 +779,13 @@ arch_star_merge (int chatter_fd,
   rmrf_file (tmp_dir);
   safe_mkdir (tmp_dir, 0777);
 
-  from_tree_root = arch_find_or_make_tmp_local_copy (chatter_fd, tmp_dir, to_tree_root, cache_dir, from_arch, from_archive, from_revision);
-  arch_project_tree_init (&from_tree, from_tree_root);
+  from_tree_root = arch_find_or_make_tmp_local_copy (chatter_fd, tmp_dir, to_tree, cache, from_arch, from_archive, from_revision);
+  from_tree = arch_project_tree_new (talloc_context, from_tree_root);
 
   if (graph_merge)
-      fq_orig_revision = arch_nearest_common_ancestor (&from_tree, from_archive, from_revision, &to_tree, to_archive, to_version, trace, tmp_dir, cache_dir);
+      fq_orig_revision = arch_nearest_common_ancestor (from_tree, from_archive, from_revision, to_tree, to_archive, to_version, trace, tmp_dir, cache);
   else
-      fq_orig_revision = arch_nearest_star_ancestor (&from_tree, from_archive, from_revision, &to_tree, to_archive, to_version, trace);
+      fq_orig_revision = arch_nearest_star_ancestor (from_tree, from_archive, from_revision, to_tree, to_archive, to_version, trace);
 
   if (!fq_orig_revision)
     {
@@ -726,7 +796,7 @@ arch_star_merge (int chatter_fd,
     {
       t_uchar * orig_tree_root = 0;
       struct arch_make_changeset_report make_report = {0, };
-      struct arch_apply_changeset_report apply_report = {0, };
+      struct arch_apply_changeset_report * apply_report = NULL;
       assoc_table inode_shortcut = 0;
 
       if (show_points_only)
@@ -734,11 +804,11 @@ arch_star_merge (int chatter_fd,
 	  arch_chatter (chatter_fd, "* merge points are:\n");
 	  arch_chatter (chatter_fd, "base: %s\n", fq_orig_revision);
 	  arch_chatter (chatter_fd, "other: %s/%s\n", from_archive, from_revision);
-	  arch_chatter (chatter_fd, "this: %s\n", to_tree.fqrevision);
+	  arch_chatter (chatter_fd, "this: %s\n", to_tree->fqrevision);
 	  return 0;
 	}
 
-      arch_chatter (chatter_fd, "* merge by delta(%s,%s/%s)[%s]\n", fq_orig_revision, from_archive, from_revision, to_tree_root);
+      arch_chatter (chatter_fd, "* merge by delta(%s,%s/%s)[%s]\n", fq_orig_revision, from_archive, from_revision, to_tree->root);
 	{
 	  t_uchar *fq_from = str_alloc_cat_many (0, from_archive,"/", from_revision, str_end);
 	  int empty = !str_cmp (fq_from, fq_orig_revision);
@@ -755,7 +825,8 @@ arch_star_merge (int chatter_fd,
       orig_archive = arch_parse_package_name (arch_ret_archive, 0, fq_orig_revision);
       orig_revision = arch_parse_package_name (arch_ret_non_archive, 0, fq_orig_revision);
 
-      orig_tree_root = arch_find_or_make_tmp_local_copy (chatter_fd, tmp_dir, to_tree_root, cache_dir, (str_cmp (orig_archive, from_archive) ? 0 : from_arch), orig_archive, orig_revision);
+      orig_tree_root = arch_find_or_make_tmp_local_copy (chatter_fd, tmp_dir, to_tree, cache, (str_cmp (orig_archive, from_archive) ? 0 : from_arch), orig_archive, orig_revision);
+      orig_tree = arch_project_tree_new (talloc_context, orig_tree_root);
 
       if (changeset_output)
         {
@@ -765,12 +836,13 @@ arch_star_merge (int chatter_fd,
       if (use_diff3)
 	{
 	  arch_chatter (chatter_fd, "* performing merge\n");
-	  answer = arch_merge3 (chatter_fd, to_tree_root, orig_tree_root, from_tree_root, escape_classes, 1);
+	  answer = arch_merge3 (chatter_fd, to_tree, orig_tree, from_tree, escape_classes, 1);
+          lim_free (0, orig_tree_root);
 	  goto star_merge_cleanup;
 	}
 
       arch_read_inode_sig_ids (0, &inode_shortcut, from_tree_root, orig_archive, orig_revision);
-      arch_make_changeset (&make_report, orig_tree_root, from_tree_root, changeset, arch_unspecified_id_tagging, arch_inventory_unrecognized, 0, inode_shortcut, 0, escape_classes);
+      arch_make_changeset (&make_report, orig_tree, from_tree, changeset, arch_unspecified_id_tagging, arch_inventory_unrecognized, 0, inode_shortcut, 0, escape_classes);
 
       if (changeset_output)
         answer = 0;
@@ -779,14 +851,11 @@ arch_star_merge (int chatter_fd,
           assoc_table older_files_table = 0;
           assoc_table yours_files_table = 0;
 
-          apply_report.callback = star_merge_callback;
-          apply_report.thunk = (void *)(long)chatter_fd;
-
           arch_chatter (chatter_fd, "* applying changeset\n");
           safe_flush (chatter_fd);
-          arch_apply_changeset (&apply_report, changeset, to_tree_root, arch_unspecified_id_tagging, arch_inventory_unrecognized, 0, forward, escape_classes);
+          apply_report = arch_apply_changeset (changeset, talloc_context, to_tree, arch_unspecified_id_tagging, arch_inventory_unrecognized, 0, forward, escape_classes, star_merge_callback, (void *)(long)chatter_fd);
 
-          if (arch_conflicts_occured (&apply_report))
+          if (arch_conflicts_occurred (apply_report))
             answer = 1;
           else
             answer = 0;
@@ -796,7 +865,7 @@ arch_star_merge (int chatter_fd,
         }
 
       arch_free_make_changeset_report_data (&make_report);
-      arch_free_apply_changeset_report_data (&apply_report);
+      talloc_free (apply_report);
 
       lim_free (0, orig_tree_root);
       free_assoc_table (inode_shortcut);
@@ -805,13 +874,14 @@ arch_star_merge (int chatter_fd,
   star_merge_cleanup:
   rmrf_file (tmp_dir);
 
-  lim_free (0, tmp_dir);
+  talloc_free (tmp_dir);
   lim_free (0, from_tree_root);
   lim_free (0, fq_orig_revision);
   lim_free (0, orig_archive);
   lim_free (0, orig_revision);
-  arch_project_tree_finalise (&to_tree);
-  arch_project_tree_finalise (&from_tree);
+  lim_free (0, changeset);
+  arch_project_tree_delete (from_tree);
+  arch_project_tree_delete (orig_tree);
 
   return answer;
 }
