@@ -17,8 +17,8 @@
 
 extern int arch_star_merge (int chatter_fd,
                             struct arch_archive * from_arch, t_uchar * from_archive, t_uchar * from_revision,
-                            t_uchar * to_tree_root, struct arch_archive * to_arch, t_uchar * to_archive, t_uchar * to_version,
-                            t_uchar * cache_dir, t_uchar * changeset_output, int use_diff3, int forward,
+                            arch_project_tree_t * to_tree, struct arch_archive * to_arch, t_uchar * to_archive, t_uchar * to_version,
+                            arch_project_tree_t * cache, t_uchar * changeset_output, int use_diff3, int forward,
                             int escape_classes, int show_points_only, int graph_merge, int trace);
 #endif  /* INCLUDE__LIBARCH__STAR_MERGE_H */
 
