@@ -21,6 +21,9 @@
 
 
 /* __STDC__ prototypes for static functions */
+static int version_cmp_n (const t_uchar * a, size_t a_l, const t_uchar * b, size_t b_l);
+static int ver_order (t_uchar c);
+static int ver_cmp_n (const t_uchar * a, size_t a_l, const t_uchar * b, size_t b_l);
 static int arch_cmp_by_field (void * va, void * vb, void * vdata);
 static int arch_lvl_cmp_by_field (void * va, void * vb, void * vdata);
 static t_uchar const * over_opt_archive_prefix (t_uchar const * name);
@@ -106,7 +109,7 @@ arch_valid_archive_name (t_uchar const * name)
 }
 
 int
-arch_valid_patch_level_name (t_uchar * name)
+arch_valid_patch_level_name (t_uchar const * const name)
 {
   t_uchar const * lvl;
 
@@ -419,12 +422,36 @@ find_version_end (t_uchar const * version_start)
   return version_end;
 }
 
+/**
+ * \brief parse a package name
+ * \param type what parts of the package to acquire
+ * \param default_archive override a default if the package does not have an archive component
+ * \param name the package name to parse
+ * \return t_uchar * parsed, heap allocated value
+ */
 t_uchar *
 arch_parse_package_name (enum arch_parse_package_name_type type,
                          t_uchar const * default_archive,
                          t_uchar const * name)
 {
   invariant (arch_valid_package_name (name, arch_maybe_archive, arch_req_package, 1));
+  return arch_parse_name (type, default_archive, name);
+}
+
+/**
+ * \brief parse a package name
+ * \param type what parts of the package to acquire
+ * \param default_archive override a default if the package does not have an archive component
+ * \param name the package name to parse
+ * \return t_uchar * parsed, heap allocated value
+ */
+t_uchar *
+arch_parse_name (enum arch_parse_package_name_type type,
+                         t_uchar const * default_archive,
+                         t_uchar const * name)
+{
+  invariant (arch_valid_package_name (name, arch_maybe_archive, arch_req_package, 1) 
+             || (type == arch_ret_archive && arch_valid_archive_name (name)));
 
   switch (type)
     {
@@ -437,6 +464,9 @@ arch_parse_package_name (enum arch_parse_package_name_type type,
       {
         t_uchar * slash;
 
+        if (arch_valid_archive_name (name))
+            return str_save (0, name);
+
         slash = str_chr_index (name, '/');
         if (!slash)
           {
@@ -580,23 +610,24 @@ arch_parse_package_name (enum arch_parse_package_name_type type,
 }
 
 t_uchar *
-arch_fully_qualify (t_uchar * default_archive, t_uchar * name)
+arch_fully_qualify (t_uchar const * const archive, t_uchar const * const name)
 {
   invariant (arch_valid_package_name (name, arch_maybe_archive, arch_req_package, 1));
-  if (default_archive)
-    invariant (arch_valid_archive_name (default_archive));
+  if (archive)
+    invariant (arch_valid_archive_name (archive));
 
   if (str_chr_index (name, '/'))
     return str_save (0, name);
   else
     {
-      if (!default_archive)
-        default_archive = arch_my_default_archive (0);
+      t_uchar * default_archive = str_save (0, archive);
+      if (!str_length (default_archive))
+        default_archive = arch_my_default_archive (NULL);
 
       if (!default_archive)
         panic ("arch: no default archive set");
 
-      return str_alloc_cat_many (0, default_archive, "/", name, str_end);
+      return str_realloc_cat_many (0, default_archive, "/", name, str_end);
     }
 }
 
@@ -688,102 +719,22 @@ arch_names_cmp (t_uchar const * a, t_uchar const * b)
   a_nxt = over_version (a);
   b_nxt = over_version (b);
 
-  {
-    int a_vsn_len;
-    int b_vsn_len;
-
-    a_vsn_len = a_nxt - a;
-    b_vsn_len = b_nxt - b;
-
-    /* precondition:
-     *    a and b point to a version number component
-     *    a_vsn_len and b_vsn_len tell how many chars remain in the version id
-     *
-     * postcondition:
-     *    RETURNS if version ids differ
-     *    a and b point past version id otherwise
-     *
-     * goal:
-     *    compare dot-separated components numerically
-     *    noting that a and b may have different numbers of components
-     */
-    while (1)
-      {
-        int errn;
-        t_uchar const * a_comp_end;
-        t_uchar const * b_comp_end;
-        t_ulong a_comp;
-        t_ulong b_comp;
-
-        a_comp_end = str_chr_index_n (a, a_vsn_len, '.');
-        if (!a_comp_end)
-          a_comp_end = a + a_vsn_len;
-
-        b_comp_end = str_chr_index_n (b, b_vsn_len, '.');
-        if (!b_comp_end)
-          b_comp_end = b + b_vsn_len;
-
-        invariant (!cvt_decimal_to_ulong (&errn, &a_comp, a, a_comp_end - a));
-        invariant (!cvt_decimal_to_ulong (&errn, &b_comp, b, b_comp_end - b));
-
-        if (a_comp < b_comp)
-          return -1;
-        else if (a_comp > b_comp)
-          return 1;
-        else
-          {
-            a_vsn_len -= (a_comp_end - a);
-            b_vsn_len -= (b_comp_end - b);
-
-            if (a_vsn_len && b_vsn_len)
-              {
-                /* both have additional components
-                 */
-                a = a_comp_end + 1;
-                b = b_comp_end + 1;
-                --a_vsn_len;
-                --b_vsn_len;
-              }
-            else if (!a_vsn_len && b_vsn_len)
-              {
-                /* b has more components, a doesn't.
-                 */
-                return -1;
-              }
-            else if (a_vsn_len && !b_vsn_len)
-              {
-                /* a has more components, b doesn't.
-                 */
-                return 1;
-              }
-            else
-              {
-                /* neither has more components
-                 */
-                a = a_comp_end;
-                b = b_comp_end;
-                break;
-              }
-          }
-      }
-
-    a_nxt = over_separator (a);
-    b_nxt = over_separator (b);
-
-    if (!a_nxt && !b_nxt)
-      return 0;
-    else if (a_nxt && !b_nxt)
-      return 1;
-    else if (!a_nxt && b_nxt)
-      return -1;
-    else
-      {
-        a = a_nxt;
-        b = b_nxt;
-      }
-  }
-
+  res = version_cmp_n (a, a_nxt - a, b, b_nxt - b);
 
+  if (res < 0)
+    return -1;
+  else if (res > 0)
+    return 1;
+  
+  a = over_separator (a_nxt);
+  b = over_separator (b_nxt);
+  
+  if (!a && !b)
+    return 0;
+  else if (a && !b)
+    return 1;
+  else if (!a && b)
+    return -1;
 
   /* compare patch level phase
    */
@@ -823,6 +774,94 @@ arch_names_cmp (t_uchar const * a, t_uchar const * b)
   }
 }
 
+static int
+version_cmp_n (const t_uchar * a, size_t a_l, const t_uchar * b, size_t b_l)
+{
+  t_uchar * a_colon;
+  t_uchar * b_colon;
+  t_ulong a_epoch;
+  t_ulong b_epoch;
+  int errn;
+
+  a_colon = str_chr_index_n (a, a_l, ':');
+  if (a_colon) {
+    invariant (!cvt_decimal_to_ulong (&errn, &a_epoch, a, a_colon - a));
+    a_l -= (a_colon - a) - 1;
+    a = a_colon + 1;
+  } else
+    a_epoch = 0;
+
+  b_colon = str_chr_index (b, ':');
+  if (b_colon) {
+    invariant (!cvt_decimal_to_ulong (&errn, &b_epoch, b, b_colon - b));
+    b_l -= (b_colon - b) - 1;
+    b = b_colon + 1;
+  } else
+    b_epoch = 0;
+
+  if (a_epoch < b_epoch)
+    return -1;
+  else if (a_epoch > b_epoch)
+    return 1;
+  else
+    return ver_cmp_n (a, a_l, b, b_l);
+}
+
+static int
+ver_order (t_uchar c)
+{
+  if (c == '~')
+    return -1;
+  else if (char_is_digit (c))
+    return 0;
+  else if (char_is_alpha (c))
+    return c;
+  else
+    return c + 256;
+}
+
+static int
+ver_cmp_n (const t_uchar * a, size_t a_l, const t_uchar * b, size_t b_l)
+{
+  while ((*a && a_l) || (*b && b_l))
+    {
+      int first_diff = 0;
+
+      while ((*a && a_l && !char_is_digit (*a))
+	     || (*b && b_l && !char_is_digit (*b)))
+      {
+	int ac, bc;
+	ac = (*a && a_l) ? ver_order (*a) : 0;
+	bc = (*b && b_l) ? ver_order (*b) : 0;
+	if (ac != bc)
+	  return ac - bc;
+
+	++a, --a_l;
+	++b, --b_l;
+      }
+
+      while ((*a == '0') && a_l) ++a, --a_l;
+      while ((*b == '0') && b_l) ++b, --b_l;
+      while (a_l && char_is_digit (*a) && b_l && char_is_digit (*b))
+      {
+	if (!first_diff)
+	  first_diff = *a - *b;
+
+	++a, --a_l;
+	++b, --b_l;
+      }
+
+      if (a_l && char_is_digit (*a))
+	return 1;
+      else if (b_l && char_is_digit (*b))
+	return -1;
+      else if (first_diff)
+	return first_diff;
+    }
+
+  return 0;
+}
+
 struct arch_sort_spec
 {
   int reverse_p;
@@ -1051,6 +1090,41 @@ arch_analyze_patch_level (t_ulong * n, t_uchar * patch_level)
   return type;
 }
 
+int
+arch_cmp_revision (t_uchar *a, t_uchar *b)
+{
+  enum arch_patch_level_type a_type;
+  enum arch_patch_level_type b_type;
+  t_ulong a_level;
+  t_ulong b_level;
+
+  a_type = arch_analyze_patch_level (&a_level, a);
+  b_type = arch_analyze_patch_level (&b_level, b);
+  if (a_type != b_type)
+    return arch_cmp_patch_level_type (a_type, b_type);
+    
+  else
+    if (a_level < b_level)
+       return -1;
+    if (b_level == a_level) 
+       return 0;
+    else return 1;
+}
+
+
+int
+arch_cmp_patch_level_type (enum arch_patch_level_type a, enum arch_patch_level_type b)
+{
+  if (a == b) 
+    return 0;
+
+  else if (a < b)
+    return -1;
+
+  else
+    return 1;
+}
+
 
 t_uchar *
 arch_form_patch_level (enum arch_patch_level_type type, t_ulong n)
@@ -1197,25 +1271,26 @@ over_separator (t_uchar const * in)
 static t_uchar const *
 over_version (t_uchar const * in)
 {
+  if (!char_is_digit (*in))
+    return 0;
+  ++in;
+
   while (1)
     {
       if (!*in)
-        break;
-      else if (!char_is_digit (*in))
-        break;
-      else
+	return in;
+      else if ((char_is_alnum (*in)) || (*in == '.') || (*in == '+')
+	       || (*in == ':') || (*in == '~'))
+	++in;
+      else if (*in == '-')
         {
-          while (char_is_digit (*in))
-            ++in;
-          if (!*in)
-            return in;
-          else if (*in == '.')
-            ++in;
-          else if ((*in == '-') && (in[1] == '-'))
-            return in;
-          else
-            break;
-        }
+	  if (in[1] == '-')
+	    return in;
+	  else
+	    ++in;
+	}
+      else
+	break;
     }
   return 0;
 }
