@@ -30,36 +30,7 @@
 #include "libarch/project-tree.h"
 #include "libarch/patch-logs.h"
 
-/* these should be provided by ar.h */
-static inline arch_patch_log_dir_t *
-ar_ref_arch_patch_log_dir (arch_patch_log_dir_t **array, int index)
-{
-    return (arch_patch_log_dir_t *)ar_ref ((void **)array,
-                                            lim_use_must_malloc,
-                                            index,
-                                            sizeof (arch_patch_log_dir_t));
-}
-
-static inline int
-ar_size_arch_patch_log_dir (arch_patch_log_dir_t *array)
-{
-    return ar_size (array, 0, sizeof (arch_patch_log_dir_t));
-}
-
-static inline void
-ar_free_arch_patch_log_dir (arch_patch_log_dir_t **array)
-{
-    ar_free ((void **)array, 0);
-    *array = NULL;
-}
-
-static inline arch_patch_log_dir_t *
-ar_pop_arch_patch_log_dir (arch_patch_log_dir_t **array)
-{
-    return (arch_patch_log_dir_t *)ar_pop ((void **)array,
-                                            lim_use_must_malloc,
-                                            sizeof (arch_patch_log_dir_t *));
-}
+AR_TYPEDEF (arch_patch_log_dir_t, arch_patch_log_dir);
 
 t_uchar *
 arch_make_log_file (t_uchar * tree_root, t_uchar * archive, t_uchar * version)
@@ -906,7 +877,12 @@ arch_print_headers_summary (int out_fd, int indent_level, assoc_table headers, i
 
           for (x = 0; x < rel_n_records (s_lines); ++x)
             {
-              s_lines[x][0] = trim_surrounding_ws (s_lines[x][0]);
+	      t_uchar *temp = str_save (0, s_lines[x][0]);
+	      temp = trim_surrounding_ws (temp);
+	      rel_replace_record (s_lines, x, rel_make_record
+				  (temp, NULL));
+	      lim_free (0, temp);
+	      
               safe_printfmt (out_fd, "%*s%s\n", indent_level, "", s_lines[x][0]);
             }
 
