@@ -16,11 +16,15 @@
 #include "hackerlab/char/str.h"
 #include "hackerlab/fmt/cvt.h"
 #include "hackerlab/fs/file-names.h"
+#include "hackerlab/mem/talloc.h"
 #include "hackerlab/vu/safe.h"
 #include "libfsutils/tmp-files.h"
 #include "libfsutils/file-contents.h"
+#include "libarch/authinfo.h"
 #include "libarch/archives.h"
+#include "libarch/debug.h"
 #include "libarch/pfs-dav.h"
+#include "po/gettext.h"
 #include <neon/ne_session.h>
 #include <neon/ne_basic.h>
 #include <neon/ne_request.h>
@@ -48,6 +52,7 @@ struct authinfo
 {
   char *username;
   char *password;
+  char *machine;
 };
 
 struct ls_data
@@ -71,6 +76,7 @@ struct arch_pfs_dav_session
   char * dav_scheme;
   char * dav_hostname;
   int dav_port;
+  struct authinfo * proxy_auth;
 
   ne_session * sess;
   ne_server_capabilities sess_opts;
@@ -78,6 +84,7 @@ struct arch_pfs_dav_session
 
 
 /* __STDC__ prototypes for static functions */
+static void fs_disconnect (struct arch_pfs_session **pfs);
 static t_uchar * pfs_file_contents (struct arch_pfs_session * p, t_uchar * path, int soft_errors);
 static rel_table pfs_directory_files (struct arch_pfs_session * p, t_uchar * path, int soft_errors);
 static int quick_file_exists (struct arch_pfs_session * p, t_uchar * path);
@@ -102,6 +109,7 @@ static void results (void * userdata, const char * uri, const ne_prop_result_set
 
 struct arch_pfs_vtable dav_pfs_fns =
 {
+  fs_disconnect, 
   pfs_file_exists,
   pfs_is_dir,
 
@@ -132,6 +140,48 @@ arch_pfs_dav_supported_protocol (t_uchar * uri)
 }
 
 
+/* Display informations about the certificate, and accept it */
+static int
+ssl_accept(void *userdata, int failures, const ne_ssl_certificate *c)
+{
+  /* (trivial) code taken from cadaver.c */
+  char *tmp, from[NE_SSL_VDATELEN], to[NE_SSL_VDATELEN];
+  const char *ident;
+
+  struct authinfo * auth = userdata;
+
+  ident = ne_ssl_cert_identity(c);
+  
+  if (ident)
+    debug (dbg_dav, 1,
+              _("WARNING: Accepting untrusted server certificate for `%s':\n"),
+              ident);
+  else
+    debug (dbg_dav, 1,
+               _("WARNING: Accepting untrusted server certificate\n"));
+
+  if (failures & NE_SSL_IDMISMATCH)
+    {
+      debug (dbg_dav,1,
+                 _("Certificate was issued to hostname `%s' instead of `%s'.\n"
+                   "This connection may be hijacked!\n"),
+                 ne_ssl_cert_identity(c),
+                 auth->machine);
+    }
+
+#define PRINT_AND_FREE(str, dn) \
+tmp = ne_ssl_readable_dname(dn); safe_printfmt(2, str, tmp); free(tmp)
+  
+  PRINT_AND_FREE("Issued to: %s\n", ne_ssl_cert_subject(c));
+  PRINT_AND_FREE("Issued by: %s\n", ne_ssl_cert_issuer(c));
+  
+  ne_ssl_cert_validity(c, from, to);
+  safe_printfmt(2, "Certificate is valid from %s to %s\n", from, to);
+  
+  return 0;
+}
+
+
 struct arch_pfs_session *
 arch_pfs_dav_connect (t_uchar * uri, int soft_errors)
 {
@@ -153,11 +203,12 @@ arch_pfs_dav_connect (t_uchar * uri, int soft_errors)
     {
       if (soft_errors)
 	  return NULL;
-      panic ("arch_pfs_dav_connect: ne_sock_init() failed.");
+      panic (_("arch_pfs_dav_connect: ne_sock_init() failed."));
     }
 
   answer = (struct arch_pfs_dav_session *)lim_malloc (0, sizeof (*answer));
   mem_set0 ((t_uchar *)answer, sizeof (*answer));
+  answer->proxy_auth = talloc (NULL, struct authinfo);
   answer->pfs.vtable = &dav_pfs_fns;
 
   if (!str_cmp_prefix ("http://", uri))
@@ -177,7 +228,7 @@ arch_pfs_dav_connect (t_uchar * uri, int soft_errors)
 	  lim_free (0, answer);
 	  return NULL;
 	}
-      panic ("bogus uri to arch_pfs_dav_connect");
+      panic (_("bogus uri to arch_pfs_dav_connect"));
     }
 
   root_path = str_chr_index (host, '/');
@@ -190,10 +241,14 @@ arch_pfs_dav_connect (t_uchar * uri, int soft_errors)
     }
 
   hostname = str_chr_index (host, '@');
+  auth = lim_malloc (0, sizeof (struct authinfo));
 
   if (!hostname)
     {
       answer->dav_hostname = str_save (0, host);
+      auth->machine = str_save (0, host);
+      auth->username = str_save(0, "");
+      auth->password = str_save(0, "");
     }
   else
     {
@@ -212,12 +267,15 @@ arch_pfs_dav_connect (t_uchar * uri, int soft_errors)
           auth->password = str_save (0, "");
         }
 
+      /* machine name will be useful to lookup in ~/.authinfo file */
+      auth->machine = str_save (0, hostname + 1);
+
       answer->dav_hostname = str_save (0, hostname + 1);
     }
-
+  
   portstr = str_chr_index (answer->dav_hostname, ':');
-   if (portstr)
-     *(portstr++) = 0;
+  if (portstr)
+    *(portstr++) = 0;
 
   answer->dav_port = ne_uri_defaultport (answer->dav_scheme);
   if (portstr && (0 > cvt_decimal_to_uint (&ign, &answer->dav_port, portstr, str_length (portstr))))
@@ -238,12 +296,11 @@ arch_pfs_dav_connect (t_uchar * uri, int soft_errors)
   if (!str_cmp ("https", answer->dav_scheme))
     {
       ne_ssl_trust_default_ca (answer->sess);
+      /* accept SSL certificate anyway */
+      ne_ssl_set_verify(answer->sess, ssl_accept, auth);
     }
-
-  if (auth)
-    {
-      ne_set_server_auth (answer->sess, dav_client_auth, auth);
-    }
+  
+  ne_set_server_auth (answer->sess, dav_client_auth, auth);
 
   ne_set_useragent (answer->sess, "arch-client/0.1");
 
@@ -296,6 +353,15 @@ arch_pfs_dav_connect (t_uchar * uri, int soft_errors)
 
           ne_session_proxy (answer->sess, proxy_path, proxy_port_num);
         }
+      
+      answer->proxy_auth->machine =
+        talloc_strdup(answer->proxy_auth, proxy_path);
+      answer->proxy_auth->username =
+        talloc_strdup(answer->proxy_auth, "");
+      answer->proxy_auth->password =
+        talloc_strdup(answer->proxy_auth, "");
+
+      ne_set_proxy_auth(answer->sess, dav_client_auth, answer->proxy_auth);
     }
 
   if (root_path)
@@ -308,7 +374,7 @@ arch_pfs_dav_connect (t_uchar * uri, int soft_errors)
 	      lim_free (0, answer);
 	      return NULL;
 	    }
-	  panic ("cannot change DAV cwd.");
+	  panic (_("cannot change DAV cwd."));
 	}
     }
 
@@ -319,6 +385,17 @@ arch_pfs_dav_connect (t_uchar * uri, int soft_errors)
   return (struct arch_pfs_session *)answer;
 }
 
+void 
+fs_disconnect (struct arch_pfs_session **pfs)
+{
+  struct arch_pfs_dav_session * local_pfs = (struct arch_pfs_dav_session *)*pfs;
+  struct authinfo * proxy_auth = local_pfs->proxy_auth;
+  talloc_free (proxy_auth);
+
+  lim_free (0, *pfs);
+  *pfs = NULL;
+}
+
 static t_uchar *
 pfs_file_contents (struct arch_pfs_session * p, t_uchar * path, int soft_errors)
 {
@@ -329,7 +406,7 @@ pfs_file_contents (struct arch_pfs_session * p, t_uchar * path, int soft_errors)
   int ne_err;
   t_uchar * answer = 0;
 
-  tmp_path = tmp_file_name ("/tmp", ",,pfs-dav-file-contents");
+  tmp_path = tmp_file_name_in_tmp (",,pfs-dav-file-contents");
   fd = safe_open (tmp_path, O_RDWR | O_CREAT | O_EXCL, 0000);
   safe_unlink (tmp_path);
 
@@ -350,6 +427,7 @@ pfs_file_contents (struct arch_pfs_session * p, t_uchar * path, int soft_errors)
     }
 
   safe_close (fd);
+  lim_free (0, tmp_path);
   lim_free (0, dav_path);
 
   return answer;
@@ -826,11 +904,35 @@ dav_client_auth (void * userdata, const char * realm, int attempt,
   if (attempt > 1)
     return attempt;
 
+  /* Since we got a void *, we need to cast it. This is an easy way
+   * to do so. */
   auth = userdata;
   len = str_length (auth->username);
-  str_cpy_n (username, auth->username, MIN (len, NE_ABUFSIZ - 1));
+
+  if (len)
+    str_cpy_n (username, auth->username, MIN (len, NE_ABUFSIZ - 1));
+  else
+    {
+      /* User is not set. Try getting it from the authinfo file */
+      t_uchar * get_user = NULL;
+      get_user = arch_auth_username (auth->machine);
+      len = str_length(get_user);
+      str_cpy_n (username, get_user, MIN (len, NE_ABUFSIZ - 1));
+      lim_free (0, get_user);
+    }
+
   len = str_length (auth->password);
-  str_cpy_n (password, auth->password, MIN (len, NE_ABUFSIZ - 1));
+  if (len)
+    str_cpy_n (password, auth->password, MIN (len, NE_ABUFSIZ - 1));
+  else
+    {
+      /* pass not set. try getting it from the authinfo file */
+      t_uchar * get_pass = NULL;
+      get_pass = arch_auth_password (username, auth->machine);
+      len = str_length(get_pass);
+      str_cpy_n (password, get_pass, MIN (len, NE_ABUFSIZ - 1));
+      lim_free (0, get_pass);
+    }
 
   return 0;
 }
