@@ -51,6 +51,7 @@ struct arch_pfs_ftp_session
 
 
 /* __STDC__ prototypes for static functions */
+static void fs_disconnect (struct arch_pfs_session **pfs);
 static int pfs_ftp_file_exists (struct arch_pfs_session * p, t_uchar * path);
 static int pfs_ftp_is_dir (struct arch_pfs_session * p, t_uchar * path);
 static t_uchar * pfs_ftp_file_contents (struct arch_pfs_session * p, t_uchar * path, int soft_errors);
@@ -209,6 +210,7 @@ static int ftp_client_pwd (alloc_limits limits,
 
 struct arch_pfs_vtable ftp_pfs_fns =
 {
+  fs_disconnect, 
   pfs_ftp_file_exists,
   pfs_ftp_is_dir,
 
@@ -240,7 +242,7 @@ arch_pfs_ftp_supported_protocol (t_uchar * uri)
 
 
 struct arch_pfs_session *
-arch_pfs_ftp_connect (t_uchar * uri)
+arch_pfs_ftp_connect (t_uchar * uri, int soft_errors)
 {
   int errn;
   int code = 0;
@@ -341,7 +343,7 @@ arch_pfs_ftp_connect (t_uchar * uri)
             }
           else
             {
-              long dir_len;
+              long dir_len = 0;
 
               if (0 > ftp_client_pwd (0, &errn, &code, &answer->remote_pwd, &dir_len, &text, &text_len, answer->in_fd, answer->out_fd))
                 {
@@ -365,9 +367,17 @@ arch_pfs_ftp_connect (t_uchar * uri)
 
   if (status)
     {
-      safe_printfmt (2, "ftp_client_connect: ftp error connecting to %s\n", hostname);
-      safe_printfmt (2, "    %d: %.*s\n", code, (int)text_len, text);
-      exit (2);
+      if (!soft_errors)
+	{
+	  safe_printfmt (2, "ftp_client_connect: ftp error connecting to %s\n", hostname);
+	  safe_printfmt (2, "    %d: %.*s\n", code, (int)text_len, text);
+	  exit (2);
+	}
+      /* FIXME: disconnect or otherwise free
+       * infd, outfd etc etc
+       */
+      lim_free (0, answer);
+      answer = NULL;
     }
 
   lim_free (0, hostname);
@@ -375,6 +385,14 @@ arch_pfs_ftp_connect (t_uchar * uri)
   return (struct arch_pfs_session *)answer;
 }
 
+void 
+fs_disconnect (struct arch_pfs_session **pfs)
+{
+    /* FIXME: close stuff yadayadayada */
+    lim_free (0, *pfs);
+    *pfs = NULL;
+}
+
 
 static int
 pfs_ftp_file_exists (struct arch_pfs_session * p, t_uchar * path)
@@ -660,11 +678,10 @@ pfs_ftp_directory_files (struct arch_pfs_session * p, t_uchar * path, int soft_e
 
                   for (x = 0; x < rel_n_records (answer); ++x)
                     {
-                      t_uchar * t = 0;
-
-                      t = file_name_tail (0, answer[x][0]);
-                      lim_free (0, answer[x][0]);
-                      answer[x][0] = t;
+                      t_uchar * t = file_name_tail (0, answer[x][0]);
+		      rel_replace_record (answer, x, rel_make_record
+					  (t, NULL));
+		      lim_free (0, t);
                     }
                 }
             }
@@ -797,9 +814,9 @@ pfs_ftp_rename (struct arch_pfs_session * p, t_uchar ** errstr, t_uchar * from,
 {
   struct arch_pfs_ftp_session * arch = (struct arch_pfs_ftp_session *)p;
   int errn;
-  int code;
+  int code = 0;
   t_uchar * text = 0;
-  long text_len;
+  long text_len = 0;
   int status;
   t_uchar * from_path = 0;
   t_uchar * to_path = 0;
@@ -1111,6 +1128,19 @@ ftp_client_login (alloc_limits limits,
  * This scheme assumes that multi-line replies may not be nested.
  */
 
+static void
+handle_reply_data (t_uchar **text, alloc_limits const limits,
+		    int const len, long * text_len,
+		    t_uchar const * const line)
+{
+    if (text)
+      {
+	*text = lim_realloc (limits, *text, len + *text_len + 1);
+	mem_move (*text + *text_len, line, len);
+	(*text)[*text_len + len] = 0;
+	*text_len = len + *text_len;
+      }
+}
 
 static int
 ftp_client_read_reply (alloc_limits limits,
@@ -1170,7 +1200,7 @@ ftp_client_read_reply (alloc_limits limits,
           return -1;
         }
 
-      if (   (len >= 4)
+      if ((len >= 4)
           && (char_is_digit(line[0]) && c0)
           && (char_is_digit(line[1]) && c1)
           && (char_is_digit(line[2]) && c2)
@@ -1178,24 +1208,21 @@ ftp_client_read_reply (alloc_limits limits,
         {
           /* Got the last line of the reply.
            */
-          if (text)
-            {
-              *text = lim_realloc (limits, *text, len + *text_len + 1);
-              mem_move (*text + *text_len, line, len);
-              (*text)[*text_len + len] = 0;
-              *text_len = len + *text_len;
-            }
+	  handle_reply_data (text, limits, len, text_len, line);
           return 0;
         }
+      else if (len == 0)
+	{
+	  /* bad ftp server hasn't finished the reply, but is looking like EOF 
+	   * observed with ftp.ece.nwu.edu 
+	   * RBC 20050316
+	   */
+	  handle_reply_data (text, limits, len, text_len, line);
+	  return 0;
+	}
       else
         {
-          if (text)
-            {
-              *text = lim_realloc (limits, *text, len + *text_len + 1);
-              mem_move (*text + *text_len, line, len);
-              (*text)[*text_len + len] = 0;
-              *text_len = len + *text_len;
-            }
+	  handle_reply_data (text, limits, len, text_len, line);
         }
     }
 
@@ -1522,8 +1549,8 @@ ftp_client_pasv (alloc_limits limits,
     }
   else
     {
-      t_ulong h;
-      t_uint16 p;
+      t_ulong h = 0;
+      t_uint16 p = 0;
 
       if (str_chr_rindex (*text, '('))
       {
