@@ -357,35 +357,35 @@ arch_pfs_sftp_connect (t_uchar * uri, int soft_errors)
     }
 
   ssh_argv = 0;
-  *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = cfg__ssh;
+  ar_push_uchar_star (&ssh_argv, cfg__ssh);
 
   if (port)
     {
-      *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = "-p";
-      *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = port;
+      ar_push_uchar_star (&ssh_argv, "-p");
+      ar_push_uchar_star (&ssh_argv, port);
     }
 
   if (user)
     {
-      *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = "-l";
-      *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = user;
+      ar_push_uchar_star (&ssh_argv, "-l");
+      ar_push_uchar_star (&ssh_argv, user);
     }
 
 #if cfg__ssh_is_lsh
-  *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = "--no-x11-forward";
-  *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = "--subsystem=sftp";
-  *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = hostname;
+  ar_push_uchar_star (&ssh_argv, "--no-x11-forward");
+  ar_push_uchar_star (&ssh_argv, "--subsystem=sftp");
+  ar_push_uchar_star (&ssh_argv, hostname);
 
 #elif cfg__ssh_is_openssh
-  *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = "-oFallBackToRsh=no";
-  *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = "-oForwardX11=no";
-  *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = "-oForwardAgent=no";
-  /*  *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = "-oClearAllForwardings=yes"; */
-  *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = "-2";
-  *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = "-s";
-  *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = hostname;
-  *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = "sftp";
-  *(t_uchar **)ar_push ((void **)&ssh_argv, 0, sizeof (t_uchar *)) = 0;
+  ar_push_uchar_star (&ssh_argv, "-oFallBackToRsh=no");
+  ar_push_uchar_star (&ssh_argv, "-oForwardX11=no");
+  ar_push_uchar_star (&ssh_argv, "-oForwardAgent=no");
+  /*  ar_push_uchar_star (&ssh_argv, "-oClearAllForwardings=yes"); */
+  ar_push_uchar_star (&ssh_argv, "-2");
+  ar_push_uchar_star (&ssh_argv, "-s");
+  ar_push_uchar_star (&ssh_argv, hostname);
+  ar_push_uchar_star (&ssh_argv, "sftp");
+  ar_push_uchar_star (&ssh_argv, 0);
 
 #else
   panic ("No ssh flavor was specified at compile time!");
@@ -816,7 +816,7 @@ pfs_file_contents (struct arch_pfs_session * p, t_uchar * path, int soft_errors)
   int fd;
   t_uchar * answer = 0;
 
-  tmp_path = tmp_file_name ("/tmp", ",,pfs-sftp-file-contents");
+  tmp_path = tmp_file_name_in_tmp (",,pfs-sftp-file-contents");
   fd = safe_open (tmp_path, O_RDWR | O_CREAT | O_EXCL, 0000);
   safe_unlink (tmp_path);
 
@@ -830,6 +830,8 @@ pfs_file_contents (struct arch_pfs_session * p, t_uchar * path, int soft_errors)
   answer = fd_contents (fd);
   safe_close (fd);
 
+  lim_free(0, tmp_path);
+
   return answer;
 }
 
@@ -1472,72 +1474,40 @@ dirfold (t_uchar *dir)
   return dir;
 }
 
-extern int arch_pfs_sftp_parse_uri (t_uchar ** user, t_uchar ** hostname, t_uchar ** port, char ** path, t_uchar * uri)
+/**
+ * \brief parse a uri into components
+ * \return 0 on success
+ */
+int
+arch_pfs_sftp_parse_uri (t_uchar ** user, t_uchar ** hostname, t_uchar ** port, char ** path, t_uchar * uri)
 {
-  t_uchar * start;
-  t_uchar * stop;
-  t_uchar * host_part = 0;
-  t_uchar * at;
-  t_uchar * h0;
-  t_uchar * h1;
-  t_uchar * colon;
+  ne_uri parsed_uri;
 
-  /* [user@]hostname[:port]/path
-   * ^    ^ ^       ^      ^
-   * |    | |       |      |
-   * |   at?|       |      stop[*]
-   * |      |       |	     0-byte of host_part
-   * |     h0       |
-   * |              |
-   * |              |
-   * |              |
-   * start[*]     colon?
-   * host_part    h1
-   * 
-   * 
-   * start and stop point to one copy of the string
-   * (the original uri which includes the /path).
-   * 
-   * host_part points to another copy of the string 
-   * (local to this block) which all the other pointers
-   * point into.   
-   * 
-   * at? and colon? are 0 if @ and : are not present.
-   * 
-   * The reason to make host_part is so that str_chr_index 
-   * searches are confined to just the host_part.
+  /* [authino@]host[:port]/path */
+
+  if (ne_uri_parse (uri, &parsed_uri) != 0)
+      return -1;
+  
+  /* can only be called for sftp uris */
+  invariant (!str_cmp(parsed_uri.scheme, "sftp"));
+  /* handle non nrmalised uris, with whacky user input. 
+   * RBC 20050304 FIXME: when we have the new registration syste,
+   * we must move this call to the uri loading code
    */
+  arch_uri_heuristics (&parsed_uri);
 
+  *user = str_save (0, parsed_uri.authinfo);
+  *hostname = str_save (0, parsed_uri.host);
+  if (parsed_uri.port)
+    {
+      t_uchar buf[64];
+      
+      cvt_long_to_decimal (buf, parsed_uri.port);
+      *port = str_save (0, buf);
+    }
+  *path = str_save (0, parsed_uri.path);
 
-  start = uri + sizeof ("sftp://") - 1;
-  stop = str_chr_index (start, '/');
-  if (!stop)
-    return -1;
-  *path = str_save (0, stop);
-
-  host_part = str_save_n (0, start, stop - start);
-
-  at = str_chr_index (host_part, '@');
-  if (at)
-    h0 = at + 1;
-  else
-    h0 = host_part;
-
-  colon = str_chr_index (h0, ':');
-  if (colon)
-    h1 = colon;
-  else
-    h1 = h0 + str_length (h0);
-
-  if (at)
-    *user = str_save_n (0, host_part, at - host_part);
-
-  if (colon)
-    *port = str_save (0, colon + 1);
-
-  *hostname = str_save_n (0, h0, h1 - h0);
-
-  lim_free (0, host_part);
+  ne_uri_free(&parsed_uri);
   return 0;
 }
 
