@@ -22,14 +22,6 @@
 
 
 /* local methods */
-/* FIXME: move to hackerlab ? */
-static void
-invariant_str_cmp(t_uchar const *left, t_uchar const *right)
-{
-    if (str_cmp (left, right))
-	safe_printfmt (2, "assertion failed: '%s' != '%s'\n", left, right);
-}
-
 static void 
 check_namespace_parse (t_uchar const *input, enum arch_parse_package_name_type type, t_uchar const *output)
 {
