@@ -12,16 +12,22 @@
 
 
 
+/** 
+ * \brief convert a rel_table to an assoc_table
+ * \param table a table to convert
+ * \param key_column the column in the rel table to be key into the assoc
+ * \param value_column the column in the rel table to be found by lookups in the table
+ */
 assoc_table
-rel_to_assoc (rel_table t, int k, int v)
+rel_to_assoc (rel_table table, int key_column, int value_column)
 {
   int lim;
   int x;
   assoc_table answer = 0;
 
-  lim = rel_n_records (t);
+  lim = rel_n_records (table);
   for (x = 0; x < lim; ++x)
-    assoc_set (&answer, t[x][k], t[x][v]);
+    assoc_set (&answer, table[x][key_column], table[x][value_column]);
 
   return answer;
 }
