@@ -1,5 +1,6 @@
 /* relational.c:
  *
+ * vim:smartindent ts=8:sts=2:sta:et:ai:shiftwidth=2
  ****************************************************************
  * Copyright (C) 2003 Tom Lord
  *
@@ -8,17 +9,27 @@
  */
 
 
+#include "hackerlab/bugs/panic.h"
+#include "hackerlab/bugs/exception.h"
+#include "hackerlab/os/errno.h"
 #include "hackerlab/os/stdarg.h"
 #include "hackerlab/vu/safe.h"
 #include "hackerlab/vu/safe-vu-utils-vfdbuf.h"
-#include "hackerlab/arrays/ar.h"
+#include "hackerlab/mem/talloc.h"
 #include "hackerlab/sort/qsort.h"
 #include "hackerlab/char/char-class.h"
 #include "hackerlab/char/str.h"
 #include "hackerlab/char/pika-escaping-utils.h"
+#define _IN_RELATIONAL_
 #include "libawk/relational.h"
 
 
+AR_TYPEDEF (int, rel_cut_spec);
+AR_TYPEDEF (rel_field, rel_record);
+AR_TYPEDEF (rel_record, rel_table);
+AR_TYPEDEF (struct rel_join_output_spec, rel_join_output_spec);
+AR_TYPEDEF (struct rel_join_output_spec *, rel_join_output_specs);
+
 /* __STDC__ prototypes for static functions */
 static int rec_cmp_by_field (void * va, void * vb, void * vdata);
 static int rec_cmp_by_field_fn (void * va, void * vb, void * vdata);
@@ -26,11 +37,13 @@ static int rec_cmp_by_fields (void * va, void * vb, void * vdata);
 static rel_record rel_read_record (int fd,
                                    int n_fields,
                                    char * err_name,
-                                   char * err_src);
+                                   char * err_src,
+                                   char const * name);
 static rel_record rel_read_pika_unescape_iso8859_1_record (int fd,
                                                            int n_fields,
                                                            char * err_name,
-                                                           char * err_src);
+                                                           char * err_src,
+                                                           char const * name);
 static void rel_print_record (int fd, rel_record rec);
 static void rel_print_pika_escape_iso8859_1_record (int fd, int escape_classes, rel_record rec);
 static void rel_print_record_sp (int fd, rel_record rec);
@@ -100,7 +113,7 @@ static void rel_print_pika_escape_iso8859_1_record_sp (int fd, int escape_classe
 int
 rel_n_records (rel_table r)
 {
-  return ar_size ((void *)r, 0, sizeof (rel_record));
+  return ar_size_rel_table (r);
 }
 
 
@@ -112,7 +125,7 @@ rel_n_records (rel_table r)
 int
 rel_n_fields (rel_record r)
 {
-  return ar_size ((void *)r, 0, sizeof (rel_field));
+  return ar_size_rel_record (r);
 }
 
 
@@ -125,7 +138,7 @@ rel_n_fields (rel_record r)
 
 
 /*(c rel_make_record)
- * rel_record rel_make_record (t_uchar * field0, ...);
+ * rel_record rel_make_record (t_uchar const * const field0, ...);
  * 
  * Allocate a new record containing the indicated fields.
  * 
@@ -137,10 +150,10 @@ rel_n_fields (rel_record r)
  * which creates a record with two fields.
  * 
  * Note that this function allocates private copies of the fields.
- * They will be freed by `rel_free_record' or `rel_free_table'.
+ * They will be freed by `ar_free_rel_record' or `rel_free_table'.
  */
 rel_record
-rel_make_record (t_uchar * field0, ...)
+_rel_make_record (char const * name, t_uchar const * const field0, ...)
 {
   va_list fp;
   rel_record answer;
@@ -149,7 +162,7 @@ rel_make_record (t_uchar * field0, ...)
     return 0;
 
   answer = 0;
-  *(rel_field *)ar_push ((void **)&answer, 0, sizeof (rel_field)) = str_save (0, field0);
+  _rel_add_field (&answer, field0, name);
   va_start (fp, field0);
   while (1)
     {
@@ -158,12 +171,18 @@ rel_make_record (t_uchar * field0, ...)
       contents = va_arg (fp, t_uchar *);
       if (!contents)
         break;
-      *(rel_field *)ar_push ((void **)&answer, 0, sizeof (rel_field)) = str_save (0, contents);
+      rel_add_field (&answer, contents);
     }
   va_end (fp);
   return answer;
 }
 
+void
+_rel_add_record (rel_table * table, rel_record r, char const * name)
+{
+    ar_push_ext_rel_table (table, r, name);
+    talloc_steal (ar_base (*table), ar_base (r));
+}
 
 /*(c rel_add_records)
  * void rel_add_records (rel_table * table, ...);
@@ -198,22 +217,91 @@ rel_make_record (t_uchar * field0, ...)
  *                         0);
  */
 void
-rel_add_records (rel_table * table, ...)
+_rel_add_records (rel_table * table, char const * name, ...)
 {
   va_list rp;
   rel_record r;
 
-  va_start (rp, table);
+  va_start (rp, name);
+  for (r = va_arg (rp, rel_record); r; r = va_arg (rp, rel_record))
+      _rel_add_record (table, r, name);
+  va_end (rp);
+}
+
+/** 
+ * \brief insert records into a rel_table. 
+ *
+ * if the index is not <= rel_n_records table, an assertion will trigger
+ * \param table the table to insert into
+ * \param index the line to insert at
+ * \param record, ...
+ * \return void
+ */
+void 
+rel_insert_records (rel_table *table, int index, ...)
+{
+  va_list rp;
+  rel_record r;
+
+  invariant (index <= rel_n_records (*table));
+  va_start (rp, index);
   for (r = va_arg (rp, rel_record); r; r = va_arg (rp, rel_record))
     {
-      *(rel_record *)ar_push ((void **)table, 0, sizeof (rel_record)) = r;
+      ar_insert_rel_table (table, index, r);
     }
   va_end (rp);
 }
 
+/**
+ * \brief replace a record in a rel table
+ * \param table the table to replace in
+ * \param index the line to replace
+ * \param record the line to replace with
+ * \return void
+ */
+void
+rel_replace_record (rel_table table, int index, rel_record record)
+{
+  ar_free_rel_record (&table[index]);
+  table[index] = record;
+  talloc_steal (ar_base (table), ar_base (record));
+}
+
+/**
+ * \brief remove records from a rel_table.
+ *
+ * \param table the table to remove from
+ * \param from_index the first record to remove
+ * \param to_index the last record to remove
+ * \return void
+ */
+void
+rel_remove_records (rel_table *table, int from_index, int to_index)
+{
+    int index;
+    if (from_index > rel_n_records (*table) - 1)
+        from_index = rel_n_records (*table) - 1;
+    if (to_index > rel_n_records (*table) - 1)
+        to_index = rel_n_records (*table) - 1;
+    if (from_index > to_index)
+        to_index = from_index;
+    if (from_index < 0)
+        return;
+    /* free resources */
+    for (index = from_index; index < to_index + 1; ++index)
+      {
+        ar_free_rel_record (&(*table)[index]);
+      }
+    /* now shrink the table */
+    for (index = from_index; index < to_index + 1; ++index)
+      {
+        ar_remove_rel_table (table, from_index);
+      }
+}
+
 
 /*(c rel_add_field)
- * void rel_add_field (rel_record * r, t_uchar * field);
+ * void rel_add_field (rel_record * r, t_uchar const * field);
  * 
  * Append a single field (the string `field') to the record `*r'.
  * 
@@ -223,28 +311,34 @@ rel_add_records (rel_table * table, ...)
  * `*r' will be update.
  */
 void
-rel_add_field (rel_record * r, t_uchar * field)
+_rel_add_field (rel_record * r, t_uchar const * field, char const * name)
 {
-  *(t_uchar **)ar_push ((void **)r, 0, sizeof (t_uchar *)) = str_save (0, field);
+  if (*r)
+      ar_push_rel_record (r, talloc_strdup (ar_base(*r), field));
+  else
+    {
+      ar_push_ext_rel_record (r, talloc_strdup (NULL, field), name);
+      talloc_steal (ar_base (*r), (*r)[ar_size_rel_record (*r) - 1]);
+    }
 }
 
 
 /*(c rel_singleton_record_n)
- * rel_record rel_singleton_record_n (t_uchar * start, size_t len);
+ * rel_record rel_singleton_record_n (t_uchar const * start, size_t len);
  * 
  * Create a new record containing a single field which is a copy of
  * the `len' characters beginning at `start' with a final 0 appended.
  */
 rel_record
-rel_singleton_record_n (t_uchar * start, size_t len)
+rel_singleton_record_n (t_uchar const * start, size_t len)
 {
   rel_record answer = 0;
 
-  *(rel_field *)ar_push ((void **)&answer, 0, sizeof (rel_field)) = str_save_n (0, start, len);
+  ar_push_rel_record (&answer, talloc_strndup (NULL, start, len));
+  talloc_steal (ar_base (answer), answer[0]);
   return answer;
 }
 
-
 
 /************************************************************************
  *(h1 "Freeing Records and Tables")
@@ -260,41 +354,14 @@ rel_singleton_record_n (t_uchar * start, size_t len)
  * Free the entire table `t'. 
  * 
  * This function will also free all records which are part
- * of `t' -- there is no need to separately call `rel_free_record'.
+ * of `t' -- there is no need to separately call `ar_free_rel_record'.
  */
 void
 rel_free_table (rel_table t)
 {
-  int lim;
-  int x;
-
-  lim = rel_n_records (t);
-  for (x = 0; x < lim; ++x)
-    rel_free_record (t[x]);
-
-  ar_free ((void **)&t, 0);
+  ar_free_rel_table (&t);
 }
 
-
-/*(c rel_free_record)
- * void rel_free_record (rel_record r);
- * 
- * Free record `r'.   This function will also individually free
- * the fields of record `r'.
- */
-void
-rel_free_record (rel_record r)
-{
-  int lim;
-  int x;
-
-  lim = rel_n_fields (r);
-  for (x = 0; x < lim; ++x)
-    lim_free (0, r[x]);
-  ar_free ((void **)&r, 0);
-}
-
-
 
 /************************************************************************
  *(h1 "Parsing Tables from Strings")
@@ -373,21 +440,21 @@ split_is_space (void const *unused, char ch)
 }
 
 /*(c rel_ws_split)
- * rel_table rel_ws_split (t_uchar * string);
+ * rel_table rel_ws_split (t_uchar const * string);
  * 
  * Allocate and return a new table formed by parsing rows,
  * each containing a single field, from `string'.   Rows 
  * are separated by arbitrary whitespace.
  */
 rel_table
-rel_ws_split (t_uchar * string)
+rel_ws_split (t_uchar const * string)
 {
   return rel_callback_split (string, split_is_space, NULL);
 }
 
 
 /*(c rel_nl_split)
- * rel_table rel_nl_split (t_uchar * string);
+ * rel_table rel_nl_split (t_uchar const * string);
  * 
  * Allocate and return a new table formed by parsing rows,
  * each containing a single field, from `string'.   Rows 
@@ -396,11 +463,11 @@ rel_ws_split (t_uchar * string)
  * into the callback method above, which would strip empty lines.
  */
 rel_table
-rel_nl_split (t_uchar * string)
+rel_nl_split (t_uchar const * string)
 {
   rel_table answer = 0;
-  t_uchar * start;
-  t_uchar * end;
+  t_uchar const * start;
+  t_uchar const * end;
 
   if (!string)
     return 0;
@@ -426,6 +493,35 @@ rel_nl_split (t_uchar * string)
     }
 }
 
+/**
+ * \brief convert a string of whitespace separated tokens to a two column table
+ * Throws EINVAL if the token count is not even.
+ * \param string the input
+ * \return rel_table the result.
+ */
+rel_table 
+rel_ws_split_pairs (t_uchar const * string)
+{
+  rel_table answer = NULL;
+  rel_table singles = NULL;
+  int index = 0;
+  int lim = 0;
+
+  singles = rel_ws_split (string);
+  lim = rel_n_records (singles);
+  /* Make sure there is an even number of entries. */
+  if (lim % 2)
+      Throw (exception (EINVAL, "rel_ws_split_pairs: string has an uneven number of tokens"));
+
+  for (index = 0; index < lim / 2; ++index)
+      rel_add_records (&answer, 
+                       rel_make_record (singles[index][0], singles[index + 1][0], 0),
+                       0);
+
+  rel_free_table (singles);
+  return answer;
+}
+
 
 
 /************************************************************************
@@ -443,7 +539,7 @@ rel_nl_split (t_uchar * string)
  * Return a freshly allocated copy of table `t'.
  */
 rel_table
-rel_copy_table (rel_table t)
+_rel_copy_table (rel_table t, char const * name)
 {
   rel_table answer;
   int records;
@@ -452,9 +548,12 @@ rel_copy_table (rel_table t)
   records = rel_n_records (t);
 
   answer = 0;
-  ar_setsize ((void **)&answer, 0, records, sizeof (rel_record));
+  ar_setsize_ext_rel_table (&answer, records, name);
   for (r = 0; r < records; ++r)
-    answer[r] = rel_copy_record (t[r]);
+    {
+      answer[r] = _rel_copy_record (t[r], name);
+      talloc_steal (ar_base (answer), ar_base (answer[r]));
+    }
 
   return answer;
 }
@@ -466,7 +565,7 @@ rel_copy_table (rel_table t)
  * Return a freshly allocated copy of record `r'.
  */
 rel_record
-rel_copy_record (rel_record r)
+_rel_copy_record (rel_record r, char const * name)
 {
   rel_record answer;
   int fields;
@@ -475,9 +574,9 @@ rel_copy_record (rel_record r)
   fields = rel_n_fields (r);
 
   answer = 0;
-  ar_setsize ((void **)&answer, 0, fields, sizeof (rel_field));
+  ar_setsize_ext_rel_record (&answer, fields, name);
   for (f = 0; f < fields; ++f)
-    answer[f] = str_save (0, r[f]);
+    answer[f] = talloc_reference (ar_base (answer), r[f]);
 
   return answer;
 }
@@ -503,7 +602,7 @@ rel_copy_record (rel_record r)
  * table.
  */
 void
-rel_append_x (rel_table * out, rel_table t)
+_rel_append_x (rel_table * out, rel_table t, char const * name)
 {
   int lim;
   int x;
@@ -511,9 +610,7 @@ rel_append_x (rel_table * out, rel_table t)
   lim = rel_n_records (t);
 
   for (x = 0; x < lim; ++x)
-    {
-      *(rel_record *)ar_push ((void **)out, 0, sizeof (rel_record)) = rel_copy_record (t[x]);
-    }
+      _rel_add_record (out, rel_copy_record (t[x]), name);
 }
 
 
@@ -718,6 +815,7 @@ rel_sort_table_by_fields (int reverse_p,
  * function (but ^note^ that, at the moment, the table is 
  * simply space-leaked).
  */
+AR_TYPEDEF(int, legint);
 int *
 rel_sort_fields (int f, ...)
 {
@@ -725,13 +823,13 @@ rel_sort_fields (int f, ...)
   int * answer;
 
   answer = 0;
-  *(int *)ar_push ((void **)&answer, 0, sizeof (int)) = f;
+  ar_push_legint (&answer, f);
 
   va_start (fp, f);
   while (1)
     {
       f = va_arg (fp, int);
-      *(int *)ar_push ((void **)&answer, 0, sizeof (int)) = f;
+      ar_push_legint (&answer, f);
       if (f < 0)
         break;
     }
@@ -794,8 +892,8 @@ rec_cmp_by_fields (void * va, void * vb, void * vdata)
  * relocated table.
  */
 void
-rel_uniq_by_field (rel_table * table,
-                   int field)
+_rel_uniq_by_field (rel_table * table,
+                   int field, char const * name)
 {
   int lim;
   int dest;
@@ -808,11 +906,11 @@ rel_uniq_by_field (rel_table * table,
 
       while ((src < (lim - 1)) && !str_cmp ((*table)[dest][field], (*table)[src + 1][field]))
         {
-          rel_free_record ((*table)[src + 1]);
+          ar_free_rel_record (&(*table)[src + 1]);
           ++src;
         }
     }
-  ar_setsize ((void **)table, 0, dest, sizeof (rel_record));
+  ar_setsize_ext_rel_table (table, dest, name);
 }
 
 
@@ -834,7 +932,7 @@ rel_uniq_by_field (rel_table * table,
  * `fields' should be a value returned by `rel_cut_list' (see below).
  */
 rel_table
-rel_cut (rel_cut_spec fields, rel_table t)
+_rel_cut (rel_cut_spec fields, rel_table t, char const * name)
 {
   rel_table answer;
   int lim;
@@ -842,11 +940,9 @@ rel_cut (rel_cut_spec fields, rel_table t)
 
   answer = 0;
 
-  lim = ar_size ((void *)t, 0, sizeof (rel_record));
+  lim = ar_size_rel_table (t);
   for (x = 0; x < lim; ++x)
-    {
-      rel_add_records (&answer, rel_cut_record (fields, t[x]), 0);
-    }
+      _rel_add_record (&answer, rel_cut_record (fields, t[x]), name);
 
   return answer;
 }
@@ -869,7 +965,7 @@ rel_cut_record (rel_cut_spec fields, rel_record r)
   answer = 0;
   for (x = 0; fields[x] >= 0; ++x)
     {
-      *(t_uchar **)ar_push ((void **)&answer, 0, sizeof (t_uchar *)) = str_save (0, r[fields[x]]);
+      rel_add_field (&answer, r[fields[x]]);
     }
   return answer;
 }
@@ -895,13 +991,13 @@ rel_cut_list (int field, ...)
   rel_cut_spec answer;
 
   answer = 0;
-  *(int *)ar_push ((void **)&answer, 0, sizeof (int)) = field;
+  ar_push_rel_cut_spec (&answer, field);
 
   va_start (fp, field);
   while (1)
     {
       field = va_arg (fp, int);
-      *(int *)ar_push ((void **)&answer, 0, sizeof (int)) = field;
+      ar_push_rel_cut_spec (&answer, field);
       if (field < 0)
         break;
     }
@@ -909,6 +1005,11 @@ rel_cut_list (int field, ...)
   return answer;
 }
 
+void
+rel_cut_spec_finalise (rel_cut_spec *spec)
+{
+    ar_free_rel_cut_spec (spec);
+}
 
 
 
@@ -946,12 +1047,13 @@ rel_cut_list (int field, ...)
  * See `rel_join_output' below.
  */
 rel_table
-rel_join (int absence_table,
+_rel_join (int absence_table,
           struct rel_join_output_spec * output,
           int table1_field,
           int table2_field,
           rel_table table1,
-          rel_table table2)
+          rel_table table2,
+          char const * name)
 {
   int f1_len;
   int f2_len;
@@ -1005,9 +1107,9 @@ rel_join (int absence_table,
           f2_rec = ((f2_pos < f2_len) ? table2[f2_pos] : 0);
           for (x = 0; x < n_output_fields; ++x)
             {
-              *(t_uchar **)ar_push ((void **)&r, 0, sizeof (char *)) = str_save (0, ((output[x].table == 1) ? f1_rec : f2_rec)[output[x].field]);
+              _rel_add_field (&r, ((output[x].table == 1) ? f1_rec : f2_rec)[output[x].field], name);
             }
-          *(rel_record *)ar_push ((void **)&answer, 0, sizeof (rel_record)) = r;
+          _rel_add_record (&answer, r, name);
         }
 
       if ((f1_pos < f1_len) && (cmp <= 0))
@@ -1033,7 +1135,8 @@ rel_join (int absence_table,
  * Each pair names a `table' (1 or 2) and a `field' (0 based).
  */
 struct rel_join_output_spec *
-rel_join_output (int table,
+_rel_join_output (char const * name,
+                 int table,
                  int field, ...)
 {
   va_list ap;
@@ -1045,7 +1148,7 @@ rel_join_output (int table,
 
   answer = 0;
 
-  for (x = 0; !answer && x < ar_size (cache, 0, sizeof (struct rel_join_output_spec *)); ++x)
+  for (x = 0; !answer && x < ar_size_rel_join_output_specs (cache); ++x)
     {
       item = cache[x];
 
@@ -1078,9 +1181,11 @@ rel_join_output (int table,
     
   if (!answer)
     {
-      item = (struct rel_join_output_spec *)ar_push ((void **)&answer, 0, sizeof (struct rel_join_output_spec));
-      item->table = table;
-      item->field = field;
+      struct rel_join_output_spec item;
+      item.table = table;
+      item.field = field;
+      
+      ar_push_ext_rel_join_output_spec (&answer, item, name);
 
       va_start (ap, field);
       while (1)
@@ -1090,17 +1195,17 @@ rel_join_output (int table,
             break;
           field = va_arg (ap, int);
 
-          item = (struct rel_join_output_spec *)ar_push ((void **)&answer, 0, sizeof (struct rel_join_output_spec));
-          item->table = table;
-          item->field = field;
+          item.table = table;
+          item.field = field;
+          ar_push_ext_rel_join_output_spec (&answer, item, name);
         }
       va_end (ap);
 
-      item = (struct rel_join_output_spec *)ar_push ((void **)&answer, 0, sizeof (struct rel_join_output_spec));
-      item->table = -1;
-      item->field = -1;
+      item.table = -1;
+      item.field = -1;
+      ar_push_ext_rel_join_output_spec (&answer, item, name);
 
-      *(struct rel_join_output_spec **)ar_push ((void **)&cache, 0, sizeof (struct rel_join_output_spec *)) = answer;
+      ar_push_ext_rel_join_output_specs (&cache, answer, name);
     }
 
   return answer;
@@ -1132,10 +1237,11 @@ rel_join_output (int table,
  * and exit with status 2.
  */
 rel_table
-rel_read_table (int fd,
+_rel_read_table (int fd,
                 int n_fields,
                 char * err_name,
-                char * err_src)
+                char * err_src,
+                char const * name)
 {
   rel_record rec;
   rel_table answer;
@@ -1143,10 +1249,10 @@ rel_read_table (int fd,
   answer = 0;
   while (1)
     {
-      rec = rel_read_record (fd, n_fields, err_name, err_src);
+      rec = rel_read_record (fd, n_fields, err_name, err_src, name);
       if (!rec)
         break;
-      *(rel_record *)ar_push ((void **)&answer, 0, sizeof (rel_record)) = rec;
+      _rel_add_record (&answer, rec, name);
     }
   return answer;
 }
@@ -1156,7 +1262,8 @@ static rel_record
 rel_read_record (int fd,
                  int n_fields,
                  char * err_name,
-                 char * err_src)
+                 char * err_src,
+                 char const * name)
 {
   t_uchar * line;
   long len;
@@ -1169,7 +1276,7 @@ rel_read_record (int fd,
     return 0;
 
   answer = 0;
-  ar_setsize ((void **)&answer, 0, n_fields, sizeof (rel_field));
+  ar_setsize_ext_rel_record (&answer, n_fields, name);
 
   pos = line;
   for (f = 0; f < n_fields; ++f)
@@ -1191,7 +1298,7 @@ rel_read_record (int fd,
             }
         }
 
-      answer[f] = str_save_n (0, line, pos - line);
+      answer[f] = talloc_strndup (ar_base(answer), line, pos - line);
 
       while (len && char_is_space (*pos))
         {
@@ -1222,10 +1329,11 @@ rel_read_record (int fd,
  * and exit with status 2.
  */
 rel_table
-rel_read_pika_unescape_iso8859_1_table (int fd,
+_rel_read_pika_unescape_iso8859_1_table (int fd,
                                         int n_fields,
                                         char * err_name,
-                                        char * err_src)
+                                        char * err_src,
+                                        char const * name)
 {
   rel_record rec;
   rel_table answer;
@@ -1233,10 +1341,10 @@ rel_read_pika_unescape_iso8859_1_table (int fd,
   answer = 0;
   while (1)
     {
-      rec = rel_read_pika_unescape_iso8859_1_record (fd, n_fields, err_name, err_src);
+      rec = rel_read_pika_unescape_iso8859_1_record (fd, n_fields, err_name, err_src, name);
       if (!rec)
         break;
-      *(rel_record *)ar_push ((void **)&answer, 0, sizeof (rel_record)) = rec;
+      _rel_add_record (&answer, rec, name);
     }
   return answer;
 }
@@ -1246,7 +1354,8 @@ static rel_record
 rel_read_pika_unescape_iso8859_1_record (int fd,
                                          int n_fields,
                                          char * err_name,
-                                         char * err_src)
+                                         char * err_src,
+                                         char const * name)
 {
   t_uchar * line;
   long len;
@@ -1259,11 +1368,12 @@ rel_read_pika_unescape_iso8859_1_record (int fd,
     return 0;
 
   answer = 0;
-  ar_setsize ((void **)&answer, 0, n_fields, sizeof (rel_field));
+  ar_setsize_ext_rel_record (&answer, n_fields, name);
 
   pos = line;
   for (f = 0; f < n_fields; ++f)
     {
+      t_uchar * temp_str;
 
       while (len && !char_is_space (*pos))
         {
@@ -1281,7 +1391,10 @@ rel_read_pika_unescape_iso8859_1_record (int fd,
             }
         }
 
-      answer[f] = pika_save_unescape_iso8859_1_n (0, 0, line, pos - line);
+      temp_str = pika_save_unescape_iso8859_1_n (0, 0, line, pos - line);
+      answer[f] = talloc_strdup (ar_base (answer), temp_str);
+      lim_free (0, temp_str);
+      
 
       while (len && char_is_space (*pos))
         {
@@ -1318,7 +1431,7 @@ rel_print_table (int fd, rel_table table)
   int recs;
   int r;
 
-  recs = ar_size ((void *)table, 0, sizeof (rel_record));
+  recs = ar_size_rel_table (table);
 
   for (r = 0; r < recs; ++r)
     rel_print_record (fd, table[r]);
@@ -1330,7 +1443,7 @@ rel_print_pika_escape_iso8859_1_table (int fd, int escape_classes, rel_table tab
   int recs;
   int r;
 
-  recs = ar_size ((void *)table, 0, sizeof (rel_record));
+  recs = ar_size_rel_table (table);
 
   for (r = 0; r < recs; ++r)
     rel_print_pika_escape_iso8859_1_record (fd, escape_classes, table[r]);
@@ -1351,7 +1464,7 @@ rel_print_table_sp (int fd, rel_table file)
   int recs;
   int r;
 
-  recs = ar_size ((void *)file, 0, sizeof (rel_record));
+  recs = ar_size_rel_table (file);
 
   for (r = 0; r < recs; ++r)
     rel_print_record_sp (fd, file[r]);
@@ -1363,7 +1476,7 @@ rel_print_pika_escape_iso8859_1_table_sp (int fd, int escape_classes, rel_table
   int recs;
   int r;
 
-  recs = ar_size ((void *)file, 0, sizeof (rel_record));
+  recs = ar_size_rel_table (file);
 
   for (r = 0; r < recs; ++r)
     rel_print_pika_escape_iso8859_1_record_sp (fd, escape_classes, file[r]);
@@ -1376,7 +1489,7 @@ rel_print_record (int fd, rel_record rec)
   int fields;
   int f;
 
-  fields = ar_size ((void *)rec, 0, sizeof (rel_field));
+  fields = ar_size_rel_record (rec);
 
   for (f = 0; f < fields; ++f)
     {
@@ -1393,7 +1506,7 @@ rel_print_pika_escape_iso8859_1_record (int fd, int escape_classes, rel_record r
   int fields;
   int f;
 
-  fields = ar_size ((void *)rec, 0, sizeof (rel_field));
+  fields = ar_size_rel_record (rec);
 
   for (f = 0; f < fields; ++f)
     {
@@ -1416,7 +1529,7 @@ rel_print_record_sp (int fd, rel_record rec)
   int fields;
   int f;
 
-  fields = ar_size ((void *)rec, 0, sizeof (rel_field));
+  fields = ar_size_rel_record (rec);
 
   for (f = 0; f < fields; ++f)
     {
@@ -1433,7 +1546,7 @@ rel_print_pika_escape_iso8859_1_record_sp (int fd, int escape_classes, rel_recor
   int fields;
   int f;
 
-  fields = ar_size ((void *)rec, 0, sizeof (rel_field));
+  fields = ar_size_rel_record (rec);
 
   for (f = 0; f < fields; ++f)
     {
@@ -1469,6 +1582,31 @@ rel_set_subtract (rel_table left, rel_table right)
   return answer;
 }
 
+/**
+ * \brief convert from a single list of strings to a reltable with 
+ * the supplied number of elements per row.
+ * count % width must be 0 or a panic will be caused.
+ * \return rel_table
+ * \param count the number of records in the suppplied array
+ * \param width how wide the output rel table should be.
+ * \para, strings the strings to convert.
+ */
+rel_table 
+rel_unflatten (int count, int width, char * strings[])
+{
+    rel_table result = NULL;
+    int position;
+    invariant (count % width == 0);
+    for (position = 0; position < count / width; ++position)
+      {
+	rel_record record = NULL;
+	int col;
+	for (col=0; col < width; ++col)
+	    rel_add_field (&record, strings[position * width + col]);
+	rel_add_record (&result, record);
+      }
+    return result;
+}
 
 
 
