@@ -15,7 +15,7 @@ srcdir=`dirname "$arg0"`
 
 setup_with_trivial_archives
 
-test_class "file-diffs"
+test_class "file-diffs (baz)"
 
 setup () {
   tla get jane@example.com--2003/hello-world--mainline--1.0 hello-world
@@ -32,8 +32,8 @@ do_file_diffs () {
 
 begin_test "diffs for unmodified file"
 setup
-do_file_diffs hello-world.c > ../hello-world.patch
-file_is_empty ../hello-world.patch
+do_file_diffs hello-world.c > ../hello-world.patch || test_fail "could not diff hello-world.c"
+file_is_empty ../hello-world.patch || test_fail "hello-world.c patch is not empty"
 cleanup
 end_test
 
