@@ -300,6 +300,9 @@ str_cmp (const t_uchar * a, const t_uchar * b)
   if ((!a || !*a) && (!b || !*b))
     return 0;
 
+  if (a == b)
+      return 0;
+
   if (!b || !*b)
     return 1;
 
@@ -337,6 +340,9 @@ str_cmp_n (const t_uchar * a, size_t a_l, const t_uchar * b, size_t b_l)
 {
   size_t l;
 
+  if (a == b && a_l == b_l)
+      return 0;
+
   l = ((a_l < b_l) ? a_l : b_l);
   while (l)
     {
@@ -375,6 +381,9 @@ str_cmp_prefix (const t_uchar * prefix, const t_uchar * s)
   if (!prefix)
     return 0;
 
+  if (prefix == s)
+    return 0;
+
   if (!s)
     {
       if (!*prefix)
@@ -418,6 +427,9 @@ str_cmp_prefix_n (const t_uchar * prefix,
   if (!s_len)
     return 1;
 
+  if (prefix == s && prefix_len <= s_len)
+      return 0;
+
   while (prefix_len && s_len)
     {
       if (*s < *prefix)
@@ -461,6 +473,9 @@ str_casecmp (const t_uchar * a, const t_uchar * b)
   if (!a || !*a)
     return -1;
 
+  if (a == b)
+    return 0;
+
   while ((char_to_lower(*a) == char_to_lower(*b)) && *a)
     {
       ++a;
@@ -489,6 +504,10 @@ str_casecmp_n (const t_uchar * a, size_t a_l,
 	       const t_uchar * b, size_t b_l)
 {
   size_t l;
+
+  if (a == b && a_l == b_l)
+    return 0;
+
   l = ((a_l < b_l) ? a_l : b_l);
   while (l)
     {
@@ -527,6 +546,9 @@ str_casecmp_prefix (const t_uchar * prefix, const t_uchar * s)
   if (!prefix)
     return 0;
 
+  if (prefix == s)
+      return 0;
+
   if (!s)
     {
       if (!*prefix)
@@ -571,6 +593,9 @@ str_casecmp_prefix_n (const t_uchar * prefix,
   if (!s_len)
     return 1;
 
+  if (prefix == s && prefix_len <= s_len)
+    return 0;
+
   while (prefix_len && s_len)
     {
       if (char_to_lower(*s) < char_to_lower (*prefix))
@@ -744,6 +769,8 @@ str_cat_n (t_uchar * to,
  *
  * Allocate storage according to `limits'.  (See xref:"Allocation
  * With Limitations".)
+ *
+ * if str is a NULL pointer, return NULL
  */
 t_uchar *
 str_save (alloc_limits limits, const t_uchar * str)
@@ -751,6 +778,8 @@ str_save (alloc_limits limits, const t_uchar * str)
   t_uchar * it;
   size_t len;
 
+  if (!str)
+      return NULL;
   len = str_length (str);
   it = (t_uchar *)lim_malloc (limits, len + 1);
   if (!it)
@@ -970,6 +999,20 @@ str_occurences (t_uchar * s, int c)
   return answer;
 }
 
+/**
+ * \brief free current and return replacement
+ *
+ * Factors out blocks of temporary variables into one shot.
+ * \return replacement
+ * \param current the string to be freed.
+ * \param replacement the string to be returned
+ */
+t_uchar *
+str_replace (t_uchar *current, t_uchar *replacement)
+{
+    lim_free (0, current);
+    return replacement;
+}
 
 /*(include-documentation "str-many.c")
  */
