@@ -10,9 +10,11 @@
 
 
 #include "hackerlab/os/alloca.h"
+#include "hackerlab/os/errno.h"
 #include "hackerlab/os/stdlib.h"
 #include "hackerlab/os/pwd.h"
 #include "hackerlab/bugs/panic.h"
+#include "hackerlab/bugs/exception.h"
 #include "hackerlab/char/str.h"
 #include "hackerlab/mem/mem.h"
 #include "hackerlab/arrays/ar.h"
@@ -235,7 +237,7 @@ file_name_tilde_expand (alloc_limits limits, t_uchar * fname)
  * If allocation fails (but does not call `panic') this function returns 0.
  */
 t_uchar *
-file_name_tail (alloc_limits limits, t_uchar * fname)
+file_name_tail (alloc_limits limits, t_uchar const * const fname)
 {
   t_uchar * s;
 
@@ -457,12 +459,12 @@ file_name_directory (alloc_limits limits, t_uchar * f)
  * 
  */
 t_uchar *
-file_name_directory_file (alloc_limits limits, t_uchar * f)
+file_name_directory_file (alloc_limits limits, t_uchar const * f)
 {
   t_uchar * intermediate;
   t_uchar * answer;
 
-  intermediate = file_name_directory (limits, f);
+  intermediate = file_name_directory (limits, (t_uchar *)f);
   if (!intermediate)
     return 0;
 
@@ -508,13 +510,13 @@ file_name_is_path_prefix (t_uchar * maybe_prefix, t_uchar * path)
 }
 
 
-/*(c file_name_in_vicinity)
- * t_uchar * file_name_in_vicinity (alloc_limits limits,
+/*(c talloc_file_name_in_vicinity)
+ * t_uchar * file_name_in_vicinity (void * context,
  * 			  	    t_uchar const * dir,
  * 			  	    t_uchar const * file);
  * 
- * Return a newly allocated string which is `dir', followed by `"/"',
- * followed by `file'.  If `dir' is a directory name and `file' is a
+ * Return a newly allocated string (in context) which is `dir', followed by 
+ * `"/"', * followed by `file'.  If `dir' is a directory name and `file' is a
  * relative file-name, then the result is a name for `file', relative
  * to `dir'.  If `dir' already ends in `"/"', no additional `"/"' is
  * added.
@@ -530,10 +532,11 @@ file_name_is_path_prefix (t_uchar * maybe_prefix, t_uchar * path)
  *	"/etc", "/usr/bin/ls"	=>	"/usr/bin/ls"
  *	0, "ls"			=>	"ls"
  *
- * If allocation fails (but does not call `panic') this function returns 0.
+ * If allocation fails (but does not call `panic') this function attempts to
+ * throw an exception with code ENOMEM.
  */
 t_uchar *
-file_name_in_vicinity (alloc_limits limits,
+talloc_file_name_in_vicinity (void * context,
 		       t_uchar const * dir,
 		       t_uchar const * name)
 {
@@ -542,7 +545,10 @@ file_name_in_vicinity (alloc_limits limits,
   t_uchar * answer;
 
   if (name && (*name == '/'))
-    return str_save (limits, name);
+    {
+      answer = talloc_strdup (context, name);
+      goto done;
+    }
 
   dl = str_length (dir);
   if (dl && (dir[dl - 1] == '/'))
@@ -552,9 +558,9 @@ file_name_in_vicinity (alloc_limits limits,
       ++dl;
     }
   nl = str_length (name);
-  answer = (t_uchar *)lim_malloc (limits, dl + nl + 2);
+  answer = talloc_size(context, dl + nl + 2);
   if (!answer)
-    return 0;
+    goto done;
   if (dir && dir[0])
     {
       mem_move (answer, dir, dl);
@@ -567,9 +573,51 @@ file_name_in_vicinity (alloc_limits limits,
       mem_move (answer, name, nl);
       answer[nl] = 0;
     }
+done:
+  if (!answer)
+      Throw (exception (ENOMEM, "Could not allocate result"));
+  talloc_set_name_const(answer, answer);
   return answer;
 }
 
+/*(c file_name_in_vicinity)
+ * t_uchar * file_name_in_vicinity (alloc_limits limits,
+ * 			  	    t_uchar const * dir,
+ * 			  	    t_uchar const * file);
+ * 
+ * Return a str allocated from limits, from the result of
+ * talloc_file_name_in_vicinity (talloc_context, dir, file);
+ *
+ * If allocation fails (but does not call `panic') this function returns 0.
+ * If an ENONMEM exception is raised, it is caught and NULL returned.
+ * All other exceptions are rethrown.
+ * 
+ */
+t_uchar *
+file_name_in_vicinity (alloc_limits limits,
+		       t_uchar const * dir,
+		       t_uchar const * name)
+{
+  t_uchar * result = NULL;
+  struct exception * e;
+  Try 
+    {
+      t_uchar * tmp = talloc_file_name_in_vicinity (talloc_context,
+						    dir,
+						    name);
+      result = str_save (0, tmp);
+    }
+  Catch (e)
+    {
+      if (e->code != ENOMEM)
+	  Throw (e);
+      talloc_free (e);
+      return NULL;
+    }
+  return result;
+}
+
+
 
 /************************************************************************
  *(h1 "Paths")
@@ -610,47 +658,47 @@ path_parse (alloc_limits limits, t_uchar * path)
     {
       if (*path == ':')
 	{
-	  loc = (t_uchar **)ar_push (&answer.void_ptr, limits, sizeof (t_uchar *));
+	  t_uchar *temp = str_save (limits, ".");
+	  if (!temp)
+	    goto espace_error;
+	  loc = ar_push_uchar_star(&answer.uchar_ptr_ptr, temp);
 	  if (!loc)
 	    {
 	      int x;
 	    espace_error:
 
-	      x = ar_size (answer.void_ptr, limits, sizeof (*answer.uchar_ptr_ptr));
+	      x = ar_size (answer.void_ptr);
 	      while (x--)
 		{
 		  lim_free (limits, answer.uchar_ptr_ptr[x]);
 		}
-	      ar_free (&answer.void_ptr, limits);
+	      ar_free_uchar_star (&answer.uchar_ptr_ptr);
 	      return 0;
 	    }
-	  *loc = (t_uchar *)str_save (limits, ".");
-	  if (!*loc)
-	    goto espace_error;
 	  ++path;
 	}
       else
 	{
 	  t_uchar * end;
+	  t_uchar * temp;
 
 	  end = str_chr_index (path, ':');
 	  if (!end)
 	    end = path + str_length (path);
-	  loc = (t_uchar **)ar_push (&answer.void_ptr, lim_use_must_malloc, sizeof (t_uchar *));
-	  if (!loc)
+	  temp = str_save_n (lim_use_must_malloc, path, end - path);
+	  if (!temp)
 	    goto espace_error;
-	  *loc = (t_uchar *)str_save_n (lim_use_must_malloc, path, end - path);
-	  if (!*loc)
+	  loc = ar_push_uchar_star (&answer.uchar_ptr_ptr, temp);
+	  if (!loc)
 	    goto espace_error;
 	  path = end;
 	  if ((*path == ':') && (*(path + 1)))
 	    ++path;
 	}
     }
-  loc = (t_uchar **)ar_push (&answer.void_ptr, lim_use_must_malloc, sizeof (t_uchar *));
+  loc = ar_push_uchar_star (&answer.uchar_ptr_ptr, NULL);
   if (!loc)
     goto espace_error;
-  *loc = 0;
   return answer.uchar_ptr_ptr;
 }
 
@@ -665,15 +713,13 @@ free_path (alloc_limits limits, t_uchar ** path)
 {
   if (path)
     {
-      t_converter aPath;
       int x;
-
-      aPath.uchar_ptr_ptr = path;
+      t_uchar **aPath = path;
 
       for (x = 0; path[x]; ++x)
 	lim_free (limits, path[x]);
 
-      ar_free (&aPath.void_ptr, limits);
+      ar_free_uchar_star (&aPath);
     }
 }
 
