@@ -858,6 +858,8 @@ struct server_memo
   char * path;                          /* If this is a unix domain socket, the file name to unlink */
 };
 
+AR_TYPEDEF (struct server_memo, server_memo);
+
 static union
 {
   struct server_memo * smp;
@@ -910,7 +912,7 @@ server_fd_for_addr (int * errn,
   int desired_flags;
   struct server_memo * memo;
 
-  size = ar_size (addr_memo.void_ptr, lim_use_must_malloc, sizeof (*addr_memo.smp));
+  size = ar_size (addr_memo.void_ptr);
   for (x = 0; x < size; ++x)
     {
       if (addr->sa_family == ADDR_MEMO(x)->sa_family)
@@ -975,6 +977,7 @@ server_fd_for_addr (int * errn,
 
   {
     int orig_flags;
+    struct server_memo temp;
     orig_flags = fcntl (fd, F_GETFL, 0);
     if (orig_flags < 0)
       {
@@ -982,14 +985,14 @@ server_fd_for_addr (int * errn,
 	close (fd);
 	return -1;
       }
-    memo = (struct server_memo *)ar_push (&addr_memo.void_ptr, lim_use_must_malloc, sizeof (struct server_memo));
-    mem_move (memo->addr_data, (t_uchar *)addr, addr_len);
-    memo->fd = fd;
-    memo->flags = orig_flags & O_NONBLOCK;
+    temp.fd = fd;
+    temp.flags = orig_flags & O_NONBLOCK;
     if (addr->sa_family == AF_INET)
-      memo->path = 0;
+      temp.path = 0;
     else
-      memo->path = str_save (lim_use_must_malloc, UN_ADDR (addr)->sun_path);
+      temp.path = str_save (lim_use_must_malloc, UN_ADDR (addr)->sun_path);
+    memo = ar_push_server_memo (&addr_memo.smp, temp);
+    mem_move (memo->addr_data, (t_uchar *)addr, addr_len);
   }
   
   vu_set_fd_handler (fd, &url_server_socket_vtable, (void *)params);
@@ -1022,7 +1025,7 @@ find_memo_for_server_fd (int fd)
   int size;
   int x;
 
-  size = ar_size (addr_memo.void_ptr, lim_use_must_malloc, sizeof (*addr_memo.smp));
+  size = ar_size (addr_memo.void_ptr);
   for (x = 0; x < size; ++x)
     {
       if (addr_memo.smp[x].fd == fd)
@@ -1043,13 +1046,13 @@ remove_memo_for_server_fd (int fd)
   int size;
   int x;
 
-  size = ar_size (addr_memo.void_ptr, lim_use_must_malloc, sizeof (*addr_memo.smp));
+  size = ar_size (addr_memo.void_ptr);
   for (x = 0; x < size; ++x)
     {
       if (addr_memo.smp[x].fd == fd)
 	{
 	  addr_memo.smp[x] = addr_memo.smp[size - 1];
-	  ar_pop (&addr_memo.void_ptr, lim_use_must_malloc, sizeof (struct server_memo));
+	  ar_pop_server_memo (&addr_memo.smp);
 	  return;
 	}
     }
@@ -1659,14 +1662,14 @@ url_socket_close (int * errn, int fd, void * closure)
   int x;
   int size;
 
-  size = ar_size (addr_memo.void_ptr, lim_use_must_malloc, sizeof (*addr_memo.smp));
+  size = ar_size (addr_memo.void_ptr);
   for (x = 0; x < size; ++x)
     if (fd == addr_memo.smp [x].fd)
       {
 	if (0 > vu_sys_close (errn, fd, 0))
 	  return -1;
 	addr_memo.smp[x] = addr_memo.smp[size - 1];
-	ar_pop (&addr_memo.void_ptr, lim_use_must_malloc, sizeof (struct server_memo));
+	ar_pop_server_memo (&addr_memo.smp);
 	return 0;
       }
   return vu_sys_close (errn, fd, 0);
