@@ -13,6 +13,7 @@
 #include "hackerlab/bugs/panic.h"
 #include "hackerlab/mem/mem.h"
 #include "hackerlab/mem/must-malloc.h"
+#include "hackerlab/mem/talloc.h"
 #include "hackerlab/arrays/ar.h"
 #include "hackerlab/vu/vu-sys.h"
 #include "hackerlab/vu/vfdbuf.h"
@@ -196,7 +197,7 @@ vfdbuf_flush_all(void)
   int x;
   int size;
 
-  size = ar_size (bufs.void_ptr, lim_use_must_malloc, sizeof (*bufs.vfd));
+  size = ar_size (bufs.void_ptr);
   for (x = 0; x < size; ++x)
     {
       if (bufs.vfd[x].sub_handler.vtable)
@@ -262,7 +263,7 @@ vfdbuf_buffer_fd (int * errn, int fd, long bufsize, int flags, int buffer_flags)
   if (bufsize <= 0)
     bufsize = 4096;
 
-  if ((ar_size (bufs.void_ptr, lim_use_must_malloc, sizeof (*bufs.vfd)) > fd) && bufs.vfd[fd].sub_handler.vtable)
+  if ((ar_size (bufs.void_ptr) > fd) && bufs.vfd[fd].sub_handler.vtable)
     {
       int had_zero_byte;
 
@@ -323,8 +324,10 @@ vfdbuf_buffer_fd (int * errn, int fd, long bufsize, int flags, int buffer_flags)
 	}
     }
 
-  if (fd >= ar_size (bufs.void_ptr, lim_use_must_malloc, sizeof (*bufs.vfd)))
-    ar_ref (&bufs.void_ptr, lim_use_must_malloc, fd, sizeof (struct vfdbuf_record));
+  if (fd >= ar_size (bufs.void_ptr))
+    {
+      ar_ref (&bufs.void_ptr, fd, sizeof (struct vfdbuf_record));
+    }
 
   bufs.vfd[fd].sub_handler = *sub_handler;
   bufs.vfd[fd].flags = O_ACCMODE & flags;
@@ -374,7 +377,7 @@ vfdbuf_set_buffer (int * errn,
 		   void (*free_buffer)(t_uchar * buf, void * closure),
 		   void * free_buffer_closure)
 {
-  if (!((ar_size (bufs.void_ptr, lim_use_must_malloc, sizeof (*bufs.vfd)) > fd) && bufs.vfd[fd].sub_handler.vtable))
+  if (!((ar_size (bufs.void_ptr) > fd) && bufs.vfd[fd].sub_handler.vtable))
     {
       /* not buffered */
     invalid:
@@ -433,7 +436,7 @@ vfdbuf_unbuffer_fd (int * errn, int fd)
 int
 vfdbuf_is_buffered (int fd)
 {
-  if ((ar_size (bufs.void_ptr, lim_use_must_malloc, sizeof (*bufs.vfd)) > fd) && bufs.vfd[fd].sub_handler.vtable)
+  if ((ar_size (bufs.void_ptr) > fd) && bufs.vfd[fd].sub_handler.vtable)
     return 1;
   return 0;
 }
@@ -1404,7 +1407,7 @@ vfdbuf_move_state (int * errn, int fd, int newfd, void * closure)
 
   /* Move the buffer to the new descriptor.
    */
-  ar_ref (&bufs.void_ptr, lim_use_must_malloc, newfd, sizeof (struct vfdbuf_record));
+  ar_ref (&bufs.void_ptr, newfd, sizeof (struct vfdbuf_record));
   bufs.vfd[newfd] = bufs.vfd[fd];
   
   /* Remove the buffer from the old descriptor.
